/*======================================================================

  This file is part of the elastix software.

  Copyright (c) University Medical Center Utrecht. All rights reserved.
  See src/CopyrightElastix.txt or http://elastix.isi.uu.nl/legal.php for
  details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE. See the above copyright notices for more information.

======================================================================*/

#ifndef __elxGeneticAlgorithm_h
#define __elxGeneticAlgorithm_h

#include "itkGeneticAlgorithmOptimizer.h"
#include "elxIncludes.h"

namespace elastix
{
using namespace itk;

  /**
   * \class GeneticAlgorithm
   * \brief An optimizer based on the itk::GeneticAlgorithmOptimizer.
   *
   * This class is a wrap around the GeneticAlgorithmOptimizer class.
   * It takes care of setting parameters and printing progress information.
   * For more information about the optimisation method, please read the documentation
   * of the GeneticAlgorithmOptimizer class.
   *
   * The parameters used in this class are:
   * \parameter Optimizer: Select this optimizer as follows:\n
   *    <tt>(Optimizer "GeneticAlgorithm")</tt>
   * \parameter MaximumNumberOfIterations: The maximum number of iterations in each resolution. \n
   *    example: <tt>(MaximumNumberOfIterations 100 100 50)</tt> \n
   *    Default value: 50.

   *
   * \ingroup Optimizers
   * \sa GeneticAlgorithmOptimizer
   */

  template <class TElastix>
    class GeneticAlgorithm :
    public
      itk::GeneticAlgorithmOptimizer,
    public
      OptimizerBase<TElastix>
  {
  public:

    /** Standard ITK.*/
    typedef GeneticAlgorithm           Self;
    typedef GeneticAlgorithmOptimizer  Superclass1;
    typedef OptimizerBase<TElastix>                   Superclass2;
    typedef SmartPointer<Self>                        Pointer;
    typedef SmartPointer<const Self>                  ConstPointer;

    /** Method for creation through the object factory. */
    itkNewMacro( Self );

    /** Run-time type information (and related methods). */
    itkTypeMacro( GeneticAlgorithm, GeneticAlgorithmOptimizer );

    /** Name of this class.
     * Use this name in the parameter file to select this specific optimizer. \n
     * example: <tt>(Optimizer "GeneticAlgorithm")</tt>\n
     */
    elxClassNameMacro( "GeneticAlgorithm" );

    /** Typedef's inherited from Superclass1.*/
    typedef Superclass1::CostFunctionType     CostFunctionType;
    typedef Superclass1::CostFunctionPointer  CostFunctionPointer;
    typedef Superclass1::StopConditionType    StopConditionType;

    /** Typedef's inherited from Elastix.*/
    typedef typename Superclass2::ElastixType           ElastixType;
    typedef typename Superclass2::ElastixPointer        ElastixPointer;
    typedef typename Superclass2::ConfigurationType     ConfigurationType;
    typedef typename Superclass2::ConfigurationPointer  ConfigurationPointer;
    typedef typename Superclass2::RegistrationType      RegistrationType;
    typedef typename Superclass2::RegistrationPointer   RegistrationPointer;
    typedef typename Superclass2::ITKBaseType           ITKBaseType;

    /** Typedef for the ParametersType. */
    typedef typename Superclass1::ParametersType        ParametersType;

    /** Methods that take care of setting parameters and printing progress information.*/
    virtual void BeforeRegistration(void);
    virtual void BeforeEachResolution(void);
    virtual void AfterEachResolution(void);
    virtual void AfterEachIteration(void);
    virtual void AfterRegistration(void);

    /** Check if any scales are set, and set the UseScales flag on or off;
     * after that call the superclass' implementation */
    virtual void StartOptimization(void);

  protected:

      GeneticAlgorithm();
      virtual ~GeneticAlgorithm() {};

  private:

      GeneticAlgorithm( const Self& ); // purposely not implemented
      void operator=( const Self& );              // purposely not implemented

  }; // end class GeneticAlgorithm


} // end namespace elastix

#ifndef ITK_MANUAL_INSTANTIATION
#include "elxGeneticAlgorithm.hxx"
#endif

#endif // end #ifndef __elxGeneticAlgorithm_h

