/*======================================================================

  This file is part of the elastix software.

  Copyright (c) University Medical Center Utrecht. All rights reserved.
  See src/CopyrightElastix.txt or http://elastix.isi.uu.nl/legal.php for
  details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE. See the above copyright notices for more information.

======================================================================*/

#ifndef __itkGeneticAlgorithmOptimizer_cxx
#define __itkGeneticAlgorithmOptimizer_cxx

#include "itkGeneticAlgorithmOptimizer.h"
#include "itkCommand.h"
#include "itkEventObject.h"
#include "itkExceptionObject.h"
#include "itkRealTimeClock.h"
#include "itkTimeStamp.h"

const double identity_similarity[] = {0,0,0,0,0,0,1};
const double identity_affine[] = {1,0,0,0,1,0,0,0,1,0,0,0};

namespace itk
{


  /**
   * ************************* Constructor ************************
   */

  GeneticAlgorithmOptimizer
    ::GeneticAlgorithmOptimizer()
  {
    itkDebugMacro( "Constructor" );

    /** Default values **/
    this->m_NumberOfIterations = 50;
    this->m_CurrentIteration = 0;
    this->m_CurrentResolution = 0;
    this->m_Value = 0.0;
    this->m_StopCondition = MaximumNumberOfIterations;
    this->m_PopulationSize = 100;
    this->m_CrossoverProbability = 0.7;
    this->m_MutationProbability = 0.05;
    this->m_TournamentSize = 3;
    this->m_Restarts = 0;
    this->m_RestartsPerformed = 0;

    RealTimeClock::Pointer realTimeClock = RealTimeClock::New();
    RealTimeClock::TimeStampType stamp = realTimeClock->GetTimeStamp();

    this->m_RandomGenerator = RandomGeneratorType::New();
    this->m_RandomGenerator->Initialize( static_cast<uint32_t>(stamp*1000) );

    this->m_Population = new PopulationType();
    
    this->m_StoredPopulation = NULL;
    this->m_StoredBestIndividual = NULL;
    this->m_BestIndividual = NULL;

  } // end Constructor


  /**
   * ************************* Destructor ************************
   */

  GeneticAlgorithmOptimizer
    ::~GeneticAlgorithmOptimizer()
  {
    itkDebugMacro( "Destructor" );

    this->ClearPopulation();
    delete this->m_Population;

  } // end Destructor


  /**
   * ************************* PrintSelf **************************
   */

  void
    GeneticAlgorithmOptimizer
    ::PrintSelf( std::ostream& os, Indent indent ) const
  {
    Superclass::PrintSelf( os, indent );

    os << indent << "NumberOfIterations: "
      << this->m_NumberOfIterations << std::endl;
    os << indent << "CurrentIteration: "
      << this->m_CurrentIteration;
    os << indent << "Value: "
      << this->m_Value;
    os << indent << "StopCondition: "
      << this->m_StopCondition;
    os << std::endl;

  } // end PrintSelf


  /**
   * *********************** StartOptimization ********************
   */
  void
    GeneticAlgorithmOptimizer
    ::StartOptimization(void)
  {
    itkDebugMacro( "StartOptimization" );

    this->m_CurrentIteration = 0;
    this->m_Stop = false;

    /** Get the number of parameters; checks also if a cost function has been set at all.
     * if not: an exception is thrown */
    this->m_NumberOfParameters = this->GetScaledCostFunction()->GetNumberOfParameters();

    /** Initialize the scaledCostFunction with the currently set scales */
    this->InitializeScales();

    /** Set the current position as the scaled initial position */
    this->SetCurrentPosition( this->GetInitialPosition() );

    /** Fill the population with random individuals or prepare the population to be used for
     * the current resolution.
     */

    if (this->m_RestartsPerformed > 0)
    {
      std::cout << "DYNAMIC BOUNDARIES" << std::endl;
      // Dynamic boundaries
      // During the second resolution transformation parameters range are restricted around the 
      // best solutions
      double b, l, h;
      for( unsigned int i = 0; i < this->m_NumberOfParameters; i++)
      {
        b = this->m_BestIndividual->genome[i];
        l = this->m_ParameterRange[i][0];
        h = this->m_ParameterRange[i][1]; 

        this->m_ParameterRange[i][0] = b -(b-l)/10;
        this->m_ParameterRange[i][1] = b +(h-b)/10;
      }
    }
    this->InitPopulation();

    if ( !this->m_Stop )
    {
      this->ResumeOptimization();
    }

  } // end StartOptimization


  /**
   * ********************** ResumeOptimization ********************
   */

  void
    GeneticAlgorithmOptimizer
    ::ResumeOptimization( void )
  {

    itkDebugMacro( "ResumeOptimization" );

    this->m_Stop = false;

    InvokeEvent( StartEvent() );
    while( ! this->m_Stop )
    {
      this->AdvanceOneStep();
      this->m_CurrentIteration++;

      if( this->m_CurrentIteration >= this->m_NumberOfIterations )
      {
        /** RESTART 
            At the end of the first resolution, check if the
            best individual's fitness is lower than the threshold.
        **/
        if( ( this->m_CurrentResolution == 0 ) && ( this->m_Restarts > 0 ) && 
            ( this->m_Value > this->m_RestartThreshold ) ) 
        {
          /** Restart */
          this->m_CurrentIteration = 0;
          this->m_Restarts--;
          this->m_RestartsPerformed++;
          
          std::cout << "Restart" << std::endl;

          // If the current solutions are better than the best we stored, 
          // replaced the stored ones
          if( ( this->m_StoredBestIndividual == NULL) || 
              ( this->m_Value < this->m_StoredBestIndividual->fitness ) )
          {
	    std::cout << "Save" << std::endl;
            if ( this->m_StoredPopulation != NULL )
            {
              while( !this->m_StoredPopulation->empty() )
              {
                delete this->m_StoredPopulation->back();
                this->m_StoredPopulation->pop_back();
              }
              delete this->m_StoredPopulation;
              delete this->m_StoredBestIndividual;
            }
            this->m_StoredPopulation =  this->m_Population;
            this->m_Population = new PopulationType( this->m_PopulationSize);
            this->m_StoredBestIndividual = this->m_BestIndividual->deep_copy();
          }

          /** Replace the individuals in the current population */
          this->ClearPopulation();
          this->m_BestIndividual = NULL;
          this->InitPopulation();
        }
        else
        {
          //If we ran out of restarts, replace the stored solutions
          //unless we got better results in the last try
          if ( ( this->m_CurrentResolution == 0 ) && ( this->m_Restarts == 0 ) 
               && (this->m_RestartsPerformed > 0) && ( this->m_Value > this->m_StoredBestIndividual->fitness ) )
          {
            std::cout << "Restore" << std::endl;
            this->ClearPopulation();
            this->m_Population = this->m_StoredPopulation;
            this->m_BestIndividual = this->m_StoredBestIndividual->deep_copy();
	    this->SetScaledCurrentPosition( this->m_BestIndividual->deep_copy()->genome );
            this->m_Value = this->m_BestIndividual->fitness;
          }

          this->m_StopCondition = MaximumNumberOfIterations;
          StopOptimization();
          break;
        }
      }
    } // while !m_stop
  } // end ResumeOptimization


  /**
   * ********************** StopOptimization **********************
   */

  void
    GeneticAlgorithmOptimizer
    ::StopOptimization( void )
  {
    itkDebugMacro( "StopOptimization" );

    this->m_Stop = true;
    InvokeEvent( EndEvent() );

  } // end StopOptimization


  /**
   * ********************** AdvanceOneStep ************************
   */

  void
    GeneticAlgorithmOptimizer
    ::AdvanceOneStep( void )
  {
    itkDebugMacro( "AdvanceOneStep" );

    PopulationType* newPopulation = new PopulationType();
    newPopulation->reserve( this->m_PopulationSize );
    
    /** Elitism. */
    newPopulation->push_back( this->m_BestIndividual->deep_copy() );

    while ( newPopulation->size() < this->m_PopulationSize )
    {
      Individual* parentAlpha = this->Selection();
      Individual* parentBeta = this->Selection();

      Individual* offspringAlpha; 
      Individual* offspringBeta; 

      if ( this->m_RandomGenerator->GetUniformVariate(0,1) <
        this->m_CrossoverProbability )  //Crossover
      {
        this->Crossover( parentAlpha, parentBeta, offspringAlpha, offspringBeta );
      }
      else //Copy
      {
        offspringAlpha = parentAlpha->deep_copy();
        offspringBeta = parentBeta->deep_copy();
      }

      // // /** Mutation. */
      this->Mutation( offspringAlpha );
      this->EvaluateIndividual( offspringAlpha );
      newPopulation->push_back( offspringAlpha );

      /** As we add two individuals at the time, we may exceed
       * the population size.
       */
      if ( newPopulation->size() < this->m_PopulationSize )
      {
        this->Mutation( offspringBeta );
        this->EvaluateIndividual( offspringBeta );
        newPopulation->push_back( offspringBeta );
      }
      else
      {
        delete offspringBeta;
      }
    }

    this->ClearPopulation();
    delete this->m_Population;

    this->m_Population = newPopulation;

    this->SetScaledCurrentPosition( this->m_BestIndividual->deep_copy()->genome );
    // std::cout << this->m_BestIndividual->genome;
    this->InvokeEvent( IterationEvent() );

  } // end AdvanceOneStep


  /**
   * ********************** EvaluateIndividual ************************
   */

  void
    GeneticAlgorithmOptimizer
  ::EvaluateIndividual( Individual* indiv , bool mandatory )
  {
    itkDebugMacro( "EvaluateIndividual" );  

    if ( !indiv->evaluated || mandatory)
    {
      try
      {
        indiv->fitness = this->GetScaledValue( indiv->genome );
      }
      catch( ExceptionObject& err )
      {
        // Terminate immediately.
        this->m_StopCondition = MetricError;
        StopOptimization();
        throw err;
      }
      
      indiv->evaluated = true;
      
      if ( ( this->m_BestIndividual == NULL ) || 
           ( indiv->fitness < this->m_BestIndividual->fitness ) )
      {
        delete this->m_BestIndividual;
        this->m_BestIndividual = indiv->deep_copy();
        this->m_Value = indiv->fitness;
      }
    }
  } // End EvaluateIndividual


  /**
   * ********************** InitPopulation ************************
   */

  void
    GeneticAlgorithmOptimizer
    ::InitPopulation( void )
  {
    itkDebugMacro( "InitPopulation" );

    /** There are three cases: 
     * 1. If the first resolution (possibly after a restart), then the population is empty
     *    and we want to create a new population. 
     * 2. Else, for all transforms but B-Spline we want to keep the current population,
     *    we just need to evaluate the fitness again.  
     * 3. Finally For B-Spline we need a fresh population anyway.
     */

    if ( this->m_NumberOfParameters > 12) 
      this->ClearPopulation();
    
    if ( this->m_Population->empty() )
    {
      if ( this->m_BestIndividual != NULL )
      {
        delete this->m_BestIndividual;
        this->m_BestIndividual = NULL;
      }
      this->m_Population->reserve( this->m_PopulationSize );

      Individual* indiv = new Individual( this->m_NumberOfParameters );

      /** Adds a special, "null transformation" individual, if allowed by the parameter ranges
        * Determine the type of transformation using the number of parameters
        * In 3D, similarity has 7, affine has 12 and B-Spline has a lot (i.e. > 20)
        */
      unsigned int start = 0;
      bool known_transform = true;
      bool b_splines = false;
      const double* id_transform = NULL;

      switch( this->m_NumberOfParameters )
      {
      case 7:
        id_transform = identity_similarity;
        break;
      case 12:
        id_transform = identity_affine;
        break;
      default:  
        if ( this->m_NumberOfParameters > 12) 
          b_splines = true;
        else
          known_transform = false;
      }

      if (known_transform)
      {
        unsigned int j;
        if (b_splines)
        {
          //Check the identity is OK with the parameter ranges

          for ( j = 0; j < this->m_NumberOfParameters; j++ )
          {
            if ( ( 0 < this->m_ParameterRange[ j ][ 0 ] ) || 
                 (  0 > this->m_ParameterRange[ j ][ 1 ]) )
              break;
            else
              indiv->genome[ j ] = 0;
          }
        }
        else
        {
          //Check the identity is OK with the parameter ranges
          for ( j = 0; j < this->m_NumberOfParameters; j++ )
          {
            if ( ( id_transform[ j ] < this->m_ParameterRange[ j ][ 0 ] ) || 
                 ( id_transform[ j ] > this->m_ParameterRange[ j ][ 1 ]) )
              break;
            else
              indiv->genome[ j ] = id_transform[ j ];
          }        
        }

          if ( j == this->m_NumberOfParameters )
          {
            this->EvaluateIndividual( indiv );
            this->m_Population->push_back( indiv );
            start = 1;
          }
          else
            delete indiv;
      }

      for ( unsigned int i = start; i < this->m_PopulationSize; i++ )
      {
        indiv = new Individual( this->m_NumberOfParameters );

        for ( unsigned int j = 0; j < this->m_NumberOfParameters; j++ )
        {
          indiv->genome[ j ] = this->m_RandomGenerator->GetUniformVariate( 
            this->m_ParameterRange[ j ][ 0 ], this->m_ParameterRange[ j ][ 1 ] );
        }

        this->EvaluateIndividual( indiv );
        this->m_Population->push_back( indiv );
      }
    }
    else
    {
      delete this->m_BestIndividual;
      this->m_BestIndividual = NULL;
      for( unsigned int i = 0; i < this->m_PopulationSize; i++ )
      {
        /** Evaluate the individuals ignoring the "evaluated" flag because the fitness function has changed **/
        this->EvaluateIndividual( this->m_Population->at( i ), true );
      }
    }
  } // end InitPopulation


  /**
   * ********************** Selection ************************
   */

  GeneticAlgorithmOptimizer::Individual*
    GeneticAlgorithmOptimizer
    ::Selection( void )
  {
    itkDebugMacro( "Selection" );

    Individual* champion = this->m_Population->at(
      this->m_RandomGenerator->GetIntegerVariate( this->m_PopulationSize -1 ) );

    Individual* contestant;
    for ( unsigned int i = 0; i < this->m_TournamentSize; i++)
    {
      contestant = this->m_Population->at (
        this->m_RandomGenerator->GetIntegerVariate( this->m_PopulationSize -1 ) );

      if ( contestant->fitness < champion->fitness  )
        champion = contestant;
    }

    return champion;

  } // end Selection


  /**
   * ********************** Crossover ************************
   */

  void
    GeneticAlgorithmOptimizer
  ::Crossover( Individual* const parA, Individual* const parB,
               Individual* & offA, Individual* & offB )
  {
    itkDebugMacro( "Crossover" );

    float alpha = 0.3;
    double cMax, cMin, I;

    offA = new Individual( this->m_NumberOfParameters );
    offB = new Individual( this->m_NumberOfParameters );

    for ( unsigned int i = 0; i < this->m_NumberOfParameters; i++ )
    {
      if (parA->genome[ i ] > parB->genome[ i ])
      {
        cMax = parA->genome[ i ];
        cMin = parB->genome[ i ];
      }
      else
      {
        cMin = parA->genome[ i ];
        cMax = parB->genome[ i ];
      }
      I = cMax - cMin;

      //Enforce the parameters to be in the appropriate range
      double u,l;

      l = std::max(cMin - I * alpha, this->m_ParameterRange[ i ][ 0 ]);
      u = std::min(cMax + I * alpha, this->m_ParameterRange[ i ][ 1 ]);

      offA->genome[ i ] = this->m_RandomGenerator->GetUniformVariate( l, u );
      offB->genome[ i ] = this->m_RandomGenerator->GetUniformVariate( l, u );
    }

    offA->evaluated = false;
    offB->evaluated = false;

  } // end Crossover


  /**
   * ********************** Mutation ************************
   */

  void
    GeneticAlgorithmOptimizer
  ::Mutation( Individual* ind )
  {
    itkDebugMacro( "Mutation" );

    for (unsigned int i = 0; i < this->m_NumberOfParameters; i++)
    {
      if ( this->m_RandomGenerator->GetUniformVariate(0,1) <
           this->m_MutationProbability )
      {
        ind->genome[ i ] = this->m_RandomGenerator->GetUniformVariate(
          this->m_ParameterRange[ i ][ 0 ], this->m_ParameterRange[ i ][ 1 ] );
      }
      ind->evaluated = false;
    }
    
  } // end Mutation


  /**
   * ********************** ClearPopulation ************************
   */

  void
    GeneticAlgorithmOptimizer
  ::ClearPopulation()
  {
    itkDebugMacro( "ClearPopulation" );

    while( !this->m_Population->empty() )
    {
    	delete this->m_Population->back();
    	this->m_Population->pop_back();
    }
  }

} // end namespace itk


#endif // end #ifndef __itkGeneticAlgorithmOptimizer_cxx
