/*======================================================================

  This file is part of the elastix software.

  Copyright (c) University Medical Center Utrecht. All rights reserved.
  See src/CopyrightElastix.txt or http://elastix.isi.uu.nl/legal.php for
  details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE. See the above copyright notices for more information.

======================================================================*/

#ifndef __itkGeneticAlgorithmOptimizer_h
#define __itkGeneticAlgorithmOptimizer_h

#include "itkScaledSingleValuedNonLinearOptimizer.h"
#include "itkThreadSafeMersenneTwisterRandomVariateGenerator.h"
#include "itkArray2D.h"

namespace itk
{

  /**
   * \class GeneticAlgorithmOptimizer
   * \brief An optimizer based on Genetic Algorithm
   *
   * bla bla
   *
   * \ingroup Optimizers
   * \sa GeneticAlgorithm
   */

  class GeneticAlgorithmOptimizer
    : public ScaledSingleValuedNonLinearOptimizer
  {
  public:

    /** Standard class typedefs. */
    typedef GeneticAlgorithmOptimizer    Self;
    typedef ScaledSingleValuedNonLinearOptimizer        Superclass;
    typedef SmartPointer<Self>                  Pointer;
    typedef SmartPointer<const Self>            ConstPointer;

    /** Method for creation through the object factory. */
    itkNewMacro( Self );

    /** Run-time type information (and related methods). */
    itkTypeMacro( GeneticAlgorithmOptimizer, ScaledSingleValuedNonLinearOptimizer );

    /** Codes of stopping conditions */
    typedef enum {
      MaximumNumberOfIterations,
        MetricError
    } StopConditionType;

    /** Advance one step following the gradient direction. */
    virtual void AdvanceOneStep( void );

    /** Start optimization. */
    void StartOptimization( void );

    /** Resume previously stopped optimization with current parameters
    * \sa StopOptimization. */
    void ResumeOptimization( void );

    /** Stop optimization.
    * \sa ResumeOptimization */
    void StopOptimization( void );

    /** Set the number of iterations. */
    itkSetMacro( NumberOfIterations, unsigned long );

    /** Get the current iteration number. */
    itkGetConstMacro( CurrentIteration, unsigned long );

    /** Get the current resolution number. */
    itkSetMacro ( CurrentResolution, unsigned int );

    /** Get the current best fitness value. */
    itkGetConstMacro( Value, double );

    /** Get Stop condition. */
    itkGetConstMacro( StopCondition, StopConditionType );

    /** Set/Get the population size. */
    itkSetMacro( PopulationSize, unsigned int );

    /** Set/Get the ranges of the parameters. */
    itkSetMacro( ParameterRange, Array2D<double> );

    /** Set/Get the crossover probability. */
    itkSetMacro( CrossoverProbability, float );

    /** Set/Get the mutation probability. */
    itkSetMacro( MutationProbability, float );

    /** Get the number of parameters. */
    itkGetConstMacro( NumberOfParameters, unsigned int );

    /** Set/Get the tournament size. */
    itkSetMacro( TournamentSize, unsigned int );

    /** Set/Get the restarts number. */
    itkSetMacro( Restarts, unsigned int );

    /** Set/Get the restarts performed. */
    itkGetMacro( RestartsPerformed, unsigned int );

    /** Set/Get the restart threshold. */
    itkSetMacro( RestartThreshold, float );

  protected:

    /** The random number generator used to generate random values. */
    typedef itk::Statistics::ThreadSafeMersenneTwisterRandomVariateGenerator RandomGeneratorType;

    GeneticAlgorithmOptimizer();
    virtual ~GeneticAlgorithmOptimizer();

    /** PrintSelf method.*/
    void PrintSelf( std::ostream& os, Indent indent ) const;

  private:

    class Individual
    {
    public:
      ParametersType genome;
      double         fitness;
      bool           evaluated;

      Individual( unsigned int length)
      {
        this->genome = ParametersType(length);
        this->evaluated = false;
      }

      Individual* deep_copy( void )
      {
        Individual* copy = new Individual( this->genome.Size()  );

        copy->genome = this->genome;    // deep copy
        copy->fitness = this->fitness;
        copy->evaluated = this->evaluated;

        return copy;
      }

      void print()
      {
	for(unsigned int i=0; i < this->genome.Size(); i++)
	{
	  std::cout << this->genome[i] << " ";
	}
	std::cout << std::endl;
	std::cout << this->fitness << std::endl;
	if (this->evaluated)
	  std::cout << "Evaluated" << std::endl;
      }

    }; // end class Individual

    typedef std::vector<Individual*> PopulationType;

    GeneticAlgorithmOptimizer( const Self& );  // purposely not implemented
    void operator=( const Self& );                    // purposely not implemented

    /** Initialize the Population. */
    void InitPopulation( void );

    /** EvaluateIndividual. */
    void EvaluateIndividual( Individual*, bool mandatory = false ); 

    /** Perform Selection. */
    Individual* Selection( void );

    /** Crossover */
    void Crossover( Individual* const, Individual* const, Individual* &, Individual* & );

    /** Mutation */
    void Mutation ( Individual* );

    /** ClearPopulation */
    void ClearPopulation ( void );

    /** Private member variables.*/
    bool                          m_Stop;
    double                        m_Value;
    StopConditionType             m_StopCondition;
    unsigned long                 m_NumberOfIterations;
    unsigned long                 m_CurrentIteration;
    unsigned long                 m_CurrentResolution;

    unsigned int                  m_PopulationSize;
    Array2D<double>               m_ParameterRange;
    float                         m_CrossoverProbability;
    float                         m_MutationProbability;
    unsigned int                  m_NumberOfParameters;
    unsigned int                  m_TournamentSize;

    RandomGeneratorType::Pointer  m_RandomGenerator;
    PopulationType*               m_Population;
    Individual*                   m_BestIndividual;

    unsigned int                  m_Restarts;
    unsigned int                  m_RestartsPerformed;
    float                         m_RestartThreshold;

    PopulationType*               m_StoredPopulation;
    Individual*                   m_StoredBestIndividual;

  }; // end class GeneticAlgorithmOptimizer

} // end namespace itk


#endif // end #ifndef __itkGeneticAlgorithmOptimizer_h

