/*======================================================================

  This file is part of the elastix software.

  Copyright (c) University Medical Center Utrecht. All rights reserved.
  See src/CopyrightElastix.txt or http://elastix.isi.uu.nl/legal.php for
  details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE. See the above copyright notices for more information.

======================================================================*/

#ifndef __elxScatterSearch_hxx
#define __elxScatterSearch_hxx

#include "elxScatterSearch.h"
#include <iomanip>
#include <string>
#include "math.h"

namespace elastix
{
  using namespace itk;


  /**
   * ********************* Constructor ****************************
   */

  template <class TElastix>
    ScatterSearch<TElastix>
    ::ScatterSearch()
  {
    ;
  } // end Constructor


  /**
   * ***************** BeforeRegistration ***********************
   */

  template <class TElastix>
    void ScatterSearch<TElastix>::
    BeforeRegistration(void)
  {
    /** Add some target cells to xout["iteration"].*/
    xout["iteration"].AddTargetCell("2:Metric");

    /** Format them as floats */
    xl::xout["iteration"]["2:Metric"]   << std::showpoint << std::fixed;

  } // end BeforeRegistration


  /**
   * ***************** BeforeEachResolution ***********************
   */

  template <class TElastix>
    void ScatterSearch<TElastix>
    ::BeforeEachResolution(void)
  {
    /** Get the current resolution level.*/
    unsigned int level = static_cast<unsigned int>(
      this->m_Registration->GetAsITKBaseType()->GetCurrentLevel() );

    this->SetCurrentResolution( level );

    /** Set the maximumNumberOfIterations.*/
    unsigned int maximumNumberOfIterations = 10;
    this->m_Configuration->ReadParameter( maximumNumberOfIterations,
      "MaximumNumberOfIterations", this->GetComponentLabel(), level, 0 );
    this->SetNumberOfIterations( maximumNumberOfIterations );

    /** Set the range of the parameters. */
    unsigned int numberOfParam = 
      this->m_Configuration->CountNumberOfParameterEntries( "ParameterRange" ) / 2;
    Array2D<double> ranges( numberOfParam, 2 );
    float minVal, maxVal;
    for ( unsigned int i = 0; i < numberOfParam; i++ )
    {
      this->m_Configuration->ReadParameter( minVal,
        "ParameterRange", 2*i, false );
      this->m_Configuration->ReadParameter( maxVal,
        "ParameterRange", 2*i + 1, false );
      ranges[ i ][ 0 ] = minVal;
      ranges[ i ][ 1 ] = maxVal;
    }
    this->SetParameterRange(ranges);

    /** Set the size of SS_P. */
    unsigned int SS_PSize = 10;
    this->m_Configuration->ReadParameter( SS_PSize,
      "SS_PSize", this->GetComponentLabel(), level, 0 );
    this->SetSS_PSize( SS_PSize );    

    /** Set SS_b. */
    unsigned int SS_b = 5;
    this->m_Configuration->ReadParameter( SS_b,
      "SS_b", this->GetComponentLabel(), level, 0 );
    this->SetSS_b1( SS_b/2 );    
    this->SetSS_b2( SS_b/2 );    

    /** Set SS_BLX_alpha. */
    double SS_BLX_alpha = 0.3;
    this->m_Configuration->ReadParameter( SS_BLX_alpha,
      "SS_BLX_alpha", this->GetComponentLabel(), level, 0 );
    this->SetSS_BLX_alpha( SS_BLX_alpha );    

    /** Set SS_LS_iterations. */
    unsigned int SS_LS_iterations = 25;
    this->m_Configuration->ReadParameter( SS_LS_iterations,
      "SS_LS_iterations", this->GetComponentLabel(), level, 0 );
    this->SetSS_LS_iterations( SS_LS_iterations );    

    /** Set the restart status and threshold */
    unsigned int restarts = 0;
    float restartThreshold = 0.5;
    bool adaptThresholdForZoom = false;
    this->m_Configuration->ReadParameter( restarts,
      "Restarts", this->GetComponentLabel(), level, 0 );
    this->SetRestarts( restarts );        
    this->m_Configuration->ReadParameter( restartThreshold,
      "RestartThreshold", this->GetComponentLabel(), level, 0 );
    this->SetRestartThreshold( restartThreshold );        
    /* Should the restart threshold be changed to take into account the zoom factor? 
    ** E.G. with Mutual Information
    */
    this->m_Configuration->ReadParameter( adaptThresholdForZoom,
      "AdaptThresholdForZoom", this->GetComponentLabel(), level, 0 );
    this->SetAdaptThresholdForZoom( adaptThresholdForZoom );        

  } // end BeforeEachResolution


  /**
   * ***************** AfterEachIteration *************************
   */

  template <class TElastix>
    void ScatterSearch<TElastix>
    ::AfterEachIteration(void)
  {
    /** Print some information */

    xl::xout["iteration"]["2:Metric"]   << this->GetValue();

    /** Select new spatial samples for the computation of the metric
     * \todo You may also choose to select new samples after evaluation
     * of the metric value */
    if ( this->GetNewSamplesEveryIteration() )
    {
      this->SelectNewSamples();
    }

  } // end AfterEachIteration


  /**
   * ***************** AfterEachResolution *************************
   */

  template <class TElastix>
    void ScatterSearch<TElastix>
    ::AfterEachResolution(void)
  {

    /**
     * enum   StopConditionType {  MaximumNumberOfIterations, MetricError }
     */
    std::string stopcondition;

    switch( this->GetStopCondition() )
    {

    case MaximumNumberOfIterations :
      stopcondition = "Maximum number of iterations has been reached";
      break;

    case MetricError :
      stopcondition = "Error in metric";
      break;

    default:
      stopcondition = "Unknown";
      break;

    }
    /** Print the stopping condition */
    elxout << "Stopping condition: " << stopcondition << "." << std::endl;

  } // end AfterEachResolution

  /**
   * ******************* AfterRegistration ************************
   */

  template <class TElastix>
    void ScatterSearch<TElastix>
    ::AfterRegistration(void)
  {
    /** Print the number of restarts that have been performed **/
    unsigned int restartsPerformed = this->GetRestartsPerformed();
    elxout << std::endl
           << "Number of restarts performed "
           << restartsPerformed
           << std::endl;

    /** Print the best metric value */
    double bestValue;
    bestValue = this->GetValue();
    elxout
      << std::endl
      << "Final metric value  = "
      << bestValue
      << std::endl;

  } // end AfterRegistration


  /**
   * ******************* StartOptimization ***********************
   */

  template <class TElastix>
    void ScatterSearch<TElastix>
    ::StartOptimization(void)
  {

    /** Check if the entered scales are correct and != [ 1 1 1 ...] */

    this->SetUseScales(false);
    const ScalesType & scales = this->GetScales();
    if ( scales.GetSize() == this->GetInitialPosition().GetSize() )
    {
      ScalesType unit_scales( scales.GetSize() );
      unit_scales.Fill(1.0);
      if (scales != unit_scales)
      {
        /** only then: */
        this->SetUseScales(true);
      }
    }

    /** Check if the ranges specified in ParameterRange are valid and match
     *  the number of parameters of the actual transformation. 
     * Alternatively, a single range can be provided in order to be used for all parameters.
     */
    bool useDefaultValues = false;
    unsigned int numberOfParams =
      this->GetScaledCostFunction()->GetNumberOfParameters();
    unsigned int NumOfRangeEntries = 
      this->m_Configuration->CountNumberOfParameterEntries( "ParameterRange" );
    float minVal, maxVal;

    // Every parameter uses the same range
    if (NumOfRangeEntries == 2)
    {

      this->m_Configuration->ReadParameter( minVal,
                                            "ParameterRange", 0, false );
      this->m_Configuration->ReadParameter( maxVal,
                                            "ParameterRange", 1, false );
      if ( maxVal < minVal )
      {
        useDefaultValues = true;
        xl::xout["warning"]
          << "WARNING: Invalid range in ParameterRange in postion " << 0
          << std::endl;
      }
      else
      {
        Array2D<double> ranges( numberOfParams, 2 );
        for ( unsigned int i = 0; i < numberOfParams; i++ )
        {
          ranges[ i ][ 0 ] = minVal;
          ranges[ i ][ 1 ] = maxVal;
        }
        this->SetParameterRange( ranges );   
      }
    }
    else
      if ( 2 * numberOfParams != NumOfRangeEntries )
      {
        useDefaultValues = true;
        xl::xout["warning"]
          << "WARNING: Invalid ParameterRange! The number of entries"
          << " in the ParameterRange option should be twice"
          << " the number of parameters of the selected transformation."
          << std::endl;
      }
      else
      {
        for ( unsigned int i = 0; i < numberOfParams; i++ )
        {
          this->m_Configuration->ReadParameter( minVal,
                                                "ParameterRange", 2*i, false );
          this->m_Configuration->ReadParameter( maxVal,
                                                "ParameterRange", 2*i + 1, false );
          if ( maxVal < minVal )
          {
            useDefaultValues = true;
            xl::xout["warning"]
              << "WARNING: Invalid range in ParameterRange in postion " << i + 1
              << std::endl;
            break;
          }
        }
      }

    if ( useDefaultValues )
    {
      xl::xout["warning"] << " The default range [-1,1] is used for all parameters."
                          << std::endl;
      Array2D<double> ranges( numberOfParams, 2 );
      for ( unsigned int i = 0; i < numberOfParams; i++ )
      {
        ranges[ i ][ 0 ] = -1;
        ranges[ i ][ 1 ] = 1;
      }
      this->SetParameterRange( ranges );
    }

    this->Superclass1::StartOptimization();

  } //end StartOptimization


} // end namespace elastix

#endif // end #ifndef __elxScatterSearch_hxx

