/*======================================================================

  This file is part of the elastix software.

  Copyright (c) University Medical Center Utrecht. All rights reserved.
  See src/CopyrightElastix.txt or http://elastix.isi.uu.nl/legal.php for
  details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE. See the above copyright notices for more information.

======================================================================*/

#ifndef __itkScatterSearchOptimizer_cxx
#define __itkScatterSearchOptimizer_cxx

#include "itkScatterSearchOptimizer.h"
#include "itkCommand.h"
#include "itkEventObject.h"
#include "itkExceptionObject.h"
#include "itkRealTimeClock.h"
#include "itkTimeStamp.h"

#include <algorithm>

namespace itk
{


  /**
   * ************************* Constructor ************************
   */

  ScatterSearchOptimizer
    ::ScatterSearchOptimizer()
  {
    itkDebugMacro( "Constructor" );

    /** Default values **/
    this->m_NumberOfIterations = 50;
    this->m_CurrentIteration = 0;
    this->m_CurrentResolution = 0;
    this->m_Value = 0.0;
    this->m_StopCondition = MaximumNumberOfIterations;
    this->m_Restarts = 0;
    this->m_RestartsPerformed = 0;

    this->m_SS_PSize = 10;
    this->m_SS_b1 = 5;
    this->m_SS_BLX_alpha = 0.3;
    this->m_SS_LS_iterations = 10;

    RealTimeClock::Pointer realTimeClock = RealTimeClock::New();
    RealTimeClock::TimeStampType stamp = realTimeClock->GetTimeStamp();

    this->m_RandomGenerator = RandomGeneratorType::New();
    this->m_RandomGenerator->Initialize( static_cast<uint32_t>(stamp*1000) );

    this->m_BestSolution = NULL;
    this->m_SS_P = NULL;
    this->m_SS_Subsets = new SubsetType();
    this->m_SS_Pool = new SolutionSetType();

    this->m_SS_StoredBestSolution = NULL;
    this->m_SS_StoredQualityRefSet = NULL;
    this->m_SS_StoredDiversityRefSet = NULL;

  } // end Constructor


  /**
   * ************************* Destructor ************************
   */

  ScatterSearchOptimizer
    ::~ScatterSearchOptimizer()
  {
    itkDebugMacro( "Destructor" );

    this->m_SS_P->clear();
    this->m_SS_QualityRefSet->clear();
    this->m_SS_DiversityRefSet->clear();
    this->m_SS_Pool->clear();
    this->m_SS_Subsets->clear();

    delete this->m_BestSolution;
    delete this->m_SS_P;
    delete this->m_SS_QualityRefSet;
    delete this->m_SS_DiversityRefSet;
    delete this->m_SS_Pool;
    delete this->m_SS_Subsets;
  } // end Destructor


  /**
   * ************************* PrintSelf **************************
   */

  void
    ScatterSearchOptimizer
    ::PrintSelf( std::ostream& os, Indent indent ) const
  {
    Superclass::PrintSelf( os, indent );

    os << indent << "NumberOfIterations: "
      << this->m_NumberOfIterations << std::endl;
    os << indent << "CurrentIteration: "
      << this->m_CurrentIteration;
    os << indent << "Value: "
      << this->m_Value;
    os << indent << "StopCondition: "
      << this->m_StopCondition;
    os << std::endl;

  } // end PrintSelf


  /**
   * *********************** StartOptimization ********************
   */
  void
    ScatterSearchOptimizer
    ::StartOptimization(void)
  {
    itkDebugMacro( "StartOptimization" );

    this->m_CurrentIteration = 0;
    this->m_Stop = false;

    /** Get the number of parameters; checks also if a cost function has been set at all.
     * if not: an exception is thrown */
    this->m_NumberOfParameters = this->GetScaledCostFunction()->GetNumberOfParameters();

    /** Initialize the scaledCostFunction with the currently set scales */
    this->InitializeScales();

    /** Set the current position as the scaled initial position */
    this->SetCurrentPosition( this->GetInitialPosition() );

    this->m_SS_FrequencyCount.SetSize(this->m_NumberOfParameters, 4);
    for(unsigned int i = 0; i < this->m_NumberOfParameters; i++)
    {
      this->m_SS_FrequencyCount[i][0] = 1;
      this->m_SS_FrequencyCount[i][1] = 1;
      this->m_SS_FrequencyCount[i][2] = 1;
      this->m_SS_FrequencyCount[i][3] = 1;
    }

    /** There are three cases: 
     * 1. If the first resolution (possibly after a restart), then the population is empty
     *    and we want to create a new population. 
     * 2. Else, for all transforms but B-Spline we want to keep the current population,
     *    we just need to evaluate the fitness again.  
     * 3. Finally For B-Spline we need a fresh start anyway.
     */

    if ( this->m_SS_P != NULL )
    {
      if ( this->m_NumberOfParameters > 12) // B-Splines
      {
        this->m_BestSolution = NULL;  //should be delete
        this->Clear( this->m_SS_P );
        this->Clear( this->m_SS_QualityRefSet );
        this->Clear( this->m_SS_DiversityRefSet );
        this->InitPAndRefSet();
      }
      else
      {
        this->m_BestSolution = NULL; // should be a delete
        for ( unsigned int i = 0; i < this->m_SS_QualityRefSet->size(); i++ )
          this->EvaluateSolution( this->m_SS_QualityRefSet->at(i), true );
        this->SortForQuality( this->m_SS_QualityRefSet );
        this->m_SS_median_fitness = this->m_SS_QualityRefSet->at( this->m_SS_b1/2 )->fitness;
        this->SortForDiversity( this->m_SS_DiversityRefSet );

        // Dynamic boundaries
        // During the second resolution transformation parameters range are restricted around the 
        // best solutions
        double b, l, h;
        for( unsigned int i = 0; i < this->m_NumberOfParameters; i++)
        {
          b = this->m_BestSolution->genome[i];
          l = this->m_ParameterRange[i][0];
          h = this->m_ParameterRange[i][1]; 

          this->m_ParameterRange[i][0] = b -(b-l)/10;
          this->m_ParameterRange[i][1] = b +(h-b)/10;
        }
      }
    }
    else
    {
      this->m_SS_P = new SolutionSetType();
      this->InitPAndRefSet();
    }

    if ( !this->m_Stop )
    {
      this->ResumeOptimization();
    }

  } // end StartOptimization


  /**
   * ********************** ResumeOptimization ********************
   */

  void
    ScatterSearchOptimizer
    ::ResumeOptimization( void )
  {
    itkDebugMacro( "ResumeOptimization" );

    this->m_Stop = false;

    InvokeEvent( StartEvent() );
    while( ! this->m_Stop )
    {
      this->AdvanceOneStep();
      this->m_CurrentIteration++;

      if( this->m_CurrentIteration >= this->m_NumberOfIterations )
      {
        /** RESTART 
            At the end of the first resolution, check if the
            best individual's fitness is lower than the threshold
        **/
        float adjusted_threshold = this->m_RestartThreshold;
        float temp;

        if ( this->m_AdaptThresholdForZoom )
        {
          switch( this->m_NumberOfParameters )
          {
          case 7:
            temp = this->m_BestSolution->genome[ 6 ];
            adjusted_threshold = this->m_AdaptThresholdForZoom * 
              ( 0.6136684 - 0.1507666 * temp + 0.5370982 * temp * temp );
            break;
          default:  
            ;
          }
        }

        if( ( this->m_CurrentResolution == 0 ) && ( this->m_Restarts > 0 ) && 
            ( this->m_Value > adjusted_threshold ) ) 
        {
          /** Restart */
          this->m_CurrentIteration = 0;
          this->m_Restarts--;
          this->m_RestartsPerformed++;

          std::cout << "Restart" << std::endl;

          // If the current solutions are better than the best we stored, 
          // replaced the stored ones
          if( ( this->m_SS_StoredBestSolution == NULL) || 
              ( this->m_Value < this->m_SS_StoredBestSolution->fitness ) )
          {
            this->m_SS_StoredQualityRefSet = this->DeepCopy( this->m_SS_QualityRefSet );
            this->m_SS_StoredDiversityRefSet = this->DeepCopy( this->m_SS_DiversityRefSet );
            this->m_SS_StoredBestSolution = this->m_BestSolution->deep_copy();
          }

          this->Clear( this->m_SS_QualityRefSet );
          delete this->m_SS_QualityRefSet;
          this->Clear( this->m_SS_DiversityRefSet );
          delete this->m_SS_DiversityRefSet;
          this->Clear( this->m_SS_P );
          this->m_BestSolution = NULL;
          this->InitPAndRefSet();
        }
        else
        {
          //If we ran out of restarts, replace the stored solutions
          //unless we got better results in the last try
          if ( ( this->m_CurrentResolution == 0 ) && ( this->m_Restarts == 0 ) 
               && (this->m_RestartsPerformed > 0) && ( this->m_Value > this->m_SS_StoredBestSolution->fitness ))
          {
            std::cout << "Restore" << std::endl;
            this->m_SS_QualityRefSet = this->DeepCopy( this->m_SS_StoredQualityRefSet );            
            this->m_SS_DiversityRefSet = this->DeepCopy( this->m_SS_StoredDiversityRefSet );            
            this->m_BestSolution = this->m_SS_StoredBestSolution->deep_copy();
            this->SetScaledCurrentPosition( this->m_BestSolution->deep_copy()->genome );
            this->m_Value = this->m_BestSolution->fitness;
          }

          this->m_StopCondition = MaximumNumberOfIterations;
          StopOptimization();
          break;
        }
      }

    } // while !m_stop


  } // end ResumeOptimization


  /**
   * ********************** StopOptimization **********************
   */

  void
    ScatterSearchOptimizer
    ::StopOptimization( void )
  {
    itkDebugMacro( "StopOptimization" );

    this->m_Stop = true;
    InvokeEvent( EndEvent() );

  } // end StopOptimization


  /**
   * ********************** AdvanceOneStep ************************
   */

  void
    ScatterSearchOptimizer
    ::AdvanceOneStep( void )
  {
    itkDebugMacro( "AdvanceOneStep" );

    this->m_SS_NewElements = true;
    this->Clear( this->m_SS_Pool ); 
    
    // std::cout << "AdvanceOneStep" << std::endl;
    // this->print(this->m_SS_QualityRefSet, "REF_SET");

    Solution* x_prime;
    while ( this->m_SS_NewElements )
    {
      this->SubsetGenerationMethod();
      this->m_SS_NewElements = false;
      
      while ( ! this->m_SS_Subsets->empty() )
      {
        SolutionSetType* sols = this->m_SS_Subsets->back();
        this->m_SS_Subsets->pop_back();
         
        Solution* x = this->SolutionCombinationMethod( sols );
        sols->clear(); //NO LEAK
        delete sols;

        if ( ( x->fitness < this->m_SS_median_fitness ) )
        {
          //          std::cout << x->fitness;
          x_prime = this->ImprovementMethod( x );
          delete x;
          x = x_prime;
          //          std::cout << x->fitness;
          this->ComputeDiversity(x, this->m_SS_QualityRefSet);
          if ( ( x->diversity > 0.05 * this->m_NumberOfParameters ) || 
               ( x->fitness < 1.01 * this->m_SS_QualityRefSet->at(0)->fitness ) )
          {
            //            std::cout << "YES" << std::endl;
            this->m_SS_Pool->push_back( x );
          }
        }
        else
        {
          this->ComputeDiversity(x, this->m_SS_QualityRefSet);
          if ( x->diversity > 0.01 * this->m_NumberOfParameters )
            this->m_SS_Pool->push_back( x );
        }
      }

      //      std::cout << "End of AdvanceOneStep" << std::endl;
      this->print(this->m_SS_QualityRefSet, "QUALITY_SET"); 
      this->print(this->m_SS_DiversityRefSet, "DIVERSITY_SET"); 
      this->print(this->m_SS_Pool, "POOL");
      //      std::cout << "Median: " << this->m_SS_median_fitness << std::endl;

      // "Elistsm"
      if ( ( this->m_BestSolution->fitness < this->m_SS_QualityRefSet->at(0)->fitness ) &&
           ( this->m_BestSolution->distance( this->m_SS_QualityRefSet->at(0) ) > 0.1 ) )
      {
        //        std::cout << "ELITISM" << std::endl;
        this->print(m_BestSolution);
        //        std::cout << "DIST: " << this->m_BestSolution->distance( this->m_SS_QualityRefSet->at(0) ) << std::endl;
        this->m_SS_Pool->push_back( this->m_BestSolution->deep_copy() );
      }

      //      std::cout << "ReferenceSetUpdateMethod" << std::endl;      
      this->ReferenceSetUpdateMethod();  // May set NewElements to true
      //      std::cout << "After RefenceSetUpdateMethod" << std::endl;

      this->print(this->m_SS_QualityRefSet, "QUALITY_SET"); 
      this->print(this->m_SS_DiversityRefSet, "DIVERSITY_SET"); 
    }

    //    std::cout << "CreateNewRefSet" << std::endl;

    this->CreateNewRefSet();

    //    std::cout << "DONE CreateNewRefSet" << std::endl;

    this->SetScaledCurrentPosition( this->m_BestSolution->deep_copy()->genome );
    // std::cout << this->m_BestSolution->genome;
    this->InvokeEvent( IterationEvent() );

  } // end AdvanceOneStep


  /**
   * ********************** EvaluateSolution ************************
   */

  void
    ScatterSearchOptimizer
  ::EvaluateSolution( Solution* indiv , bool mandatory )
  {
    itkDebugMacro( "EvaluateSolution" );  

    bool bad = false;
    double a,b;

    if ( !indiv->evaluated || mandatory)
    {
      do 
      {
        bad = false;
        try
        {
          indiv->fitness = this->GetScaledValue( indiv->genome );
        }
        catch( ExceptionObject& err )
        {
          bad = true;
        }

        if (bad)
        {
          //Replace the solution with a random one
          for ( unsigned int j = 0; j < this->m_NumberOfParameters; j++ )
          {
            a = this->m_ParameterRange[j][0];
            b = this->m_ParameterRange[j][1];
            indiv->genome[j] = this->m_RandomGenerator->GetUniformVariate(a, b);
          }
        }  
      }
      while(bad);
      
      indiv->evaluated = true;
      
      if ( ( this->m_BestSolution == NULL ) || 
           ( indiv->fitness < this->m_BestSolution->fitness ) )
      {
    	delete this->m_BestSolution;
        this->m_BestSolution = indiv->deep_copy();
        this->m_Value = indiv->fitness;
      }
    }
  } // End EvaluateSolution


  void
    ScatterSearchOptimizer
    ::InitPAndRefSet( void )
  {
    itkDebugMacro( "InitPAndRefSet" );

    // Fills P
    this->FillP();

    // Call the Improvement Method over the elements of P, using P as pool for the crossover
    // The elements of P are replaced with the improved version
    Solution* improv_sol;
    for ( unsigned int i=0; i < this->m_SS_PSize; i++)
    {
      improv_sol = this->ImprovementMethod( this->m_SS_P->at( i ), this->m_SS_P );
      delete this->m_SS_P->at(i);
      this->m_SS_P->at(i) = improv_sol;
    }
    this->SortForQuality( this->m_SS_P );
    
    this->m_SS_QualityRefSet = new SolutionSetType();
    this->m_SS_DiversityRefSet = new SolutionSetType();

    //Move the SS_b1 best solutions to SS_QualityRefSet; recall P is sorted for quality
    for(unsigned int i = 0; i < this->m_SS_b1; i++)
    {
      this->m_SS_QualityRefSet->push_back( this->m_SS_P->at(i) );
      this->m_SS_P->at(i) = NULL;
    }
    
    //Sort remaing solutions in P according to diversity and
    //move the SS_b2 most diverse solution to SS_DiversityRefSet
    this->SortForDiversity( this->m_SS_P, this->m_SS_b1);
    for(unsigned int i = this->m_SS_b1; i < this->m_SS_b1 + this->m_SS_b2; i++)
    {
      this->m_SS_DiversityRefSet->push_back( this->m_SS_P->at(i) );
      this->m_SS_P->at(i) = NULL;
    }

    //Clear the rest of P
    this->Clear( this->m_SS_P );

    //Update the median fitness
    this->m_SS_median_fitness = this->m_SS_QualityRefSet->at( this->m_SS_b1/2 )->fitness;
    this->m_SS_P = new SolutionSetType();
    
  } // end InitPAndRefSet


  void
    ScatterSearchOptimizer
    ::CreateNewRefSet( void )
  {
    itkDebugMacro( "InitPAndRefSet" );

    this->SortForDiversity( this->m_SS_DiversityRefSet );

    this->FillP();
    // Call the Improvement Method over the elements of P, using P as pool for the crossover
    // The elements of P are replaced with the improved version
    Solution* improv_sol;
    for ( unsigned int i=0; i < this->m_SS_PSize; i++)
    {
      improv_sol = this->ImprovementMethod( this->m_SS_P->at( i ), this->m_SS_P );
      delete this->m_SS_P->at(i);
      this->m_SS_P->at(i) = improv_sol;
    }
    this->SortForQuality( this->m_SS_P );

    //Merge P and QualityRefSet
    SolutionSetType* new_set = new SolutionSetType();

    unsigned int i,j;
    i = 0;
    j = 0;

    while ( new_set->size() < this->m_SS_b1 )
    {
      //      std::cout << i << ", " << j << std::endl;
      if ( ( j >= m_SS_P->size() ) 
           || ( this->m_SS_QualityRefSet->at(i)->fitness <= this->m_SS_P->at(j)->fitness ) )
      {
        new_set->push_back( this->m_SS_QualityRefSet->at(i) );
        this->m_SS_QualityRefSet->at(i) = NULL;
        i++;
      }
      else
      {
        new_set->push_back( this->m_SS_P->at(j) );
        this->m_SS_P->at(j) = NULL;
        j++;
      }
    }

    this->Clear( this->m_SS_QualityRefSet );
    delete this->m_SS_QualityRefSet;
    this->m_SS_QualityRefSet = new_set;

    // Update the median fitness value
    this->m_SS_median_fitness = this->m_SS_QualityRefSet->at( this->m_SS_b1 / 2 )->fitness;

    this->SortForDiversity( this->m_SS_P, this->m_SS_b1 );

    //Merge between P and DiversityRefSet
    new_set = new SolutionSetType();

    i = 0;
    j = this->m_SS_b1;

    while ( new_set->size() < this->m_SS_b2 )
    {
      if ( ( j >= m_SS_P->size() ) 
           || ( this->m_SS_DiversityRefSet->at(i)->diversity > this->m_SS_P->at(j)->diversity ) ) 
      {
        new_set->push_back( this->m_SS_DiversityRefSet->at(i) );
        this->m_SS_DiversityRefSet->at(i) = NULL;
        //        std::cout << "I: " << i << std::endl;
        i++;
      }
      else
      {
        new_set->push_back( this->m_SS_P->at(j) );
        this->m_SS_P->at(j) = NULL;
        //        std::cout << "J: " << j << std::endl;
        j++;
      }
    }

    this->Clear( this->m_SS_DiversityRefSet );
    delete this->m_SS_DiversityRefSet;
    this->m_SS_DiversityRefSet = new_set;

    //Clear the rest of P
    this->Clear( this->m_SS_P );
    this->m_SS_P = new SolutionSetType();
    
  } // end CreateNewRefSet
    

  /**
   * ********************** FillP ************************
   */

  void
    ScatterSearchOptimizer
  ::FillP( )
  {
    itkDebugMacro( "FillP" );

    Solution* solution;
    for ( unsigned int i = 0; i < this->m_SS_PSize; i++ )
    {      
      solution = this->DiversificationGenerationMethod();
      this->m_SS_P->push_back( solution );
    }
    this->SortForQuality( this->m_SS_P );

  } // end FillP


  /**
   * ********************** ReferenceSetUpdateMethod ************************
   */

  void
    ScatterSearchOptimizer
    ::ReferenceSetUpdateMethod( void )
  {
    itkDebugMacro( "ReferenceSetUpdateMethod" );

    /* Perform partial merge of Pool and RefSet to obtain a ordered vector 
     * with the best SS_b individuals 
     */
    this->SortForQuality( this->m_SS_Pool );
    //    std::cout << "SORT POOL done" << std::endl;

    SolutionSetType* new_set = new SolutionSetType();

    unsigned int i,j;
    i = 0;
    j = 0;

    if (this->m_SS_Pool->size() > 0)
    {

      while ( new_set->size() < this->m_SS_b1 )
      {
        //        std::cout << i << ", " << j << std::endl;
        if ( ( j >= m_SS_Pool->size() ) 
             || ( this->m_SS_QualityRefSet->at(i)->fitness <= this->m_SS_Pool->at(j)->fitness ) )
        {
          new_set->push_back( this->m_SS_QualityRefSet->at(i) );
          this->m_SS_QualityRefSet->at(i) = NULL;
          i++;
        }
        else
        {
          new_set->push_back( this->m_SS_Pool->at(j) );
          this->m_SS_Pool->at(j) = NULL;
          j++;
          this->m_SS_NewElements = true;
          // std::cout << "NewElements is now true" << std::endl;
        }
      }

      //      std::cout << "MERGE DONE" << std::endl;

      this->Clear( this->m_SS_Pool );
      this->Clear( this->m_SS_QualityRefSet );
      delete this->m_SS_QualityRefSet;
      this->m_SS_QualityRefSet = new_set;

      // Update the median fitness value
      this->m_SS_median_fitness = this->m_SS_QualityRefSet->at( this->m_SS_b1 / 2 )->fitness;
    }

  } // end ReferenceSetUpdateMethod


  /**
   * ********************** SubsetGenerationMethod ************************
   */

  void
    ScatterSearchOptimizer
    ::SubsetGenerationMethod( void )
  {
    itkDebugMacro( "SubsetGenerationMethod" );

    //Quality only
    for ( unsigned int i = 0; i < this->m_SS_QualityRefSet->size(); i++ )
    {
      for ( unsigned int j = i+1; j < this->m_SS_QualityRefSet->size(); j++)
      {
        SolutionSetType* set = new SolutionSetType();
        set->push_back( this->m_SS_QualityRefSet->at(i) );
        set->push_back( this->m_SS_QualityRefSet->at(j) );

        this->m_SS_Subsets->push_back( set );
      }
    }

    //Diversity only
    for ( unsigned int i = 0; i < this->m_SS_DiversityRefSet->size(); i++ )
    {
      for ( unsigned int j = i+1; j < this->m_SS_DiversityRefSet->size(); j++)
      {
        SolutionSetType* set = new SolutionSetType();
        set->push_back( this->m_SS_DiversityRefSet->at(i) );
        set->push_back( this->m_SS_DiversityRefSet->at(j) );

        this->m_SS_Subsets->push_back( set );
      }
    }

    //Mix quality and diversity solutions
    for ( unsigned int i = 0; i < this->m_SS_DiversityRefSet->size(); i++ )
    {
      for ( unsigned int j = 0; j < this->m_SS_QualityRefSet->size(); j++)
      {
        SolutionSetType* set = new SolutionSetType();
        set->push_back( this->m_SS_DiversityRefSet->at(i) );
        set->push_back( this->m_SS_QualityRefSet->at(j) );

        this->m_SS_Subsets->push_back( set );
      }
    }

  } // end SubsetGenerationMethod


  /**
   * ********************** SortForQuality ************************
   */

  void
    ScatterSearchOptimizer
    ::SortForQuality( SolutionSetType* set )
  {
    itkDebugMacro( "SortForQuality" );

    std::sort( set->begin(), set->end(), Solution::hasBetterFitness );

  } // end SortForQuality

  /**
   * ********************** SortForDiversity ************************
   */

  void
    ScatterSearchOptimizer
  ::SortForDiversity( SolutionSetType* set, unsigned int start_pos )
  {
    itkDebugMacro( "SortForDiversity" );

    //Compute the diversity measure wrt the QualityRefSet
    //then sorts the set

    for( unsigned int i = start_pos; i < set->size(); i++ )
      this->ComputeDiversity( set->at(i), this->m_SS_QualityRefSet);

    std::sort( set->begin() +  start_pos, set->end(), Solution::hasHigherDiversity);

  } // end SortForDiversity

  /**
   * ********************** ComputeDiversity ************************
   */

  void
    ScatterSearchOptimizer
  ::ComputeDiversity( Solution* sol, SolutionSetType* set )
  {
    itkDebugMacro( "ComputeDiversity" );

    float dist;
    float lowest_diff = 10000000;
    for( unsigned int j = 0; j < set->size(); j++ )
    {
      //SSD between sol and set(j)
      Solution* other_sol = set->at(j);

      dist = sol->distance(other_sol);

      if ( dist < lowest_diff )
        lowest_diff = dist;
    }
    sol->diversity = lowest_diff;

  } // end ComputeDiversity


  /**
   * ********************** DiversificationGenerationMethod ************************
   */

  ScatterSearchOptimizer::Solution*
    ScatterSearchOptimizer
    ::DiversificationGenerationMethod( void )
  {
    itkDebugMacro( "DiversificationGenerationMethod" );

    unsigned int t1,t2,t3,t4,interval;
    double tot;
    double p, low, high, a, b;
    Solution* solution = new Solution( this->m_NumberOfParameters );
    for ( unsigned int j = 0; j < this->m_NumberOfParameters; j++ )
    {
      t1 = this->m_SS_FrequencyCount[j][0];
      t2 = this->m_SS_FrequencyCount[j][1];
      t3 = this->m_SS_FrequencyCount[j][2];
      t4 = this->m_SS_FrequencyCount[j][3];
      tot = t1 + t2 + t3 + t4;

      p = this->m_RandomGenerator->GetUniformVariate(0,1);

      if ( p < (1-t1/tot)/3 )
        interval = 0;
      else
      {
        if ( p < (2-(t1+t2)/tot)/3 )
          interval = 1;
        else
        {
          if ( p < (3-(t1+t2+t3)/tot)/3 )
            interval = 2;
          else
            interval = 3;
        }
      }
      
      this->m_SS_FrequencyCount[ j ][ interval ] += 1; 

      a = this->m_ParameterRange[j][0];
      b = this->m_ParameterRange[j][1];
      low  = a + interval     * (b-a)/4;
      high = a + (interval+1) * (b-a)/4;
      solution->genome[j] = this->m_RandomGenerator->GetUniformVariate(low, high);
    }
    
    this->EvaluateSolution( solution, true );
    return solution;

  } // end DiversificationGenerationMethod


  /**
   * ********************** ImprovementMethod ************************
   */

  ScatterSearchOptimizer::Solution*
    ScatterSearchOptimizer
  ::ImprovementMethod( Solution* solution )
  {
    itkDebugMacro( "ImprovementMethod" );

    return this->ImprovementMethod( solution, this->m_SS_QualityRefSet );

  } // end ImprovementMethod



  /**
   * ********************** ImprovementMethod ************************
   */

  ScatterSearchOptimizer::Solution*
    ScatterSearchOptimizer
  ::ImprovementMethod( Solution* solution, SolutionSetType* pool )
  {
    itkDebugMacro( "ImprovementMethod" );

    /* Crossover based local serch method. PBX-alpha is used to generate neighbours *
     * and the best one is selected.
     */

    //Ensures that bestSolution is actually a "new solution"
    // i.e. you can delete it, both here or outside

    Solution* bestSolution = solution->deep_copy();

    Solution* otherParent;
    Solution* offspring;
    unsigned int pos;

    for ( unsigned int i = 0; i < this->m_SS_LS_iterations; i++)
    {
      pos = this->m_RandomGenerator->GetIntegerVariate( pool->size() -1 );
      otherParent = pool->at( pos );

      offspring = PBX_Alpha_Crossover( bestSolution, otherParent );

      if ( offspring->fitness < bestSolution->fitness )
      {
        delete bestSolution;
        bestSolution = offspring;
        //        std::cout << "X";
      }
      else
      {
        delete offspring;
        //        std::cout << "_";
      }
    }

    //    std::cout << std::endl;

    return bestSolution;

  } // end ImprovementMethod


  /**
   * ********************** PBX_Alpha_Crossover ************************
   */

  ScatterSearchOptimizer::Solution* 
    ScatterSearchOptimizer
  ::PBX_Alpha_Crossover( Solution* parA, Solution* parB )
  {
    itkDebugMacro( "PBX_Alpha_Crossover" );

    const double PLX_alpha = 0.5;

    Solution* offA = new Solution( this->m_NumberOfParameters );
    Solution* offB = new Solution( this->m_NumberOfParameters );

    double x,y,I;
    for ( unsigned int i = 0; i < this->m_NumberOfParameters; i++ )
    {
      x = parA->genome[ i ];
      y = parB->genome[ i ];
      I = std::fabs(x - y);

      //Enforce the parameters to be in the appropriate range
      double u,l;

      l = std::max( x - I * PLX_alpha, this->m_ParameterRange[ i ][ 0 ] );
      u = std::min( x + I * PLX_alpha, this->m_ParameterRange[ i ][ 1 ] );
      offA->genome[ i ] = this->m_RandomGenerator->GetUniformVariate( l, u );

      l = std::max( y - I * PLX_alpha, this->m_ParameterRange[ i ][ 0 ] );
      u = std::min( y + I * PLX_alpha, this->m_ParameterRange[ i ][ 1 ] );
      offB->genome[ i ] = this->m_RandomGenerator->GetUniformVariate( l, u );
    }

    this->EvaluateSolution( offA, true );
    this->EvaluateSolution( offB, true );

    if ( offA->fitness < offB->fitness )
    {
      delete offB;
      return offA;
    }
    else
    {
      delete offA;
      return offB;
    }

  } // end PBX_Alpha_Crossover


  /**
   * ********************** SolutionCombinationMethod ************************
   */

  ScatterSearchOptimizer::Solution* 
    ScatterSearchOptimizer
  ::SolutionCombinationMethod( SolutionSetType* set )
  {
    itkDebugMacro( "SolutionCombinationMethod" );

    Solution* parA = set->at(0);
    Solution* parB = set->at(1);

    double cMax, cMin, I;

    Solution* offA = new Solution( this->m_NumberOfParameters );
    Solution* offB = new Solution( this->m_NumberOfParameters );

    for ( unsigned int i = 0; i < this->m_NumberOfParameters; i++ )
    {
      if (parA->genome[ i ] > parB->genome[ i ])
      {
        cMax = parA->genome[ i ];
        cMin = parB->genome[ i ];
      }
      else
      {
        cMin = parA->genome[ i ];
        cMax = parB->genome[ i ];
      }
      I = cMax - cMin;

      //Enforce the parameters to be in the appropriate range
      double u,l;

      l = std::max(cMin - I * this->m_SS_BLX_alpha, this->m_ParameterRange[ i ][ 0 ]);
      u = std::min(cMax + I * this->m_SS_BLX_alpha, this->m_ParameterRange[ i ][ 1 ]);

      offA->genome[ i ] = this->m_RandomGenerator->GetUniformVariate( l, u );
      offB->genome[ i ] = this->m_RandomGenerator->GetUniformVariate( l, u );
    }

    this->EvaluateSolution( offA, true );
    this->EvaluateSolution( offB, true );

    if ( offA->fitness < offB->fitness )
    {
      delete offB;
      return offA;
    }
    else
    {
      delete offA;
      return offB;
    }

  } // end SolutionCombinationMethod


  void ScatterSearchOptimizer::Clear(SolutionSetType* set)
  {
    for ( unsigned int i = 0; i < set->size(); i++)
      delete set->at(i);
    set->clear();
  }

  ScatterSearchOptimizer::SolutionSetType* ScatterSearchOptimizer::DeepCopy( SolutionSetType* set)
  {
    SolutionSetType* result = new SolutionSetType();

    Solution* sol = NULL;
    for(unsigned int i = 0; i < set->size(); i++)
    {
      sol  = set->at(i)->deep_copy();
      result->push_back(sol);
    }
    return result;
  }

  void ScatterSearchOptimizer::print(SolutionSetType* set, std::string name)
  {
    // std::cout << name << "************************" << std::endl;
    // for ( unsigned int i = 0; i < set->size(); i++)
    //   this->print(set->at(i));
    // std::cout << "************************" << std::endl;
    // // ***********************************************************
  }

  void ScatterSearchOptimizer::print(Solution* sol)
  {
      // std::cout << "[ ";
      // for ( unsigned int j = 0; j < this->m_NumberOfParameters; j++)
      // {
      //   std::cout << sol->genome[j] << " ";
      // }
      // std::cout << " *** " << sol->fitness << std::endl;
  }


} // end namespace itk

#endif // end #ifndef __itkScatterSearchOptimizer_cxx
