/*======================================================================

  This file is part of the elastix software.

  Copyright (c) University Medical Center Utrecht. All rights reserved.
  See src/CopyrightElastix.txt or http://elastix.isi.uu.nl/legal.php for
  details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE. See the above copyright notices for more information.

======================================================================*/

#ifndef __itkScatterSearchOptimizer_h
#define __itkScatterSearchOptimizer_h

#include "itkScaledSingleValuedNonLinearOptimizer.h"
#include "itkThreadSafeMersenneTwisterRandomVariateGenerator.h"
#include "itkArray2D.h"

namespace itk
{
  /**
   * \class ScatterSearchOptimizer
   * \brief An optimizer based on Scatter Search
   *
   * bla bla
   *
   * \ingroup Optimizers
   * \sa ScatterSearch
   */

  class ScatterSearchOptimizer
    : public ScaledSingleValuedNonLinearOptimizer
  {
  public:

    /** Standard class typedefs. */
    typedef ScatterSearchOptimizer    Self;
    typedef ScaledSingleValuedNonLinearOptimizer        Superclass;
    typedef SmartPointer<Self>                  Pointer;
    typedef SmartPointer<const Self>            ConstPointer;

    /** Method for creation through the object factory. */
    itkNewMacro( Self );

    /** Run-time type information (and related methods). */
    itkTypeMacro( ScatterSearchOptimizer, ScaledSingleValuedNonLinearOptimizer );

    /** Codes of stopping conditions */
    typedef enum {
      MaximumNumberOfIterations,
        MetricError
    } StopConditionType;

    /** Advance one step following the gradient direction. */
    virtual void AdvanceOneStep( void );

    /** Start optimization. */
    void StartOptimization( void );

    /** Resume previously stopped optimization with current parameters
    * \sa StopOptimization. */
    void ResumeOptimization( void );

    /** Stop optimization.
    * \sa ResumeOptimization */
    void StopOptimization( void );

    /** Set the number of iterations. */
    itkSetMacro( NumberOfIterations, unsigned long );

    /** Get the current iteration number. */
    itkGetConstMacro( CurrentIteration, unsigned int );

    /** Get the current resolution number. */
    itkSetMacro ( CurrentResolution, unsigned int );

    /** Get the current best fitness value. */
    itkGetConstMacro( Value, double );

    /** Get Stop condition. */
    itkGetConstMacro( StopCondition, StopConditionType );

    /** Set/Get the ranges of the parameters. */
    itkSetMacro( ParameterRange, Array2D<double> );

    /** Set the size of SS_P. */
    itkSetMacro( SS_PSize, unsigned int );

    /** Set SS_b1 and SS_b2. */
    itkSetMacro( SS_b1, unsigned int );
    itkSetMacro( SS_b2, unsigned int );

    /** Set alpha in BLX crossover. */
    itkSetMacro( SS_BLX_alpha, float );

    /** Set the number of iterations for the local search procedure. */
    itkSetMacro( SS_LS_iterations, unsigned int );

    /** Set/Get the restart threshold. */
    itkSetMacro( Restarts, unsigned int );

    /** Set/Get the restarts performed. */
    itkGetMacro( RestartsPerformed, unsigned int );

    /** Set/Get the restart threshold. */
    itkSetMacro( RestartThreshold, float );

    /** Restart threshold is adapted to the zoom factor
     ** e.g., with Mutual Information
     */
    itkSetMacro( AdaptThresholdForZoom, bool );

    class Solution
    {
    public:
      ParametersType genome;
      double         fitness;
      float          diversity;
      bool           evaluated;

      Solution( unsigned int length)
      {
        this->genome = ParametersType(length);
        this->fitness = 0;
        this->diversity = 0;
        this->evaluated = false;
      }

      Solution* deep_copy( void )
      {
        Solution* copy = new Solution( this->genome.Size() );

        copy->genome = this->genome;    // deep copy
        copy->fitness = this->fitness;
        copy->diversity = this->diversity;
        copy->evaluated = this->evaluated;

        return copy;
      }

      double distance( Solution* other)
      {
        double sum = 0;
        double temp;
        for(unsigned int i=0; i < this->genome.Size(); i++)
        {
          temp = this->genome[i] - other->genome[i];
          sum += temp * temp;
        }
        return sum;
      }

      bool static hasBetterFitness( Solution* a, Solution* b )
      {
        return (a->fitness < b->fitness);
      }

      bool static hasHigherDiversity( Solution* a, Solution* b )
      {
        return (a->diversity > b->diversity);
      }

    }; // end class Solution

    typedef std::vector<Solution*> SolutionSetType;
    typedef std::vector<SolutionSetType*> SubsetType;

  protected:

    /** The random number generator used to generate random values. */
    typedef itk::Statistics::ThreadSafeMersenneTwisterRandomVariateGenerator RandomGeneratorType;

    ScatterSearchOptimizer();
    virtual ~ScatterSearchOptimizer();

    /** PrintSelf method.*/
    void PrintSelf( std::ostream& os, Indent indent ) const;

  private:

    ScatterSearchOptimizer( const Self& );  // purposely not implemented
    void operator=( const Self& );                    // purposely not implemented

    /** Fill P with solution using the DiversificationGeneration method. 
     */
    void FillP( void );

    /** Generate a diverse solution. */
    Solution* DiversificationGenerationMethod( void );

    /** Improve a solution.  */
    Solution* ImprovementMethod( Solution* );
    Solution* ImprovementMethod( Solution* , SolutionSetType* );

    /** Init P and RefSet **/
    void InitPAndRefSet( void );

    /** Create a new P and build a new RefSet **/
    void CreateNewRefSet( void );

    /** EvaluateSolution. */
    void EvaluateSolution( Solution*, bool mandatory = false ); 

    /** Solution combination method. */
    Solution* SolutionCombinationMethod( SolutionSetType* );

    /** PLX_alhpa Crossover. */
    Solution* PBX_Alpha_Crossover( Solution*, Solution* );

    /** Sort a set of solutions according to their fitness value. */
    void SortForQuality( SolutionSetType* );

    /** Compute the diversity value and sort WRT it. */
    void SortForDiversity( SolutionSetType*, unsigned int start = 0 );

    /** Compute the diversity WRT the QualityRefSet. */
    void ComputeDiversity( Solution*, SolutionSetType* );

    /** Subset generation method. */
    void SubsetGenerationMethod( void );

    /** Reference set update method. */
    void ReferenceSetUpdateMethod( void );

    /** Clear a solution set. */
    void Clear( SolutionSetType* );

    /** Deep copy a solution set. */
    SolutionSetType* DeepCopy( SolutionSetType* );

    /** Private member variables.*/
    bool                          m_Stop;
    double                        m_Value;
    StopConditionType             m_StopCondition;
    unsigned long                 m_NumberOfIterations;
    unsigned long                 m_CurrentIteration;
    unsigned long                 m_CurrentResolution;
    Array2D<double>               m_ParameterRange;
    RandomGeneratorType::Pointer  m_RandomGenerator;
    unsigned int                  m_NumberOfParameters;
    Solution*                     m_BestSolution;

    SolutionSetType*              m_SS_P;
    unsigned int                  m_SS_PSize;
    SolutionSetType*              m_SS_QualityRefSet;
    SolutionSetType*              m_SS_DiversityRefSet;
    unsigned int                  m_SS_b1;
    unsigned int                  m_SS_b2;
    SolutionSetType*              m_SS_Pool;
    bool                          m_SS_NewElements;
    SubsetType*                   m_SS_Subsets;
    float                         m_SS_median_fitness;
    float                         m_SS_BLX_alpha;
    unsigned int                  m_SS_LS_iterations;

    unsigned int                  m_Restarts;
    unsigned int                  m_RestartsPerformed;
    float                         m_RestartThreshold;
    bool                          m_AdaptThresholdForZoom;
    
    SolutionSetType*              m_SS_StoredQualityRefSet;
    SolutionSetType*              m_SS_StoredDiversityRefSet;
    Solution*                     m_SS_StoredBestSolution;

    Array2D<double>               m_SS_FrequencyCount;

    void print(SolutionSetType* set, std::string name);  //**************************** TEMP
    void print(Solution* );  //**************************** TEMP


  }; // end class ScatterSearchOptimizer

} // end namespace itk


#endif // end #ifndef __itkScatterSearchOptimizer_h

