/*  
   Compute the measure of overlap between two images
   Images are thresholded, then the intersection is computed and measured
   If the images are different in size or spacing, the smallest is upsampled to match the other.
*/


#include <stdlib.h>

#include <itkImage.h>
#include <itkImageFileReader.h>
#include <itkImageFileWriter.h>
#include "itkResampleImageFilter.h"
#include "itkNearestNeighborInterpolateImageFunction.h"

using namespace std;

typedef short PixelType;
typedef itk::Image< PixelType, 2 >  ImageType;
typedef itk::ImageFileReader< ImageType > ReaderType;

const int threshold = 0;

int main( int argc, char ** argv )
{

  string fixedImageFilename;
  string movingImageFilename;

  fixedImageFilename = argv[1];  
  movingImageFilename = argv[2];

  ImageType::Pointer fixedImage;
  ImageType::Pointer movingImage;

  // Read the fixed image
  ReaderType::Pointer reader = ReaderType::New();
  reader->SetFileName( fixedImageFilename );
  try
  {
    reader->Update();
  }
  catch( itk::ExceptionObject & err )
  {
    cerr << err << endl;
    return EXIT_FAILURE;
  }
  fixedImage = reader->GetOutput();

  // Read the moving image
  ReaderType::Pointer reader2 = ReaderType::New();
  reader2->SetFileName( movingImageFilename );
  try
  {
    reader2->Update();
  }
  catch( itk::ExceptionObject & err )
  {
    cerr << err << endl;
    return EXIT_FAILURE;
  }
  movingImage = reader2->GetOutput();

  //Determine the common region between the images
  ImageType::RegionType commonRegion = fixedImage->GetRequestedRegion();
  commonRegion.Crop( movingImage->GetRequestedRegion() );

  itk::ImageRegionConstIterator< ImageType > fixItr( fixedImage, commonRegion );
  itk::ImageRegionConstIterator< ImageType > movItr( movingImage, commonRegion );

  unsigned long total_count = 0;
  unsigned long common_count = 0;

  PixelType a, b;
  for( fixItr.GoToBegin(), movItr.GoToBegin(); !fixItr.IsAtEnd(); ++fixItr, ++movItr )
  {
    a = fixItr.Get();
    b = movItr.Get();

    if ( a > threshold ) 
      total_count++;
    if ( b > threshold )
      total_count++;
    if ( ( a > threshold ) && ( b > threshold ) )
      common_count++;

    if (!( ( ( a == 0 ) || ( a == 1 ) || ( a == 255 ) ) && ( ( b == 0 ) || ( b == 1 ) || ( b == 255 ) ) ))
    {
      cerr << "Warning: "
           << "Pixel is not 0, 1 nor 255" << endl;
      cerr << a << " " << b << endl;
    }
  }

  // Compute the overlap
  cout << 2*common_count/(double)total_count << endl;

  return EXIT_SUCCESS;
}

