/* Extract slices on a given axis
   Change the intensity to fit 0-255
*/

#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkExtractImageFilter.h"
#include "itkImage.h"
#include <itkRescaleIntensityImageFilter.h>

#include <string>

int main( int argc, char ** argv )
{
  // Verify the number of parameters in the command line
  if( argc < 4 )
    {
    std::cerr << "Usage: " << std::endl;
    std::cerr << argv[0] << " input3DImageFile  dimension slice_number" << std::endl;
    return EXIT_FAILURE;
    }

  typedef float         InputPixelType;
  typedef unsigned char OutputPixelType;

  typedef itk::Image< InputPixelType,  3 >    InputImageType;
  typedef itk::Image< InputPixelType,  2 >    SliceImageType;
  typedef itk::Image< OutputPixelType, 2 >    OutputImageType;

  typedef itk::ImageFileReader< InputImageType  >  ReaderType;
  typedef itk::ImageFileWriter< OutputImageType >  WriterType;

  typedef itk::ExtractImageFilter< InputImageType, SliceImageType > ExtractFilterType;
  typedef itk::RescaleIntensityImageFilter< SliceImageType, OutputImageType > RescaleFilterType;

  // Command line arguments
  std::string inputFilename(argv[1]);
  unsigned int dimension = atoi( argv[2] );
  unsigned int sliceNumber = atoi( argv[3] );

  ReaderType::Pointer reader = ReaderType::New();
  reader->SetFileName( inputFilename );
  reader->Update();

  InputImageType::RegionType inputRegion =
           reader->GetOutput()->GetLargestPossibleRegion();
  InputImageType::SizeType size;
  InputImageType::IndexType start;
  InputImageType::RegionType desiredRegion;
  ExtractFilterType::Pointer extractFilter;

  size = inputRegion.GetSize();
  size[dimension] = 0;

  start = inputRegion.GetIndex();
  start[dimension] = sliceNumber;
  
  desiredRegion.SetSize(  size  );
  desiredRegion.SetIndex( start );

  extractFilter = ExtractFilterType::New();
  extractFilter->SetExtractionRegion( desiredRegion );
  extractFilter->SetInput( reader->GetOutput() );
  extractFilter->Update();

  RescaleFilterType::Pointer rescaleFilter = RescaleFilterType::New();
  rescaleFilter->SetInput( extractFilter->GetOutput() );
  rescaleFilter->SetOutputMinimum( 0 );
  rescaleFilter->SetOutputMaximum( 255 );
  rescaleFilter->Update();

  WriterType::Pointer writer;

  std::ostringstream outputFilename;
  outputFilename << inputFilename.substr(0, inputFilename.length()-4) << "_" << dimension << "_" << sliceNumber;
  std::string outFileName( outputFilename.str() );
  
  // write Metaimage
  writer = WriterType::New();
  writer->SetFileName( outFileName + ".mhd" );
  writer->SetUseCompression(true);
  writer->SetInput( rescaleFilter->GetOutput() );
  try
  {
    writer->Update();
  }
  catch( itk::ExceptionObject & err )
  {
      std::cerr << "ExceptionObject caught !" << std::endl;
      std::cerr << err << std::endl;
      return EXIT_FAILURE;
  }
  
  return EXIT_SUCCESS;
}
