#!/usr/bin/python
# Run ofeli, compute a validation measure and write a table with the results.

from subprocess import Popen, PIPE
import os
from os.path import join
from time import time, sleep
import itertools
import multiprocessing
from collections import namedtuple
from optparse import OptionParser
import re
import random
import glob
import shutil

default_config = {}

# Read the configuration file
try:
    from config import *
except Exception, e:
    print "Configuration file incorrect or missing"
    print e
    exit()

Task = namedtuple('Task', 'img output_dir extconf options' )

class Writer(multiprocessing.Process):
    def __init__(self, results_queue, file_header):
        multiprocessing.Process.__init__(self)
        self.queue = results_queue
        self.file_header = file_header

    def run(self):
        results_file = open(results_file_name, 'w')
        results_file.write( self.file_header + "\n" )
        while True:
            try:
                res = self.queue.get()
            except KeyboardInterrupt, e:
                break
            if res is None:
                #Poison pill
                results_file.close()
                break
            results_file.write(res)
            results_file.flush()

class Executor(multiprocessing.Process):

    def __init__(self, task_queue, result_queue):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue

    def run(self):
        while True:
            task = self.task_queue.get()
            if task is None:
                # Poison pill means shutdown
                self.task_queue.task_done()
                break
            # Wait for a small random time, because of Elastix's poor random generator
            sleep(random.random())

            # Run elastix
            starttime = time()

            args = [ ofeli_cmd, join(input_data_path, task.img) + input_data_ext,
                     join(input_data_path, task.img) + prior_data_ext, "-out_dir", task.output_dir ]

            # Write the parameters file
            for key in sorted(task.extconf.iterkeys()):
                args += [ key, str(task.extconf[key]) ]

            proc = Popen(args)
            try:
                out, err = proc.communicate()
            except KeyboardInterrupt, e:
                break
            except Exception, e:
                print e
                self.task_queue.task_done()
                continue

            if err != '' and err is not None:
                print 'Error while running for', task.img
                print err
                self.task_queue.task_done()
                continue

            #Collect result
            timing = time() - starttime

            if not task.options.compute_vfc_only:
                #The output is the "last" out file
                last_out_file = join(task.output_dir,"out_last")

                args = [ validation_program, join(validation_data_path, task.img) + validation_data_ext,
                         last_out_file + output_data_ext ]

                proc = Popen(args, stdout=PIPE, stderr=PIPE)
                try:
                    out, err = proc.communicate()
                except Exception, e:
                    print 'An exception was raised when calling the validation program'
                    print e
                    self.task_queue.task_done()
                    continue

                if err != '' and err is not None:
                    print 'Error while performing validation for', task.img
                    print err
                    self.task_queue.task_done()
                    continue

                validation_measure = float(out)


                # Prepare the output for the results file
                # Print the parameters that vary in the configurations

                temp_out = ""
                if task.extconf:
                    temp_out += ", ".join( [ str(task.extconf[key])
                                         for key in sorted(task.extconf.iterkeys()) ] )

                output_line = "{0}, {1}, {2}, {3}\n".format( task.img, temp_out, validation_measure, timing )
                print validation_label+":", validation_measure, "Time:", timing, \
                      "Remaining:", self.task_queue.qsize()-1

                # Remove the quotes symbols
                self.result_queue.put( output_line )

            self.task_queue.task_done()

if __name__ == '__main__':

    parser = OptionParser()
    options = None
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose", default=False)
    parser.add_option("-o", "--compute_vfc_only",
                      action="store_true", dest="compute_vfc_only", default=False)
    (options, args) = parser.parse_args()

    if not os.path.exists(ofeli_cmd):
        print "ERROR elastix not found. It is supposed to be in " + ofeli_cmd
        exit()

    if not os.path.exists(input_data_path):
        print "ERROR Images not found, are you sure you are in the right directory?"
        exit()

    if not os.path.exists(validation_program):
        print "WARNING Validation program not found. It is supposed to be " + validation_program
        exit()

    # Purge previous results
    for directory in glob.glob("results_*"):
        shutil.rmtree(directory)

    tasks = multiprocessing.JoinableQueue()
    results = multiprocessing.Queue()

    # Create a list of dictionaries with all combination of values from variable_parameters
    extra_conf = [ dict( zip( variable_parameters.keys(), x ) )
                   for x in itertools.product( *variable_parameters.values() ) ]

    # Start executors
    executors = [ Executor(tasks, results) for i in xrange(num_threads) ]
    for w in executors:
        w.start()

    config = default_config
    for config_num, ext in enumerate(extra_conf):
        config.update(ext)
        #Check the parameters values for inconsistency
        for img in scenarios:
            # Create the output directory
            output_dir = "results_{0:06d}/{1}".format(config_num, img)

            if not os.path.exists(output_dir):
                os.makedirs(output_dir)

            #Run the test by calling Elastix
            tasks.put(Task( img, output_dir, extra_conf[ config_num ], options))

    # Add a poison pill for each executors
    for i in xrange( num_threads ):
        tasks.put(None)

    # Create results file header
    temp_out = ""
    if extra_conf[0]:
        temp_out += ", ".join(sorted(extra_conf[0].iterkeys()))

    header = "img, {0}, {1}, time".format(temp_out, validation_label)
    header = header.replace("-",'')

    # Process that writes the output as soon as it is produced
    writer = Writer(results, header)
    writer.start()

    try:
        # Wait for all of the tasks to finish
        tasks.join()
        # Poison pill for the writer
        results.put(None)
        # Wait for the writer to finish
        writer.join()
    except KeyboardInterrupt, e:
        print "Interrupted by user"
        os._exit(1)
