#!/usr/bin/python
# Run elastix (2 phases), ofeli and validation

from subprocess import Popen, PIPE
import os
from os.path import join
from time import time, sleep
import itertools
import multiprocessing
from collections import namedtuple
from optparse import OptionParser
import re
import random
import glob
import shutil
import tempfile

# Read the configuration file
try:
    from config import *
except Exception, e:
    print "Configuration file incorrect or missing"
    print e
    exit()

Task = namedtuple('Task', 'fixed moving output_dir extconf_Ofeli rep' )

class Writer(multiprocessing.Process):
    def __init__(self, results_queue, file_header):
        multiprocessing.Process.__init__(self)
        self.queue = results_queue
        self.file_header = file_header

    def run(self):
        results_file = open(results_file_name, 'w')
        results_file.write( self.file_header + "\n" )
        while True:
            try:
                res = self.queue.get()
            except KeyboardInterrupt, e:
                break
            if res is None:
                #Poison pill
                results_file.close()
                break
            results_file.write(res)
            results_file.flush()

class Executor(multiprocessing.Process):

    def __init__(self, task_queue, result_queue):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue

    def run(self):
        while True:
            task = self.task_queue.get()
            if task is None:
                # Poison pill means shutdown
                self.task_queue.task_done()
                break

            output_dir = tempfile.mkdtemp( dir=os.getcwd() )

            # Wait for a small random time, because of Elastix's poor random generator
            sleep(random.random())

            # Select a random prior
            prior_img = "{0}_{1}".format(task.fixed, task.moving)
            prior_img_file = glob.glob(join(prior_path, prior_img + "*"))[task.rep]

            # Get the prior validation value
            args = [ validation_program, join(validation_data_path, task.fixed) + validation_data_ext,
                     prior_img_file]

            proc = Popen(args, stdout=PIPE, stderr=PIPE)
            try:
                out, err = proc.communicate()
            except Exception, e:
                print e
                self.task_queue.task_done()
                continue

            if err != '' and err is not None:
                print 'Error while performing validation for', task.fixed, 'and', task.moving
                print err
                self.task_queue.task_done()
                continue

            last_line = out.rstrip("\n").rpartition("\n")[2]
            validation_prior = float(last_line)

            #Run Ofeli #######
            start_time_ofeli = time()

            args = [ ofeli_cmd, join(input_data_path, task.fixed) + ofeli_input_data_ext,
                     prior_img_file, "-out_dir", join(output_dir, 'ofeli') ]

            for key in sorted(task.extconf_Ofeli.iterkeys()):
                args += [ key, str(task.extconf_Ofeli[key]) ]

            proc = Popen(args)
            try:
                out, err = proc.communicate()
            except KeyboardInterrupt, e:
                break
            except Exception, e:
                print e
                self.task_queue.task_done()
                continue

            if err != '' and err is not None:
                print 'Error while running for', task.fixed
                print err
                self.task_queue.task_done()
                continue

            #Collect result
            timing = time() - start_time_ofeli

            #The output is the "last" out file
            last_out_file = join(output_dir, 'ofeli', "out_last")

            args = [ validation_program, join(validation_data_path, task.fixed) + ofeli_validation_data_ext,
                     last_out_file + ofeli_output_data_ext ]
            proc = Popen(args, stdout=PIPE, stderr=PIPE)
            try:
                out, err = proc.communicate()
            except Exception, e:
                print 'An exception was raised when calling the validation program'
                print e
                self.task_queue.task_done()
                continue

            if err != '' and err is not None:
                print 'Error while performing validation for', task.fixed
                print err
                self.task_queue.task_done()
                continue

            validation_measure_ofeli = float(out)

            args = [ validation_program2, join(validation_data_path, task.fixed) + ofeli_validation_data_ext,
                     last_out_file + ofeli_output_data_ext ]
            proc = Popen(args, stdout=PIPE, stderr=PIPE)
            try:
                out, err = proc.communicate()
            except Exception, e:
                print 'An exception was raised when calling the validation program'
                print e
                self.task_queue.task_done()
                continue

            if err != '' and err is not None:
                print 'Error while performing validation for', task.fixed
                print err
                self.task_queue.task_done()
                continue

            validation_measure_ofeli2 = float(out)
            
            # save results
            self.task_queue.task_done()

            # Prepare the output for the results file

            temp_out = ""
            temp_out += ", ".join( [ str(task.extconf_Ofeli[key])
                                     for key in sorted(task.extconf_Ofeli.iterkeys()) ] )

            output_line = ", ".join(map(str, [ task.fixed, task.moving, temp_out, validation_prior,
                                               validation_measure_ofeli, validation_measure_ofeli2, timing ])) + "\n"
            print validation_label+":", validation_measure_ofeli, "Time:", timing, \
              "Remaining:", self.task_queue.qsize()-1

            # Remove the quotes symbols
            self.result_queue.put( output_line.replace( '"', '' ) )
            shutil.rmtree( output_dir )

options = None
config_Ofeli = {}

if __name__ == '__main__':

    parser = OptionParser()
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose", default=False)
    (options, args) = parser.parse_args()

    if not os.path.exists(ofeli_cmd):
        print "ERROR elastix not found. It is supposed to be in " + elastix
        exit()
    if not os.path.exists(input_data_path):
        print "ERROR Images not found, are you sure you are in the right directory?"
        exit()
    if not os.path.exists(validation_program):
        print "WARNING Validation program not found. It is supposed to be " + validation_program
        exit()

    # # Purge previous results
    # for directory in glob.glob("results_*"):
    #     shutil.rmtree(directory)

    tasks = multiprocessing.JoinableQueue()
    results = multiprocessing.Queue()

    # Remove reported parameters that are already in variable parameters
    reported_parameters = filter(lambda x: x not in variable_parameters_pOne.keys() and x not in variable_parameters_pTwo.keys(),
                                 reported_parameters)

    # Create a list of dictionaries with all combination of values from variable_parameters
    extra_conf_Ofeli = [ dict( zip( variable_parameters_Ofeli.keys(), x ) )
                   for x in itertools.product( *variable_parameters_Ofeli.values() ) ]

    # Create results file header
    temp_out = ""
    temp_out += ", ".join(map(lambda x: 'ofeli_'+x[1:], sorted(extra_conf_Ofeli[0].iterkeys())))

    header = "fixed_img, moving_img, {0}, registration.{1}, ofeli.{1}, ofeli.{2}, time".format(temp_out, validation_label, validation_label2)

    # Process that writes the output as soon as it is produced
    writer = Writer(results, header)
    writer.start()

    # Start executors
    executors = [ Executor(tasks, results) for i in xrange(num_threads) ]
    for w in executors:
        w.start()

    for config_num_Ofeli, ext_Ofeli in enumerate(extra_conf_Ofeli):
        config_Ofeli.update(ext_Ofeli)

        for (fixed, moving) in scenarios:
            for rep in xrange(repetitions):
                tasks.put(Task( fixed, moving, "", ext_Ofeli, rep ))

    # Add a poison pill for each executors
    for i in xrange( num_threads ):
        tasks.put(None)

    try:
        # Wait for all of the tasks to finish
        tasks.join()
        # Poison pill for the writer
        results.put(None)
        # Wait for the writer to finish
        writer.join()
    except KeyboardInterrupt, e:
        print "Interrupted by user"
        os._exit(1)



