﻿/****************************************************************************
**
**
****************************************************************************/

#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iterator>

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>

#include "hybrid.hpp"
#include "linked_list.hpp"

#include "boost/filesystem.hpp"

#include "../vfc/vfc.h"
#include "fftw3.h"

#include <boost/interprocess/sync/file_lock.hpp>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>

using namespace std;
using namespace cv;
namespace fs = boost::filesystem;
namespace ip = boost::interprocess;

class cache_record
{
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & img;
    ar & file_x;
    ar & file_y;
    ar & gamma;
  }
public:
  string img;
  string file_x;
  string file_y;
  double gamma;
  bool matches( string name1, double gamma1)
  {
    if ( (img == name1 ) 
         && (gamma == gamma1) )
      return true;
    else
      return false;
  }
};

class cache_vector
{
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version)
  {
    ar & array;
  }
public:
  std::vector<cache_record> array;
};


//Proto
void print_usage_and_exit();
void process_command_line_parameters(int argc, char *argv[]);
void compute_VFC_with_cache(string, Mat &, dmat & x, dmat & y);
void CVMatToUnsignedChar(Mat, unsigned char * &);
void img2phi(Mat, char* &);
void charToCVMat(Mat *, char *);
void combinePhiAndInput(Mat *, Mat *, char *);
char * rectangle_phi(unsigned int, unsigned int, double, double, double, double);
Mat imreadResize(const string&);
void imwriteResize(const string&, Mat);
void phiToCVMat(Mat * img, char * img_char);

// Chan & Vese
double lambda_in = 1;
double lambda_out = 1;

// Iterations
unsigned int Na = 45;

//2-Cycle, smoothing
bool hasSmoothingCycle = true;
unsigned int Ns = 5;
unsigned int kernel_curve = 5; 
double std_curve = 1.0;
//Extra
bool hasCleaning = true;

string outDir = "out";

//VFC
double gammaVFC = 2.5;

double weightCV = 1;
double weightPrior = 1;
double weightVFC = 1;

bool verbose = false;
bool compute_vfc_only = false;
string phi_init_type = "prior";
string cacheDir = "cache";

bool use_resize = false;
unsigned int orig_rows;
unsigned int orig_cols;

//Command line parameters parsing
const unsigned int n_flags = 16;
const char* flags[] = { "-lambda_in", "-lambda_out", "-Na", "-Ns", "-kernel_curve", "-std_curve", "-out_dir", "-weightCV", 
                        "-weightPrior", "-weightVFC", "-verbose", "-phi_init_type", "-gammaVFC", "-cache_dir", "-compute_vfc_only", "-resize" };
const char  type[]  = { 'f'         , 'f'          , 'i'  , 'i'  , 'i'            , 'i'         , 's'       , 'f'        , 
                        'f'           , 'f'         , 'b'       , 's'            , 'f'         , 's'         , 'b', 'b' };
void* var[]         = { &lambda_in  , &lambda_out  , &Na  , &Ns  , &kernel_curve  , &std_curve  , &outDir   ,  &weightCV ,
                        &weightPrior  , &weightVFC  , &verbose  ,   &phi_init_type, &gammaVFC  , &cacheDir   , &compute_vfc_only, &use_resize };


int main(int argc, char *argv[])
{
  // *** Parameters parsing ***
  if (argc < 2)
    print_usage_and_exit();
 
  string imageFileName = argv[1];
  string priorFileName = argv[2];

  process_command_line_parameters(argc, argv);

  // *** Read the input image ***
  unsigned char * img_uchar;
  unsigned char * prior_uchar;

  // *** Compute the VFC or read it from the cache *** 
  dmat vfc_x, vfc_y;

  //Also reads the input image
  Mat input_img;
  compute_VFC_with_cache(imageFileName, input_img, vfc_x, vfc_y);

  unsigned int rows = input_img.rows;
  unsigned int cols = input_img.cols;
  
  if (verbose)
  {
    saveDmatAsImage("vfc_x.png", vfc_x, 255);
    saveDmatAsImage("vfc_y.png", vfc_y, 255);
    
    float scale = 15;
    cv::Mat pic;
    cv::cvtColor(input_img, pic, CV_GRAY2RGB);
    
    for(int y=0; y < rows; y += 35)
      for(int x=0; x < cols; x += 35)
      {
        cv::Point dest = cv::Point(x + scale*vfc_x[y][x], y + scale*vfc_y[y][x]);
        cv::circle(pic, dest, 3, CV_RGB(255,0,0));
        cv::line(pic, cv::Point(x,y), dest, CV_RGB(255, 0, 0), 2, 8, 0 );
      }

       cv::imwrite("vfc_force.png", pic);
    // cv::imshow("pic", pic);
    // cv::waitKey();
  }

  // *** Read the prior image, compute the initial phi ***

  CVMatToUnsignedChar(input_img, img_uchar);

  Mat prior_img = imreadResize(priorFileName);
  CVMatToUnsignedChar(prior_img, prior_uchar);

  // Create the initial level set directly from the prior
  char * phi_init;

  if (phi_init_type == "prior")
    img2phi(prior_img, phi_init);
    else if (phi_init_type == "rectangle")
      // FOR TESTING without prior
      phi_init = rectangle_phi(cols, rows, 0.3, 0.3, 0, 0);
    else
      print_usage_and_exit();

  ofeli::Hybrid * hybrid 
    = new ofeli::Hybrid(img_uchar, cols, rows, phi_init,
                        hasSmoothingCycle, kernel_curve, std_curve, hasCleaning, 
                        Na, Ns, lambda_out, lambda_in, prior_uchar,
                        weightCV, weightPrior, weightVFC, gammaVFC, vfc_x, vfc_y, verbose);

  char * out_phi;
  Mat * out_img = new Mat(rows, cols, CV_8UC1);

  // *** Create and clean output directory ***
  fs::path dir(outDir);
  if (fs::exists(dir))
  {
    fs::directory_iterator endit;
    for( fs::directory_iterator dir_iter(dir) ; dir_iter != endit ; ++dir_iter)
      if (dir_iter->path().extension() == ".png")
           fs::remove(*dir_iter);
  }
  else
    fs::create_directory(dir);

  unsigned int itr = 0;
  ostringstream filename;

  // *** Actual Level Set Evolution ***
  if (!verbose)
  {
    hybrid->evolve_to_final_state();
    out_phi = hybrid->get_phi();
    charToCVMat( out_img, out_phi);
    filename << outDir << "/out_last.png";
    imwriteResize(filename.str(), *out_img);
  }
  else
  {
    Mat * combined_img = new Mat(rows, cols, CV_8UC3);
    while( !hybrid->get_hasAlgoStopping() )
    {
      hybrid->evolve_one_iteration();
      out_phi = hybrid->get_phi();

      filename.str("");
      // Write the current segmentation
      charToCVMat( out_img, out_phi);
      filename << outDir << "/out" << itr << ".png";
      imwriteResize(filename.str(), *out_img);

      filename.str("");
      // Write the current phi over the input image
      combinePhiAndInput( combined_img, &input_img, out_phi);
      filename << outDir << "/combined" << itr << ".png";
      imwriteResize(filename.str(), *combined_img);

      filename.str("");
      // Write the current phi 
      phiToCVMat( out_img, out_phi);
      filename << outDir << "/phi" << itr << ".png";
      imwriteResize(filename.str(), *out_img);

      itr++;
    }
      filename.str("");
      // Write the current segmentation
      charToCVMat( out_img, out_phi);
      filename << outDir << "/out_last.png";
      imwrite(filename.str(), *out_img);
  }

  return 0;
}

void combinePhiAndInput(Mat * out, Mat * input, char * phi)
{
  unsigned int i = 0;
  MatIterator_<Vec3b> it_out;
  MatConstIterator_<unsigned char> it_in;

  for(it_out = out->begin<Vec3b>(), 
        it_in = input->begin<unsigned char>();
      it_out != out->end<Vec3b>(); it_out++, it_in++, i++ )
  {
    *it_out = Vec3b(*it_in, *it_in, *it_in);

    if ( phi[i] == 1)
      *it_out = Vec3b(255,0,0);
    else if (phi[i] == -1)
      *it_out = Vec3b(0,0,255);
  }
}

void phiToCVMat(Mat * img, char * img_char)
{
  unsigned int i = 0;
  for(MatIterator_<char> it = img->begin<char>(); 
      it != img->end<char>(); it++, i++ )
    switch(img_char[i])
    {
    case 3:
      *it = 0;
      break;
    case 1:
      *it = 50;
      break;
    case -1:
      *it = 100;
      break;
    case -3:
      *it = 150;
      break;
    default:
      *it = 255;
    }
}


void charToCVMat(Mat * img, char * img_char)
{
  unsigned int i = 0;
  for(MatIterator_<unsigned char> it = img->begin<unsigned char>(); 
      it != img->end<unsigned char>(); it++, i++ )
    {
      if (img_char[i] >= 0)
        *it = 0;
      else
        *it = 255;
    }
}

void CVMatToUnsignedChar(Mat img, unsigned char * & img_uchar)
{
  // Convert the image to an array of uchar 
  img_uchar = new unsigned char[img.rows * img.cols];
  unsigned int i = 0;
  for(MatConstIterator_<unsigned char> it = img.begin<unsigned char>(); 
      it != img.end<unsigned char>(); it++, i++ )
    img_uchar[i] = *it;
  
}

void img2phi(Mat img, char* & phi_init)
{
  unsigned int img_size = img.rows * img.cols;
  phi_init = new char[img_size];

  unsigned int i = 0;
  for(MatConstIterator_<unsigned char> it = img.begin<unsigned char>(); 
      it != img.end<unsigned char>(); it++, i++ )
  {
    if( *it < 128 )
      phi_init[i] = 1;
    else
      phi_init[i] = -1;
  }
}

char * rectangle_phi(unsigned int img_width, unsigned int img_height, double init_width, double init_height, double center_x, double center_y)
{
  int img_size = img_width * img_height;
  char * phi = new char[img_size];
  int x,y;
  for( int i = 0; i < img_size; i++ )
  {
    y = i/img_width;
    x = i-y*img_width;

    if( double(y) > ((1.0-init_height)*double(img_height)/2.0 + center_y*double(img_height)) 
        && double(y) < ((double(img_height)-(1.0-init_height)*double(img_height)/2.0) + center_y*double(img_height)) 
        && double(x) > ((1.0-init_width)*double(img_width)/2.0 + center_x*double(img_width)) 
        && double(x) < ((double(img_width)-(1.0-init_width)*double(img_width)/2.0) + center_x*double(img_width)) )
      phi[i] = -1;
    else
      phi[i] = 1;
  }

  return phi;
}

void print_usage_and_exit()
{
  cout << "Usage: command image_file prior_file [-param_name param_value]" << endl;
  cout << "Available parameters: ";
  for(unsigned int i=0; i<n_flags; i++)
    cout << flags[i] << " ";
  cout << endl;
  exit(0);
}

void process_command_line_parameters(int argc, char *argv[])
{
  // cout << "Command line parameters: " << endl;
  unsigned int j;
  char* str;
  char* val;
  for(int i = 3; i < argc; i++)
  {
    str = argv[i];
    for( j = 0; j < n_flags; j++)
    {
      if (strcmp(str, flags[j]) == 0)
        break;
    }
    if (j == n_flags)
      print_usage_and_exit();

    if (type[j] == 'b')
    {
      *((bool *) var[j]) = !*((bool *) var[j]);
      // cout << "-verbose" << endl;
    }
    else
    {
      val = argv[++i];
      switch(type[j])
      {
      case 'f':
        *((double *) var[j]) = atof( val );
        // cout << "-" << flags[j] << "\t" << atof(val) << endl;
        break;
      case 'i':
        *((int *) var[j]) = atoi( val );
        // cout << "-" << flags[j] << "\t" << atoi(val) << endl;
        break;
      case 's':
        *((string *) var[j]) = string( val );
        // cout << "-" << flags[j] << "\t" << string(val) << endl;
        break;
      default:
        print_usage_and_exit();
        break;
      }
    }
  }
}

void compute_VFC_with_cache(string imageFileName, Mat & input_img, dmat & vx, dmat & vy)
{
  //Create the cache directory
  fs::path dir(cacheDir);
  fs::path lockfilename(dir / fs::path(string(".lock")));
  fs::path cachefilename(dir / fs::path(string(".cache")));

  if (!fs::exists(dir))
    fs::create_directory(dir);

  if (!fs::exists(lockfilename))
  {
    ofstream outfile( lockfilename.string().c_str() );
    outfile.close();
  }

  ip::file_lock filelock(lockfilename.string().c_str());
  filelock.lock();

  if (!fs::exists(cachefilename))
  {
    ofstream cachefile( cachefilename.string().c_str() );
    cachefile.close();
  }

  // Read stuff from cache
  cache_vector cache;
  cache_record rec;
  bool in_cache = false;
  ifstream infile(cachefilename.string().c_str(), ios::binary);

  //Is file empty?
  if (infile.peek() != ifstream::traits_type::eof())
  {
    boost::archive::text_iarchive ia(infile);
    ia >> cache;

    for(vector<cache_record>::iterator it=cache.array.begin(); it != cache.array.end(); it++)
    {
      if ( it->matches(imageFileName, gammaVFC) )
      {
        in_cache = true;
        rec = *it;
        break;
      }
    }
  }
  infile.close();

  // Is our VFC in the cache?
  if ( in_cache )
  {
    filelock.unlock();
    if (compute_vfc_only)
      //We are done
      exit(0);

    input_img = imreadResize(imageFileName);
    read_vfc_component( rec.file_x.c_str(), vx, input_img.cols, input_img.rows);
    read_vfc_component( rec.file_y.c_str(), vy, input_img.cols, input_img.rows);
  }
  else
  {
    fftw_init_threads();
    fftw_plan_with_nthreads(4);

    input_img = imreadResize(imageFileName);    
    Mat src(input_img);
    getVFC("", src, ANISOTROPIC_NORMALIZED, vx, vy, gammaVFC, 2, 2, verbose);      // src is PROCESSED

    //Random file names
    fs::path vfc_file_x = fs::unique_path();
    fs::path vfc_file_y = fs::unique_path();

    // Save on disk
    save_vfc_component( ( dir / vfc_file_x ).string().c_str(), vx);
    save_vfc_component( ( dir / vfc_file_y ).string().c_str(), vy);

    // Save to cache file
    cache_record rec;

    rec.img = imageFileName;
    rec.file_x = ( dir / vfc_file_x ).string();
    rec.file_y = ( dir / vfc_file_y ).string();
    rec.gamma = gammaVFC;

    cache.array.push_back(rec);
    
    ofstream outfile(cachefilename.string().c_str(), ios::binary);
    boost::archive::text_oarchive oa(outfile);
    oa << cache;
    outfile.close();

    filelock.unlock();
  }

  if (compute_vfc_only)
    //We are done
    exit(0);
}

Mat imreadResize(const string& fname)
{
  Mat outimg;
  Mat temp = imread(fname, CV_LOAD_IMAGE_GRAYSCALE);

  orig_rows = temp.rows;
  orig_cols = temp.cols;

  if ((orig_rows > 1000 ) || (orig_cols > 1000 ))
    use_resize = true;

  if (use_resize)
  {
    resize(temp, outimg, Size(), 0.5, 0.5);

    return outimg;
  }

  return temp;
}

void imwriteResize(const string& fname, Mat img)
{
  Mat temp;
  if (use_resize)
  {
    resize(img, temp, Size(orig_cols, orig_rows), 0, 0, INTER_NEAREST);
    imwrite(fname, temp);
  }
  else
    imwrite(fname, img);
}
