﻿/****************************************************************************
**
** Copyright (C) 2010-2011 Fabien Bessy.
** All rights reserved.
** Contact: fabien.bessy@gmail.com
**
** This file is part of Ofeli.
**
** You may use this file under the terms of the CeCILL license version 2 as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Fabien Bessy nor the names of its contributors
**     may be used to endorse or promote products derived from this
**     software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
**                    For all legal intents and purposes.
**
****************************************************************************/

#include "ac_withoutedges.hpp"
#include <iostream> // for the object "std::cerr"

namespace ofeli
{

ACwithoutEdges::ACwithoutEdges(const unsigned char* img_data1, int img_width1, int img_height1) : ActiveContour(img_width1, img_height1, false, 0.9, 0.9, 0.0, 0.0, true, 5, 2.0, true, 30, 3), img_data(img_data1), lambda_out(1), lambda_in(1)
{
    if( img_data1 == NULL )
    {
        std::cerr << std::endl << "The pointer img1 must be a non-null pointer, it must be allocated." << std::endl;
    }
    else
    {
        initialize_sums();
        calculate_means();
    }
}

ACwithoutEdges::ACwithoutEdges(const unsigned char* img_data1, int img_width1, int img_height1, bool hasEllipse1, double init_width1, double init_height1, double center_x1, double center_y1, bool hasSmoothingCycle1, int kernel_length1, double sigma1, bool hasCleaning1, int Na1, int Ns1, int lambda_out1, int lambda_in1) : ActiveContour(img_width1, img_height1, hasEllipse1, init_width1, init_height1, center_x1, center_y1, hasSmoothingCycle1, kernel_length1, sigma1, hasCleaning1, Na1, Ns1), img_data(img_data1), lambda_out(lambda_out1), lambda_in(lambda_in1)
{
    if( img_data1 == NULL )
    {
        std::cerr << std::endl << "The pointer img1 must be a non-null pointer, it must be allocated." << std::endl;
    }
    else
    {
        initialize_sums();
        calculate_means();
    }
}

ACwithoutEdges::ACwithoutEdges(const unsigned char* img_data1, int img_width1, int img_height1, const char* phi_init1, bool hasSmoothingCycle1, int kernel_length1, double sigma1, bool hasCleaning1, int Na1, int Ns1, int lambda_out1, int lambda_in1) : ActiveContour(img_width1, img_height1, phi_init1, hasSmoothingCycle1, kernel_length1, sigma1, hasCleaning1, Na1, Ns1), img_data(img_data1), lambda_out(lambda_out1), lambda_in(lambda_in1)
{
    if( img_data1 == NULL )
    {
        std::cerr << std::endl << "The pointer img1 must be a non-null pointer, it must be allocated." << std::endl;
    }
    else
    {
        initialize_sums();
        calculate_means();
    }
}

void ACwithoutEdges::initialize_sums()
{
    sum_in = 0;
    sum_out = 0;
    n_in = 0;
    n_out = 0;

    for( int i = 0; i < img_size; i++ )
    {
        if( phi[i] > 0 )
        {
            sum_out += int(img_data[i]);
            n_out++;
        }
        else
        {
            sum_in += int(img_data[i]);
            n_in++;
        }
    }

    return;
}

inline void ACwithoutEdges::calculate_means()
{
    // protection against /0
    if( n_out == 0 )
    {
        Cout = 0;
    }
    else
    {
        Cout = sum_out/n_out;
    }

    // protection against /0
    if( n_in == 0 )
    {
        Cin = 255;
    }
    else
    {
        Cin = sum_in/n_in;
    }

    return;
}

inline int ACwithoutEdges::compute_external_speed_Fd(int i, int x, int y)
{
    // intensity of the current pixel
    I = int(img_data[i]);

    return lambda_out*square(I-Cout) - lambda_in*square(I-Cin);
}


inline void ACwithoutEdges::do_specific_in_step(int i, int x, int y)
{
    sum_out -= I;
    n_out--;

    sum_in += I;
    n_in++;

    return;
}

inline void ACwithoutEdges::do_specific_out_step(int i, int x, int y)
{
    sum_in -= I;
    n_in--;

    sum_out += I;
    n_out++;

    return;
}

inline void ACwithoutEdges::do_specific_in_step2(int i, int x, int y)
{
    // intensity of the current pixel
    I = int(img_data[i]);

    sum_out -= I;
    n_out--;

    sum_in += I;
    n_in++;

    return;
}

inline void ACwithoutEdges::do_specific_out_step2(int i, int x, int y)
{
    // intensity of the current pixel
    I = int(img_data[i]);

    sum_in -= I;
    n_in--;

    sum_out += I;
    n_out++;

    return;
}

void ACwithoutEdges::do_specific_beginning_step()
{
    calculate_means();

    return;
}

int ACwithoutEdges::get_Cout() const
{
    return Cout;
}

int ACwithoutEdges::get_Cin() const
{
    return Cin;
}

// initilization for video tracking
void ACwithoutEdges::initialize_for_each_frame()
{
    do_general_initialization_for_each_frame();
    initialize_sums();
    calculate_means();

    return;
}

// initilization for video tracking
void ACwithoutEdges::initialize_for_each_frame(const unsigned char* img_data1)
{
    img_data = img_data1;
    this->initialize_for_each_frame();

    return;
}

}
