﻿/****************************************************************************
**
** Copyright (C) 2010-2011 Fabien Bessy.
** All rights reserved.
** Contact: fabien.bessy@gmail.com
**
** This file is part of Ofeli.
**
** You may use this file under the terms of the CeCILL license version 2 as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Fabien Bessy nor the names of its contributors
**     may be used to endorse or promote products derived from this
**     software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
**                    For all legal intents and purposes.
**
****************************************************************************/


#include "filters.hpp"
#include <cmath> // for the function "exp"
#include <cstring> // for the function "memcpy"
#include <iostream> // for the object "std::cerr"
#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>


inline double my_round( double d )
{
    return floor( d + 0.5 );
}

namespace ofeli
{

Filters::Filters(const unsigned char* img_data1, int img_width1, int img_height1) : img_data(img_data1), img_width(img_width1), img_height(img_height1), img_size(img_width1*img_height1), columns_histo(new int [256*img_width1]), kernel_histo(new int [256])
{

    if( img_data1 == NULL )
    {
        std::cerr << std::endl << "The pointer img1 must be a non-null pointer, it must be allocated." << std::endl;
    }
    else
    {
        filtered = new unsigned char[img_size];

        initialyze_filtered();

        filtered_modif = new unsigned char[img_size];

        // buffer for the top-hat transform
        previous_filtered = new unsigned char[img_size];

        // buffers for the anisotropic filter
        diff_img = new double[img_size];
        diff_img1 = new double[img_size];
    }
}

void Filters::initialyze_filtered()
{
    memcpy(filtered,img_data,img_size);
    return;
}

Filters::~Filters()
{
    delete[] previous_filtered;
    delete[] filtered_modif;
    delete[] filtered;

    // buffers deleted for the anisotropic filter
    delete[] diff_img;
    delete[] diff_img1;

    // buffers deleted for the Perreault's algorithm
    delete[] columns_histo;
    delete[] kernel_histo;
}

void Filters::anisotropic_diffusion(int max_itera, double lambda, double kappa, int option)
{

    if( kappa < 0.000000001 )
    {
        kappa = 0.000000001;
    }

    // initial condition of the PDE
    for( int i = 0; i < img_size; i++)
    {
        diff_img[i] = double(filtered[i]);
    }

    // euclidian distance square of the neighbor pixels to the center
    //const double dist_sqr[8] = { dd*dd, dy*dy, dd*dd, dx*dx, dx*dx, dd*dd, dy*dy, dd*dd };
    const double dist_sqr[8] = { 2.0, 1.0, 2.0, 1.0, 1.0, 2.0, 1.0, 2.0 };

    // index for 0 to 7 to access of the 8 gradients and the 8 diffusion coefficients
    int index;
    // sum calculated on the 8 directions for the PDE solving
    double sigma;

    // 8 gradients for each direction for the current pixel n
    double nabla[8];

    // intensity of the current central pixel
    double centrI;

    // 8 coefficients of diffusion for each direction
    double c[8];

    // temporary pointer to swap the pointers of diffused images
    double* ptemp_aniso;

    int x, y; // position of the current pixel


    // choice of the function of diffusion
    switch( option )
    {

    // c(x,y,t) = exp(-(nabla/kappa)^2)
    case 1 :
    {
        // anisotropic diffusion
        for( int itera = 0; itera < max_itera; itera++ )
        {
            // finite differences
            for( int i = 0; i < img_size; i++ )
            {
                y = i/img_width;
                x = i-y*img_width;

                centrI = diff_img[i];

                index = 0;

                // if the current pixel is not in the border
                if( x > 0 && x < img_width-1 && y > 0 && y < img_height-1 )
                {
                    // no tests of neighbors

                    for( int dy = -1; dy <= 1; dy++ )
                    {
                        for( int dx = -1; dx <= 1; dx++ )
                        {
                            if( !( dx == 0 && dy == 0 ) )
                            {
                                nabla[ index++ ] = diff_img[ (x+dx)+(y+dy)*img_width ]-centrI;
                            }
                        }
                    }
                }

                else
                {
                    for( int dy = -1; dy <= 1; dy++ )
                    {
                        for( int dx = -1; dx <= 1; dx++ )
                        {
                            if( !( dx == 0 && dy == 0 ) )
                            {
                                if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                                {
                                    nabla[ index++ ] = diff_img[ (x+dx)+(y+dy)*img_width ]-centrI;
                                }
                                else
                                {
                                    nabla[ index++ ] = 0.0;
                                }
                            }
                        }
                    }
                }

                sigma = 0.0;
                for( index = 0; index < 8; index++ )
                {
                    c[index] = exp( -(nabla[index]/kappa)*(nabla[index]/kappa) );

                    sigma += (1.0/(dist_sqr[index]))*c[index]*nabla[index];
                }

                // solution of the discret PDE
                diff_img1[i] = centrI+lambda*sigma;
            }
            ptemp_aniso = diff_img;
            diff_img = diff_img1;
            diff_img1 = ptemp_aniso;
        }

        break;
    }

    // c(x,y,t) = 1/(1+(1+(nabla/kappa)^2)
    case 2 :
    {
        // anisotropic diffusion
        for( int itera = 0; itera < max_itera; itera++ )
        {
            // finite differences
            for( int i = 0; i < img_size; i++ )
            {
                y = i/img_width;
                x = i-y*img_width;

                centrI = diff_img[i];

                // if the current pixel is not in the border
                if( x > 0 && x < img_width-1 && y > 0 && y < img_height-1 )
                {
                    // no tests of neighbors
                    index = 0;
                    for( int dy = -1; dy <= 1; dy++ )
                    {
                        for( int dx = -1; dx <= 1; dx++ )
                        {
                            if( !( dx == 0 && dy == 0 ) )
                            {
                                nabla[ index++ ] = diff_img[ (x+dx)+(y+dy)*img_width ]-centrI;
                            }
                        }
                    }
                }

                else
                {
                    index = 0;
                    for( int dy = -1; dy <= 1; dy++ )
                    {
                        for( int dx = -1; dx <= 1; dx++ )
                        {
                            if( !( dx == 0 && dy == 0 ) )
                            {
                                if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                                {
                                    nabla[ index++ ] = diff_img[ (x+dx)+(y+dy)*img_width ]-centrI;
                                }
                                else
                                {
                                    nabla[ index++ ] = 0.0;
                                }
                            }
                        }
                    }
                }

                sigma = 0.0;
                for( index = 0; index < 8; index++ )
                {
                    c[index] = 1.0/(1.0+((nabla[index]/kappa)*(nabla[index]/kappa)));

                    sigma += (1.0/(dist_sqr[index]))*c[index]*nabla[index];
                }

                // solution of the discret PDE
                diff_img1[i] = centrI+lambda*sigma;
            }
            ptemp_aniso = diff_img;
            diff_img = diff_img1;
            diff_img1 = ptemp_aniso;
        }
    }
    }

    // result
    for( int i = 0; i < img_size; i++ )
    {
        filtered[i] = (unsigned char)my_round(diff_img[i]);
    }

    return;
}

void Filters::morphological_gradient(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int border_length = (kernel_length-1)/2;

    int x, y; // position of the current pixel

    unsigned char max, min;

    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        // initialization
        max = 0;
        min = 255;

        // if the current pixel is not in the border
        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    // no tests of neighbors
                    if( filtered[ (x+dx)+(y+dy)*img_width ] > max )
                    {
                        max = filtered[ (x+dx)+(y+dy)*img_width ];
                    }
                    if( filtered[ (x+dx)+(y+dy)*img_width ] < min )
                    {
                        min = filtered[ (x+dx)+(y+dy)*img_width ];
                    }
                }
            }
        }
        // if in the border
        else
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    // neighbors tests
                    if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                    {
                        if( filtered[ (x+dx)+(y+dy)*img_width ] > max )
                        {
                            max = filtered[ (x+dx)+(y+dy)*img_width ];
                        }
                        if( filtered[ (x+dx)+(y+dy)*img_width ] < min )
                        {
                            min = filtered[ (x+dx)+(y+dy)*img_width ];
                        }
                    }
                }
            }
        }
        filtered_modif[i] = max-min;
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

void Filters::morphological_gradient_o1(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // variable called radius r in the Perreault and Hébert's article
    // border_length = r
    // kernel_length = 2*r+1
    const int border_length = (kernel_length-1)/2;

    int I;
    unsigned char max, min;

    int x, y; // position of the current pixel

    /////////////////////////////////////////
    // processing of the top of the image //
    ////////////////////////////////////////

    // clear
    for( I = 0; I < 256*img_width; I++ )
    {
        columns_histo[I] = 0;
    }

    // initialization
    for( x = 0; x < img_width; x++ )
    {
        for( y = 0; y < kernel_length; y++ )
        {
            columns_histo[ 256*x+filtered[x+y*img_width] ]++;
        }
    }

    // downward moving in the image
    for( y = 0; y < border_length+1; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y+border_length+1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y+border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ (x-border_length)+y*img_width ] = max-min;
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++)
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length+1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = max-min;
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = max-min;
        }
    }

    //////////////////////////////////////////////////////////////
    // processing of the image without top and bottom stripes  ///
    //////////////////////////////////////////////////////////////

    // clear
    for( I = 0; I < 256*img_width; I++ )
    {
        columns_histo[I] = 0;
    }

    // initialization
    for( x = 0; x < img_width; x++ )
    {
        for( y = 0; y < kernel_length; y++ )
        {
            columns_histo[ 256*x+filtered[x+y*img_width] ]++;
        }
    }

    // downward moving in the image
    for( y = border_length+1; y <= img_height-border_length-2; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y+border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ (x-border_length)+y*img_width ] = max-min;
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++)
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = max-min;
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = max-min;
        }
    }

    ///////////////////////////////////////////
    // processing of the bottom of the image //
    ///////////////////////////////////////////

    // no need to clear and to initialize columns_histo

    for( y = img_height-border_length-1; y < img_height; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y-border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ (x-border_length)+y*img_width ] = max-min;
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = max-min;
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max !=0 )
            {
                max--;
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = max-min;
        }
    }

    if( kernel_length != 1 )
    {
        unsigned char* ptemp = filtered;
        filtered = filtered_modif;
        filtered_modif = ptemp;
    }

    return;
}

void Filters::dilation(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int border_length = (kernel_length-1)/2;

    int x, y; // position of the current pixel

    unsigned char max;

    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        // initialisation
        max = 0;

        // if the current pixel is not in the border
        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    // no neighbors tests
                    if( filtered[ (x+dx)+(y+dy)*img_width ] > max )
                    {
                        max = filtered[ (x+dx)+(y+dy)*img_width ];
                    }
                }
            }
        }
        // if in the border
        else
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    // neighbors tests
                    if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                    {
                        if( filtered[ (x+dx)+(y+dy)*img_width ] > max )
                        {
                            max = filtered[ (x+dx)+(y+dy)*img_width ];
                        }
                    }
                }
            }

        }
        filtered_modif[i] = max;
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

void Filters::dilation_o1(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // variable called radius r in the Perreault and Hébert's article
    // border_length = r
    // kernel_length = 2*r+1
    const int border_length = (kernel_length-1)/2;

    int I;
    unsigned char max;

    int x, y; // position of the current pixel

    /////////////////////////////////////////
    // processing of the top of the image //
    ////////////////////////////////////////

    // clear
    for( I = 0; I < 256*img_width; I++ )
    {
        columns_histo[I] = 0;
    }

    // initialization
    for( x = 0; x < img_width; x++ )
    {
        for( y = 0; y < kernel_length; y++ )
        {
            columns_histo[ 256*x+filtered[x+y*img_width] ]++;
        }
    }

    // downward moving in the image
    for( y = 0; y < border_length+1; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y+border_length+1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y+border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                filtered_modif[ (x-border_length)+y*img_width ] = max;
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length+1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            filtered_modif[ x+y*img_width ] = max;
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            filtered_modif[ x+y*img_width ] = max;
        }
    }

    //////////////////////////////////////////////////////////////
    // processing of the image without top and bottom stripes  ///
    //////////////////////////////////////////////////////////////

    // clear
    for( I = 0; I < 256*img_width; I++ )
    {
        columns_histo[I] = 0;
    }

    // initialization
    for( x = 0; x < img_width; x++ )
    {
        for( y = 0; y < kernel_length; y++ )
        {
            columns_histo[ 256*x+filtered[x+y*img_width] ]++;
        }
    }

    // downward moving in the image
    for( y = border_length+1; y <= img_height-border_length-2; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y+border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                filtered_modif[ (x-border_length)+y*img_width ] = max;
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++)
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            filtered_modif[ x+y*img_width ] = max;
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            filtered_modif[ x+y*img_width ] = max;
        }
    }

    ///////////////////////////////////////////
    // processing of the bottom of the image //
    ///////////////////////////////////////////

    // no need to clear and to initialize columns_histo

    for( y = img_height-border_length-1; y < img_height; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y-border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                filtered_modif[ (x-border_length)+y*img_width ] = max;
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            filtered_modif[ x+y*img_width ] = max;
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the maximum value in the kernel histogram
            max = 255;
            while( kernel_histo[max] == 0 && max != 0 )
            {
                max--;
            }

            filtered_modif[ x+y*img_width ] = max;
        }
    }

    if( kernel_length != 1 )
    {
        unsigned char* ptemp = filtered;
        filtered = filtered_modif;
        filtered_modif = ptemp;
    }

    return;
}

void Filters::erosion(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int border_length = (kernel_length-1)/2;

    int x, y; // position of the current pixel

    unsigned char min;

    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        // initialization
        min = 255;

        // if the current pixel is not in the border
        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    // no neighbors tests
                    if( filtered[ (x+dx)+(y+dy)*img_width ] < min )
                    {
                        min = filtered[ (x+dx)+(y+dy)*img_width ];
                    }
                }
            }
        }
        // if in the border
        else {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    // neighbors tests
                    if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                    {
                        if( filtered[ (x+dx)+(y+dy)*img_width ] < min )
                        {
                            min = filtered[ (x+dx)+(y+dy)*img_width ];
                        }
                    }
                }
            }
        }
        filtered_modif[i] = min;
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

void Filters::erosion_o1(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // variable called radius r in the Perreault and Hébert's article
    // border_length = r
    // kernel_length = 2*r+1
    const int border_length = (kernel_length-1)/2;

    int I;
    unsigned char min;

    int x, y; // position of the current pixel

    /////////////////////////////////////////
    // processing of the top of the image //
    ////////////////////////////////////////

    // clear
    for( I = 0; I < 256*img_width; I++ )
    {
        columns_histo[I] = 0;
    }

    // initialization
    for( x = 0; x < img_width; x++ )
    {
        for( y = 0; y < kernel_length; y++ )
        {
            columns_histo[ 256*x+filtered[x+y*img_width] ]++;
        }
    }

    // downward moving in the image
    for( y = 0; y < border_length+1; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y+border_length+1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y+border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ (x-border_length)+y*img_width ] = min;
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length+1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I]-columns_histo[256*(x-border_length-1)+I ];
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = min;
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        for( x = img_width-border_length-1; x < img_width; x++)
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++)
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = min;
        }
    }

    //////////////////////////////////////////////////////////////
    // processing of the image without top and bottom stripes  ///
    //////////////////////////////////////////////////////////////

    // clear
    for( I = 0; I < 256*img_width; I++ )
    {
        columns_histo[I] = 0;
    }

    // initialization
    for( x = 0; x < img_width; x++ )
    {
        for( y = 0; y < kernel_length; y++ )
        {
            columns_histo[ 256*x+filtered[x+y*img_width] ]++;
        }
    }

    // downward moving in the image
    for( y = border_length+1; y <= img_height-border_length-2; y++ )
    {

        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y+border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ (x-border_length)+y*img_width ] = min;
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = min;
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = min;
        }
    }

    ///////////////////////////////////////////
    // processing of the bottom of the image //
    ///////////////////////////////////////////

    // no need to clear and to initialize columns_histo

    for( y = img_height-border_length-1; y < img_height; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y-border_length)*img_width] ]++;
            for( I = 0; I < 256; I++)
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ (x-border_length)+y*img_width ] = min;
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = min;
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the minimum value in the kernel histogram
            min = 0;
            while( kernel_histo[min] == 0 && min != 255 )
            {
                min++;
            }

            filtered_modif[ x+y*img_width ] = min;
        }
    }

    if( kernel_length !=1 )
    {
        unsigned char* ptemp = filtered;
        filtered = filtered_modif;
        filtered_modif = ptemp;
    }

    return;
}

void Filters::closing(int kernel_length)
{
    dilation(kernel_length);

    // for the top-hat function eventually
    unsigned char* ptemp = filtered_modif;
    filtered_modif = previous_filtered;
    previous_filtered = ptemp;

    erosion(kernel_length);

    return;
}

void Filters::closing_o1(int kernel_length)
{
    dilation_o1(kernel_length);

    // for the top-hat function eventually
    unsigned char* ptemp = filtered_modif;
    filtered_modif = previous_filtered;
    previous_filtered = ptemp;

    erosion_o1(kernel_length);

    return;
}

void Filters::opening(int kernel_length)
{
    erosion(kernel_length);

    // for the top-hat function eventually
    unsigned char* ptemp = filtered_modif;
    filtered_modif = previous_filtered;
    previous_filtered = ptemp;

    dilation(kernel_length);

    return;
}

void Filters::opening_o1(int kernel_length)
{
    erosion_o1(kernel_length);

    // for the top-hat function eventually
    unsigned char* ptemp = filtered_modif;
    filtered_modif = previous_filtered;
    previous_filtered = ptemp;

    dilation_o1(kernel_length);

    return;
}

void Filters::black_top_hat(int kernel_length)
{
    closing(kernel_length);

    for( int i = 0; i < img_size; i++ )
    {
        if( filtered[i] >= previous_filtered[i] )
        {
            filtered[i] -= previous_filtered[i];
        }
        else
        {
            filtered[i] = 0;
        }
    }

    return;
}

void Filters::black_top_hat_o1(int kernel_length)
{
    closing_o1(kernel_length);

    for( int i = 0; i < img_size; i++ )
    {
        if( filtered[i] >= previous_filtered[i] )
        {
            filtered[i] -= previous_filtered[i];
        }
        else
        {
            filtered[i] = 0;
        }
    }

    return;
}

void Filters::white_top_hat(int kernel_length)
{
    opening(kernel_length);

    for( int i = 0; i < img_size; i++ )
    {
        if( previous_filtered[i] >= filtered[i] )
        {
            filtered[i] = previous_filtered[i]-filtered[i];
        }
        else
        {
            filtered[i] = 0;
        }
    }

    return;
}

void Filters::white_top_hat_o1(int kernel_length)
{
    opening_o1(kernel_length);

    for( int i = 0; i < img_size; i++ )
    {
        if( previous_filtered[i] >= filtered[i] )
        {
            filtered[i] = previous_filtered[i]-filtered[i];
        }
        else
        {
            filtered[i] = 0;
        }
    }

    return;
}

void Filters::mean_filtering(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    int x, y; // position of the current pixel

    const int border_length = (kernel_length-1)/2;
    int sum;

    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        sum = 0;

        // if the current pixel is not in the border
        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    sum += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }
        }
        // if in the border
        else
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {

                    if( (x+dx >= 0 && x+dx < img_width) && (y+dy >= 0 && y+dy < img_height) )
                    {
                        sum += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                    }
                    if( (x+dx >= 0 && x+dx < img_width) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                    {
                        sum += int( filtered[ (x+dx)+(y-dy)*img_width ] );
                    }
                    if( ( !(x+dx >= 0 && x+dx < img_width) ) && (y+dy >= 0 && y+dy < img_height) )
                    {
                        sum += int( filtered[ (x-dx)+(y+dy)*img_width ] );
                    }
                    if( ( !(x+dx >= 0 && x+dx < img_width) ) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                    {
                        sum += int( filtered[ (x-dx)+(y-dy)*img_width ] );
                    }
                }
            }
        }
        filtered_modif[i] = (unsigned char)(sum/(kernel_length*kernel_length));

    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

const double* const Filters::gaussian_kernel(int kernel_length, double sigma)
{
    // to protect the input : kernel_length impair et strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // to protect against /0
    if( sigma < 0.000000001 )
    {
        sigma = 0.000000001;
    }

    const int kernel_size = kernel_length*kernel_length;

    double* const kernel = new double[kernel_size];

    int x;
    int y;
    const int k = (kernel_length-1)/2;

    double sum = 0.0;

    for(int i = 0; i < kernel_size; i++)
    {
        y = i/kernel_length;
        x = i-y*kernel_length;

        kernel[i] = exp( - ((double(y)-double(k))*(double(y)-double(k))+(double(x)-double(k))*(double(x)-double(k))) / (2*sigma*sigma));

        sum += kernel[i];
    }

    // to normalize
    for( int i = 0; i < kernel_size; i++ )
    {
        kernel[i] = kernel[i]/sum;
    }

    return kernel;
}

void Filters::gaussian_filtering(int kernel_length, double sigma)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // to protect against /0
    if( sigma < 0.000000001 )
    {
        sigma = 0.000000001;
    }

    const double* const mask = gaussian_kernel(kernel_length, sigma);

    int x, y; // position of the current pixel

    const int border_length = (kernel_length-1)/2;
    double sum;

    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        sum = 0.0;

        // if the current pixel is not in the border
        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }
        }
        // if is in the border
        else {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    if( (x+dx >= 0 && x+dx < img_width) && (y+dy >= 0 && y+dy < img_height) )
                    {
                        sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ (x+dx)+(y+dy)*img_width ] );
                    }
                    if( (x+dx >= 0 && x+dx < img_width) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                    {
                        sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ (x+dx)+(y-dy)*img_width ] );
                    }
                    if( ( !(x+dx >= 0 && x+dx < img_width) ) && (y+dy >= 0 && y+dy < img_height) )
                    {
                        sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ (x-dx)+(y+dy)*img_width ] );
                    }
                    if( ( !(x+dx >= 0 && x+dx < img_width) ) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                    {
                        sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ (x-dx)+(y-dy)*img_width ] );
                    }
                }
            }
        }
        filtered_modif[i] = (unsigned char)my_round(sum);

    }

    delete[] mask;

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

inline void Filters::swap(unsigned char* const array, int a, int b)
{
    unsigned char temp = array[a];
    array[a] = array[b];
    array[b] = temp;

    return;
}

void Filters::quick_sort(unsigned char* const array, int begin, int end)
{
    int left = begin-1;
    int right = end+1;
    const unsigned char pivot = array[begin];

    // if the length of the array is 0, to do nothing
    if( begin >= end )
    {
        return;
    }

    while( 1 )
    {
        do
        {
            right--;
        }
        while( array[right] > pivot );

        do
        {
            left++;
        }
        while( array[left] < pivot );

        if( left < right )
        {
            swap(array, left, right);
        }
        else
        {
            break;
        }
    }

    quick_sort(array, begin, right);
    quick_sort(array, right+1, end);

    return;
}

void Filters::median_filtering_oNlogN(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0)
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    int x, y; // position of the current pixel

    const int border_length = (kernel_length-1)/2;
    int m;

    const int length = kernel_length*kernel_length;
    unsigned char* const median_kernel = new unsigned char[length];

    // i, offset
    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        // m : index of the kernel
        m = 0;

        // if the current pixel is not in the border
        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    median_kernel[m++] = filtered[ (x+dx)+(y+dy)*img_width ];
                }
            }
        }
        // if in the border
        else
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    if( (x+dx >= 0 && x+dx < img_width) && (y+dy >= 0 && y+dy < img_height) )
                    {
                        median_kernel[m++] = filtered[ (x+dx)+(y+dy)*img_width ];
                    }
                    if( (x+dx >= 0 && x+dx < img_width) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                    {
                        median_kernel[m++] = filtered[ (x+dx)+(y-dy)*img_width ];
                    }
                    if( ( !(x+dx >= 0 && x+dx < img_width) ) && (y+dy >= 0 && y+dy < img_height) )
                    {
                        median_kernel[m++] = filtered[ (x-dx)+(y+dy)*img_width ];
                    }
                    if( ( !(x+dx >= 0 && x+dx < img_width) ) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                    {
                        median_kernel[m++] = filtered[ (x-dx)+(y-dy)*img_width ];
                    }
                }
            }
        }

        quick_sort(median_kernel,0,length-1);

        filtered_modif[i] = median_kernel[(length-1)/2];
    }

    delete[] median_kernel;

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;

}

void Filters::median_filtering_o1(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int median_position = 1+kernel_length*kernel_length/2;

    // variable called radius r in the Perreault and Hébert's article
    // border_length = r
    // kernel_length = 2*r+1
    const int border_length = (kernel_length-1)/2;

    int I;

    int x, y; // position of the current pixel

    int n, m;

    /////////////////////////////////////////
    // processing of the top of the image //
    ////////////////////////////////////////

    // clear
    for( I = 0; I < 256*img_width; I++ )
    {
        columns_histo[I] = 0;
    }

    // initialization
    for( x = 0; x < img_width; x++ )
    {
        for( y = 0; y < kernel_length; y++ )
        {
            columns_histo[ 256*x+filtered[x+y*img_width] ]++;
        }
    }

    // downward moving in the image
    for( y = 0; y < border_length+1; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y+border_length+1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y+border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the value I of the median in the kernel histogram
                n = 0;
                I = -1;
                while( n <= kernel_length*(x+1)/2 )
                {
                    n += kernel_histo[++I];
                }
                filtered_modif[ (x-border_length)+y*img_width ] = (unsigned char)(I);
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length+1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++)
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the value I of the median in the kernel histogram
            n = 0;
            I = -1;
            while( n <= (kernel_length*kernel_length-1)/2 )
            {
                n += kernel_histo[++I];
            }

            filtered_modif[ x+y*img_width ] = (unsigned char)(I);
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        m = 1;
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the value I of the median in the kernel histogram
            n = 0;
            I = -1;
            while( n <= (kernel_length*(kernel_length-m))/2 )
            {
                n += kernel_histo[++I];
            }

            filtered_modif[ x+y*img_width ] = (unsigned char)(I);
            m++;
        }
    }

    //////////////////////////////////////////////////////////////
    // processing of the image without top and bottom stripes  ///
    //////////////////////////////////////////////////////////////

    // clear
    for( I = 0; I < 256*img_width; I++)
    {
        columns_histo[I] = 0;
    }

    // initialization
    for( x = 0; x < img_width; x++ )
    {
        for( y = 0; y < kernel_length; y++ )
        {
            columns_histo[ 256*x+filtered[x+y*img_width] ]++;
        }
    }

    // downward moving in the image
    for( y = border_length+1; y <= img_height-border_length-2; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y+border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the value I of the median in the kernel histogram
                n = 0;
                I = -1;
                while( n <= kernel_length*(x+1)/2 )
                {
                    n += kernel_histo[++I];
                }
                filtered_modif[ (x-border_length)+y*img_width ] = (unsigned char)(I);
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y+border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++)
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the value I of the median in the kernel histogram
            n = 0;
            I = -1;
            while( n < median_position )
            {
                n += kernel_histo[++I];
            }

            filtered_modif[ x+y*img_width ] = (unsigned char)(I);
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        m = 1;
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the value I of the median in the kernel histogram
            n = 0;
            I = -1;
            while( n <= (kernel_length*(kernel_length-m))/2 )
            {
                n += kernel_histo[++I];
            }

            filtered_modif[ x+y*img_width ] = (unsigned char)(I);
            m++;
        }
    }

    ///////////////////////////////////////////
    // processing of the bottom of the image //
    ///////////////////////////////////////////

    // no need to clear and to initialize columns_histo

    for( y = img_height-border_length-1; y < img_height; y++ )
    {
        // clear
        for( I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        // initialization for each new current row
        for( x = 0; x < kernel_length; x++ )
        {
            columns_histo[ 256*x+filtered[x+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*x+filtered[x+(y-border_length)*img_width] ]++;
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*x+I ];
            }

            ///////////////////////////
            // left border processed //
            ///////////////////////////

            if( x >= border_length )
            {
                // to find the value I of the median in the kernel histogram
                n = 0;
                I = -1;
                while( n <= kernel_length*(x+1)/2 )
                {
                    n += kernel_histo[++I];
                }
                filtered_modif[ (x-border_length)+y*img_width ] = (unsigned char)(I);
            }
        }

        //////////////////////
        // center processed //
        //////////////////////

        // rightward moving in the image
        for( x = border_length+1; x <= img_width-border_length-2; x++ )
        {
            // to update the column histogram H(x+border_length)
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length-1)*img_width] ]--;
            columns_histo[ 256*(x+border_length)+filtered[(x+border_length)+(y-border_length)*img_width] ]++;

            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the value I of the median in the kernel histogram
            n = 0;
            I = -1;
            while( n < median_position )
            {
                n += kernel_histo[++I];
            }

            filtered_modif[ x+y*img_width ] = (unsigned char)(I);
        }

        ////////////////////////////
        // right border processed //
        ////////////////////////////

        // rightward moving in the image
        m = 1;
        for( x = img_width-border_length-1; x < img_width; x++ )
        {
            // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
            }

            // to find the value I of the median in the kernel histogram
            n = 0;
            I = -1;
            while( n <= (kernel_length*(kernel_length-m))/2 )
            {
                n += kernel_histo[++I];
            }

            filtered_modif[ x+y*img_width ] = (unsigned char)(I);
            m++;
        }
    }

    if( kernel_length != 1 )
    {
        unsigned char* ptemp = filtered;
        filtered = filtered_modif;
        filtered_modif = ptemp;
    }

    return;
}

void Filters::gaussian_white_noise(double sigma)
{
    // generator
    boost::mt19937 rng(time(0));
    // distribution
    boost::normal_distribution<> nd(0.0, sigma);

    // link between the generator and the distribution
    boost::variate_generator<boost::mt19937&,
                             boost::normal_distribution<> > var_nor(rng, nd);

    int filtered_int;

    for( int i = 0; i < img_size; i++ )
    {
        filtered_int = int(filtered[i]) + (int)(var_nor());

        if( filtered_int < 0 )
        {
            filtered_int = 0;
        }
        if( filtered_int > 255 )
        {
            filtered_int = 255;
        }
        filtered[i] = (unsigned char)(filtered_int);
    }

    return;
}

void Filters::salt_pepper_noise(double proba)
{
    // generator
    boost::mt19937 rng(time(0));
    // distribution
    boost::uniform_01<> ud;

    // link between the generator and the distribution
    boost::variate_generator<boost::mt19937&,
                             boost::uniform_01<> > var_uni(rng, ud);

    for( int i = 0; i < img_size; i++ )
    {
        if( var_uni() < proba )
        {
            if( var_uni() < 0.5 )
            {
                filtered[i] = 0;
            }
            else
            {
                filtered[i] = 255;
            }
        }
    }

    return;
}

void Filters::speckle(double sigma)
{
    // generator
    boost::mt19937 rng(time(0));
    // distribution
    boost::uniform_01<> ud;

    // link between the generator and the distribution
    boost::variate_generator<boost::mt19937&,
                             boost::uniform_01<> > var_uni(rng, ud);

    int filtered_int;

    for( int i = 0; i < img_size; i++ )
    {
        // img = img + X * img
        // with X a random variable from a uniform distribution centered at 0
        // and of a standard deviation sigma
        // sigma * sqrt(12) is the interval of the uniform distribution
        filtered_int = int(filtered[i]) + int (double(filtered[i]) * sigma * sqrt(12.0) * (var_uni()-0.5));

        if( filtered_int < 0 )
        {
            filtered_int = 0;
        }
        if(filtered_int > 255 )
        {
            filtered_int = 255;
        }
        filtered[i] = (unsigned char)(filtered_int);
    }

    return;
}

void Filters::local_binary_pattern(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int border_length = (kernel_length-1)/2;

    int x, y; // position of the current pixel

    unsigned char Icentr;

    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        Icentr = filtered[i];

        filtered_modif[i] = 0;

        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            if( filtered[ (x-border_length)+(y-border_length)*img_width ] >= Icentr )
            {
                filtered_modif[ x+y*img_width ] += 1;
            }
            if( filtered[ x+(y-border_length)*img_width ] >= Icentr )
            {
                filtered_modif[ x+y*img_width ] += 2;
            }
            if( filtered[ (x+border_length)+(y-border_length)*img_width] >= Icentr )
            {
                filtered_modif[ x+y*img_width ] += 4;
            }
            if( filtered[ (x-border_length)+y*img_width ] >= Icentr )
            {
                filtered_modif[ x+y*img_width ] += 8;
            }
            if( filtered[ (x+border_length)+y*img_width ] >= Icentr )
            {
                filtered_modif[ x+y*img_width ] += 16;
            }
            if( filtered[ (x-border_length)+(y+border_length)*img_width ] >= Icentr )
            {
                filtered_modif[ x+y*img_width ] += 32;
            }
            if( filtered[ x+(y+border_length)*img_width ] >= Icentr )
            {
                filtered_modif[ x+y*img_width ] += 64;
            }
            if( filtered[ (x+border_length)+(y+border_length)*img_width ] >= Icentr )
            {
                filtered_modif[ x+y*img_width ] += 128;
            }
        }
    }

    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        // clear
        for( int I = 0; I < 256; I++ )
        {
            kernel_histo[I] = 0;
        }

        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    kernel_histo[ filtered_modif[(x+dx)+(x+dy)*img_width] ]++;
                }
            }
        }
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}


void Filters::nagao_filtering(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    int x, y; // position of the current pixel

    const int border_length = (kernel_length-1)/2;

    int var[9];
    int sum[9];

    int min_var;

    int nagao_size_init = square(border_length+1);
    if( nagao_size_init <= 0 )
    {
        nagao_size_init = 1;
    }
    const int nagao_size = nagao_size_init;

    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        for( int n = 0; n < 9; n++ )
        {
            sum[n] = 0;
            var[n] = 0;
        }

        // if not in the border
        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length)
        {
            for( int dy = 0; dy >= -border_length; dy-- )
            {
                for( int dx = dy; dx <= -dy; dx++ )
                {
                    sum[0] += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }
            for( int dx = 0; dx <= border_length; dx++ )
            {
                for( int dy = -dx; dy <= dx; dy++ )
                {
                    sum[1] += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }
            for( int dy = 0; dy <= border_length; dy++ )
            {
                for( int dx = -dy; dx <= dy; dx++ )
                {
                    sum[2] += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }
            for( int dx = 0; dx >= -border_length; dx-- )
            {
                for( int dy = dx; dy <= -dx; dy++ )
                {
                    sum[3] += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }

            for( int dy = -border_length; dy <= 0; dy++ )
            {
                for( int dx = 0; dx <= border_length; dx++ )
                {
                    sum[4] += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }
            for( int dy = 0; dy <= border_length; dy++ )
            {
                for( int dx = 0; dx <= border_length; dx++ )
                {
                    sum[5] += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }
            for( int dy = 0; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= 0; dx++ )
                {
                    sum[6] += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }
            for( int dy = -border_length; dy <= 0; dy++ )
            {
                for( int dx = -border_length; dx <= 0; dx++ )
                {
                    sum[7] += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }

            for( int dy = -border_length+1; dy <= border_length-1; dy++ )
            {
                for( int dx = -border_length+1; dx <= border_length-1; dx++ )
                {
                    sum[8] += int( filtered[ (x+dx)+(y+dy)*img_width ] );
                }
            }
        }

        // if not in the border
        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            for(int dy = 0; dy >= -border_length; dy-- )
            {
                for( int dx = dy; dx <= -dy; dx++ )
                {
                    var[0] += square( sum[0]-int(filtered[ (x+dx)+(y+dy)*img_width ]) );
                }
            }
            for( int dx = 0; dx <= border_length; dx++ )
            {
                for( int dy = -dx; dy <= dx; dy++ )
                {
                    var[1] += square( sum[1]-int(filtered[ (x+dx)+(y+dy)*img_width ]) );
                }
            }
            for( int dy = 0; dy <= border_length; dy++ )
            {
                for(int dx = -dy; dx <= dy; dx++ )
                {
                    var[2] += square( sum[2]-int(filtered[ (x+dx)+(y+dy)*img_width] ) );
                }
            }
            for( int dx = 0; dx >= -border_length; dx-- )
            {
                for( int dy = dx; dy <= -dx; dy++ )
                {
                    var[3] += square( sum[3]-int(filtered[ (x+dx)+(y+dy)*img_width] ) );
                }
            }

            for( int dy = -border_length; dy <= 0; dy++ )
            {
                for( int dx = 0; dx <= border_length; dx++ )
                {
                    var[4] += square( sum[4]-int(filtered[ (x+dx)+(y+dy)*img_width ]) );
                }
            }
            for( int dy = 0; dy <= border_length; dy++ )
            {
                for( int dx = 0; dx <= border_length; dx++ )
                {
                    var[5] += square( sum[5]-int(filtered[ (x+dx)+(y+dy)*img_width ]) );
                }
            }
            for( int dy = 0; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= 0; dx++ )
                {
                    var[6] += square( sum[6]-int(filtered[ (x+dx)+(y+dy)*img_width ]) );
                }
            }
            for( int dy = -border_length; dy <= 0; dy++ )
            {
                for( int dx = -border_length; dx <= 0; dx++ )
                {
                    var[7] += square( sum[7]-int(filtered[ (x+dx)+(y+dy)*img_width ]) );
                }
            }

            for( int dy = -border_length+1; dy <= border_length-1; dy++ )
            {
                for( int dx = -border_length+1; dx <= border_length-1; dx++ )
                {
                    var[8] += square(sum[8]-int(filtered[(x+dx)+(y+dy)*img_width]));
                }
            }
        }

        min_var = 999999999;
        for( int n = 0; n < 9; n++ )
        {
            if( var[n] < min_var )
            {
                min_var = var[n];
            }
        }

        for( int n = 0; n < 9; n++ )
        {
            if( var[n] == min_var )
            {
                filtered_modif[i] = (unsigned char)(sum[n]/nagao_size);
            }
        }
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

void Filters::fas(int kernel_length)
{
    for( int i = 3; i < kernel_length; i += 2 )
    {
        opening(i);
        closing(i);
    }
}

int Filters::square(int value)
{
    return value*value;
}

const unsigned char* const Filters::get_filtered() const
{
    return filtered;
}

}
