﻿/****************************************************************************
**
** Copyright (C) 2010-2011 Fabien Bessy.
** All rights reserved.
** Contact: fabien.bessy@gmail.com
**
** This file is part of Ofeli.
**
** You may use this file under the terms of the CeCILL license version 2 as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Fabien Bessy nor the names of its contributors
**     may be used to endorse or promote products derived from this
**     software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
**                    For all legal intents and purposes.
**
****************************************************************************/


#include "filters_rgb.hpp"
#include <cmath> // for the function "exp"
#include <cstring> // for the function "memcpy"
#include <iostream> // for the object "std::cerr"
#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>

inline double my_round( double d )
{
  return floor( d + 0.5 );
}

namespace ofeli
{

FiltersRGB::FiltersRGB(const unsigned char* img_rgb_data1, int img_width1, int img_height1) : img_rgb_data(img_rgb_data1), img_width(img_width1), img_height(img_height1), img_size(img_width1*img_height1), columns_histo(new int [256*img_width1]), kernel_histo(new int [256])
{
    if( img_rgb_data1 == NULL )
    {
        std::cerr << std::endl << "The pointer img1 must be a non-null pointer, it must be allocated." << std::endl;
    }
    else
    {
        filtered = new unsigned char[3*img_size];

        initialyze_filtered();

        filtered_modif = new unsigned char[3*img_size];

        // buffer for the top-hat transform
        previous_filtered = new unsigned char[3*img_size];

        // buffers for the anisotropic filter
        diff_img = new double[img_size];
        diff_img1 = new double[img_size];

        gradient = new unsigned char[img_size];
    }
}

void FiltersRGB::initialyze_filtered()
{
    memcpy(filtered,img_rgb_data,3*img_size);
    return;
}

FiltersRGB::~FiltersRGB()
{
    delete[] previous_filtered;
    delete[] filtered_modif;
    delete[] filtered;

    // buffers deleted for the anisotropic filter
    delete[] diff_img;
    delete[] diff_img1;

    // buffers deleted for the Perreault's algorithm
    delete[] columns_histo;
    delete[] kernel_histo;
}

void FiltersRGB::anisotropic_diffusion(int max_itera, double lambda, double kappa, int option)
{
    if( kappa < 0.000000001 )
    {
        kappa = 0.000000001;
    }

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        // initial condition of the PDE
        for( int i = 0; i < img_size; i++)
        {
            diff_img[i] = double(filtered[3*i+color_channel]);
        }

        // euclidian distance square of the neighbor pixels to the center
        //const double dist_sqr[8] = { dd*dd, dy*dy, dd*dd, dx*dx, dx*dx, dd*dd, dy*dy, dd*dd };
        const double dist_sqr[8] = { 2.0, 1.0, 2.0, 1.0, 1.0, 2.0, 1.0, 2.0 };

        // index for 0 to 7 to access of the 8 gradients and the 8 diffusion coefficients
        int index;
        // sum calculated on the 8 directions for the PDE solving
        double sigma;

        // 8 gradients for each direction for the current pixel n
        double nabla[8];

        // intensity of the current central pixel
        double centrI;

        // 8 coefficients of diffusion for each direction
        double c[8];

        // temporary pointer to swap the pointers of diffused images
        double* ptemp_aniso;

        int x, y; // position of the current pixel


        // choice of the function of diffusion
        switch( option )
        {
        // c(x,y,t) = exp(-(nabla/kappa)^2)
        case 1 :
        {
            // anisotropic diffusion
            for( int itera = 0; itera < max_itera; itera++ )
            {
                // finite differences
                for( int i = 0; i < img_size; i++ )
                {
                    y = i/img_width;
                    x = i-y*img_width;

                    centrI = diff_img[i];

                    index = 0;

                    // if the current pixel is not in the border
                    if( x > 0 && x < img_width-1 && y > 0 && y < img_height-1 )
                    {
                        // no tests of neighbors
                        for( int dy = -1; dy <= 1; dy++ )
                        {
                            for( int dx = -1; dx <= 1; dx++ )
                            {
                                if( !( dx == 0 && dy == 0 ) )
                                {
                                    nabla[ index++ ] = diff_img[ (x+dx)+(y+dy)*img_width ]-centrI;
                                }
                            }
                        }
                    }

                    else
                    {
                        for( int dy = -1; dy <= 1; dy++ )
                        {
                            for( int dx = -1; dx <= 1; dx++ )
                            {
                                if( !( dx == 0 && dy == 0 ) )
                                {
                                    if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                                    {
                                        nabla[ index++ ] = diff_img[ (x+dx)+(y+dy)*img_width ]-centrI;
                                    }
                                    else
                                    {
                                        nabla[ index++ ] = 0.0;
                                    }
                                }
                            }
                        }

                    }

                    sigma = 0.0;
                    for( index = 0; index < 8; index++ )
                    {
                        c[index] = exp( -(nabla[index]/kappa)*(nabla[index]/kappa) );

                        sigma += (1.0/(dist_sqr[index]))*c[index]*nabla[index];
                    }

                    // solution of the discret PDE
                    diff_img1[i] = centrI+lambda*sigma;

                }
                ptemp_aniso = diff_img;
                diff_img = diff_img1;
                diff_img1 = ptemp_aniso;
            }

            break;
        }

        // c(x,y,t) = 1/(1+(1+(nabla/kappa)^2)
        case 2 :
        {
            // anisotropic diffusion
            for( int itera = 0; itera < max_itera; itera++ )
            {
                // finite differences
                for( int i = 0; i < img_size; i++ )
                {
                    y = i/img_width;
                    x = i-y*img_width;

                    centrI = diff_img[i];

                    // if the current pixel is not in the border
                    if( x > 0 && x < img_width-1 && y > 0 && y < img_height-1 )
                    {
                        // no tests of neighbors

                        index = 0;
                        for( int dy = -1; dy <= 1; dy++ )
                        {
                            for( int dx = -1; dx <= 1; dx++ )
                            {
                                if( !( dx == 0 && dy == 0 ) )
                                {
                                    nabla[ index++ ] = diff_img[ (x+dx)+(y+dy)*img_width ]-centrI;
                                }
                            }
                        }
                    }

                    else
                    {
                        index = 0;
                        for( int dy = -1; dy <= 1; dy++ )
                        {
                            for( int dx = -1; dx <= 1; dx++ )
                            {
                                if( !( dx == 0 && dy == 0 ) )
                                {
                                    if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                                    {
                                        nabla[ index++ ] = diff_img[ (x+dx)+(y+dy)*img_width ]-centrI;
                                    }
                                    else
                                    {
                                        nabla[ index++ ] = 0.0;
                                    }
                                }
                            }
                        }

                    }

                    sigma = 0.0;
                    for( index = 0; index < 8; index++ )
                    {
                        c[index] = 1.0/(1.0+((nabla[index]/kappa)*(nabla[index]/kappa)));

                        sigma += (1.0/(dist_sqr[index]))*c[index]*nabla[index];
                    }

                    // solution of the discret PDE
                    diff_img1[i] = centrI+lambda*sigma;

                }
                ptemp_aniso = diff_img;
                diff_img = diff_img1;
                diff_img1 = ptemp_aniso;
            }
        }
        }

        // result
        for( int i = 0; i < img_size; i++ )
        {
            filtered[3*i+color_channel] = (unsigned char)my_round(diff_img[i]);
        }

    }

    return;
}

void FiltersRGB::morphological_gradient(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int border_length = (kernel_length-1)/2;

    int x, y; // position of the current pixel

    unsigned char max, min;

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        for( int i = 0; i < img_size; i++ )
        {
            y = i/img_width;
            x = i-y*img_width;

            // initialization
            max = 0;
            min = 255;

            // if the current pixel is not in the border
            if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        // no tests of neighbors
                        if( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] > max )
                        {
                            max = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                        }
                        if( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] < min )
                        {
                            min = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                        }
                    }
                }
            }
            // if in the border
            else
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        // neighbors tests
                        if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                        {
                            if( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] > max )
                            {
                                max = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                            }
                            if( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] < min )
                            {
                                min = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                            }
                        }
                    }
                }
            }
            filtered_modif[3*i+color_channel] = max-min;
        }
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

void FiltersRGB::morphological_gradient_o1(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // variable called radius r in the Perreault and Hébert's article
    // border_length = r
    // kernel_length = 2*r+1
    const int border_length = (kernel_length-1)/2;

    int I;
    unsigned char max, min;

    int x, y; // position of the current pixel

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        /////////////////////////////////////////
        // processing of the top of the image //
        ////////////////////////////////////////

        // clear
        for( I = 0; I < 256*img_width; I++ )
        {
            columns_histo[I] = 0;
        }

        // initialization
        for( x = 0; x < img_width; x++ )
        {
            for( y = 0; y < kernel_length; y++ )
            {
                columns_histo[ 256*x+filtered[3*(x+y*img_width)+color_channel] ]++;
            }
        }

        // downward moving in the image
        for( y = 0; y < border_length+1; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y+border_length+1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y+border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the maximum value in the kernel histogram
                    max = 255;
                    while( kernel_histo[max] == 0 && max != 0 )
                    {
                        max--;
                    }

                    // to find the minimum value in the kernel histogram
                    min = 0;
                    while( kernel_histo[min] == 0 && min != 255 )
                    {
                        min++;
                    }

                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = max-min;
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++)
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length+1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max-min;
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max-min;
            }
        }

        //////////////////////////////////////////////////////////////
        // processing of the image without top and bottom stripes  ///
        //////////////////////////////////////////////////////////////

        // clear
        for( I = 0; I < 256*img_width; I++ )
        {
            columns_histo[I] = 0;
        }

        // initialization
        for( x = 0; x < img_width; x++ )
        {
            for( y = 0; y < kernel_length; y++ )
            {
                columns_histo[ 256*x+filtered[3*(x+y*img_width)+color_channel] ]++;
            }
        }

        // downward moving in the image
        for( y = border_length+1; y <= img_height-border_length-2; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y+border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the maximum value in the kernel histogram
                    max = 255;
                    while( kernel_histo[max] == 0 && max != 0 )
                    {
                        max--;
                    }

                    // to find the minimum value in the kernel histogram
                    min = 0;
                    while( kernel_histo[min] == 0 && min != 255 )
                    {
                        min++;
                    }

                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = max-min;
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++)
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max-min;
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max-min;
            }
        }

        ///////////////////////////////////////////
        // processing of the bottom of the image //
        ///////////////////////////////////////////

        // no need to clear and to initialize columns_histo

        for( y = img_height-border_length-1; y < img_height; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y-border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the maximum value in the kernel histogram
                    max = 255;
                    while( kernel_histo[max] == 0 && max != 0 )
                    {
                        max--;
                    }

                    // to find the minimum value in the kernel histogram
                    min = 0;
                    while( kernel_histo[min] == 0 && min != 255 )
                    {
                        min++;
                    }

                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = max-min;
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max-min;
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max !=0 )
                {
                    max--;
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max-min;
            }
        }
    }

    if( kernel_length != 1 )
    {
        unsigned char* ptemp = filtered;
        filtered = filtered_modif;
        filtered_modif = ptemp;
    }

    return;
}

void FiltersRGB::dilation(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int border_length = (kernel_length-1)/2;

    int x, y; // position of the current pixel

    unsigned char max;

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        for( int i = 0; i < img_size; i++ )
        {
            y = i/img_width;
            x = i-y*img_width;

            // initialisation
            max = 0;

            // if the current pixel is not in the border
            if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        // no neighbors tests
                        if( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] > max )
                        {
                            max = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                        }
                    }
                }
            }
            // if in the border
            else
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        // neighbors tests
                        if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                        {
                            if( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] > max )
                            {
                                max = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                            }
                        }
                    }
                }

            }
            filtered_modif[3*i+color_channel] = max;
        }
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

void FiltersRGB::dilation_o1(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // variable called radius r in the Perreault and Hébert's article
    // border_length = r
    // kernel_length = 2*r+1
    const int border_length = (kernel_length-1)/2;

    int I;
    unsigned char max;

    int x, y; // position of the current pixel

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        /////////////////////////////////////////
        // processing of the top of the image //
        ////////////////////////////////////////

        // clear
        for( I = 0; I < 256*img_width; I++ )
        {
            columns_histo[I] = 0;
        }

        // initialization
        for( x = 0; x < img_width; x++ )
        {
            for( y = 0; y < kernel_length; y++ )
            {
                columns_histo[ 256*x+filtered[3*(x+y*img_width)+color_channel] ]++;
            }
        }

        // downward moving in the image
        for( y = 0; y < border_length+1; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y+border_length+1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y+border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the maximum value in the kernel histogram
                    max = 255;
                    while( kernel_histo[max] == 0 && max != 0 )
                    {
                        max--;
                    }

                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = max;
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length+1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max;
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max;
            }
        }

        //////////////////////////////////////////////////////////////
        // processing of the image without top and bottom stripes  ///
        //////////////////////////////////////////////////////////////

        // clear
        for( I = 0; I < 256*img_width; I++ )
        {
            columns_histo[I] = 0;
        }

        // initialization
        for( x = 0; x < img_width; x++ )
        {
            for( y = 0; y < kernel_length; y++ )
            {
                columns_histo[ 256*x+filtered[3*(x+y*img_width)+color_channel] ]++;
            }
        }

        // downward moving in the image
        for( y = border_length+1; y <= img_height-border_length-2; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y+border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the maximum value in the kernel histogram
                    max = 255;
                    while( kernel_histo[max] == 0 && max != 0 )
                    {
                        max--;
                    }

                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = max;
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++)
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max;
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max;
            }
        }

        ///////////////////////////////////////////
        // processing of the bottom of the image //
        ///////////////////////////////////////////

        // no need to clear and to initialize columns_histo

        for( y = img_height-border_length-1; y < img_height; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y-border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the maximum value in the kernel histogram
                    max = 255;
                    while( kernel_histo[max] == 0 && max != 0 )
                    {
                        max--;
                    }

                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = max;
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max;
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the maximum value in the kernel histogram
                max = 255;
                while( kernel_histo[max] == 0 && max != 0 )
                {
                    max--;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = max;
            }
        }

    }

    if( kernel_length != 1 )
    {
        unsigned char* ptemp = filtered;
        filtered = filtered_modif;
        filtered_modif = ptemp;
    }

    return;
}

void FiltersRGB::erosion(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int border_length = (kernel_length-1)/2;

    int x, y; // position of the current pixel

    unsigned char min;

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        for( int i = 0; i < img_size; i++ )
        {
            y = i/img_width;
            x = i-y*img_width;

            // initialization
            min = 255;

            // if the current pixel is not in the border
            if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        // no neighbors tests
                        if( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] < min )
                        {
                            min = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                        }
                    }
                }
            }
            // if in the border
            else
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        // neighbors tests
                        if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                        {
                            if( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] < min )
                            {
                                min = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                            }
                        }
                    }
                }
            }
            filtered_modif[3*i+color_channel] = min;
        }
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

void FiltersRGB::erosion_o1(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // variable called radius r in the Perreault and Hébert's article
    // border_length = r
    // kernel_length = 2*r+1
    const int border_length = (kernel_length-1)/2;

    int I;
    unsigned char min;

    int x, y; // position of the current pixel

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        /////////////////////////////////////////
        // processing of the top of the image //
        ////////////////////////////////////////

        // clear
        for( I = 0; I < 256*img_width; I++ )
        {
            columns_histo[I] = 0;
        }

        // initialization
        for( x = 0; x < img_width; x++ )
        {
            for( y = 0; y < kernel_length; y++ )
            {
                columns_histo[ 256*x+filtered[3*(x+y*img_width)+color_channel] ]++;
            }
        }

        // downward moving in the image
        for( y = 0; y < border_length+1; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ ) {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y+border_length+1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y+border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the minimum value in the kernel histogram
                    min = 0;
                    while( kernel_histo[min] == 0 && min != 255 )
                    {
                        min++;
                    }

                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = min;
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length+1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I]-columns_histo[256*(x-border_length-1)+I ];
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = min;
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            for( x = img_width-border_length-1; x < img_width; x++)
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++)
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = min;
            }
        }

        //////////////////////////////////////////////////////////////
        // processing of the image without top and bottom stripes  ///
        //////////////////////////////////////////////////////////////

        // clear
        for( I = 0; I < 256*img_width; I++ )
        {
            columns_histo[I] = 0;
        }

        // initialization
        for( x = 0; x < img_width; x++ )
        {
            for( y = 0; y < kernel_length; y++ )
            {
                columns_histo[ 256*x+filtered[3*(x+y*img_width)+color_channel] ]++;
            }
        }

        // downward moving in the image
        for( y = border_length+1; y <= img_height-border_length-2; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y+border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the minimum value in the kernel histogram
                    min = 0;
                    while( kernel_histo[min] == 0 && min != 255 )
                    {
                        min++;
                    }

                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = min;
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = min;
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = min;
            }
        }

        ///////////////////////////////////////////
        // processing of the bottom of the image //
        ///////////////////////////////////////////

        // no need to clear and to initialize columns_histo

        for( y = img_height-border_length-1; y < img_height; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y-border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++)
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the minimum value in the kernel histogram
                    min = 0;
                    while( kernel_histo[min] == 0 && min != 255 )
                    {
                        min++;
                    }

                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = min;
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = min;
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the minimum value in the kernel histogram
                min = 0;
                while( kernel_histo[min] == 0 && min != 255 )
                {
                    min++;
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = min;
            }
        }
    }

    if( kernel_length !=1 )
    {
        unsigned char* ptemp = filtered;
        filtered = filtered_modif;
        filtered_modif = ptemp;
    }

    return;
}

void FiltersRGB::closing(int kernel_length)
{
    dilation(kernel_length);

    // for the top-hat function eventually
    unsigned char* ptemp = filtered_modif;
    filtered_modif = previous_filtered;
    previous_filtered = ptemp;

    erosion(kernel_length);

    return;
}

void FiltersRGB::closing_o1(int kernel_length)
{
    dilation_o1(kernel_length);

    // for the top-hat function eventually
    unsigned char* ptemp = filtered_modif;
    filtered_modif = previous_filtered;
    previous_filtered = ptemp;

    erosion_o1(kernel_length);

    return;
}

void FiltersRGB::opening(int kernel_length)
{
    erosion(kernel_length);

    // for the top-hat function eventually
    unsigned char* ptemp = filtered_modif;
    filtered_modif = previous_filtered;
    previous_filtered = ptemp;

    dilation(kernel_length);

    return;
}

void FiltersRGB::opening_o1(int kernel_length)
{
    erosion_o1(kernel_length);

    // for the top-hat function eventually
    unsigned char* ptemp = filtered_modif;
    filtered_modif = previous_filtered;
    previous_filtered = ptemp;

    dilation_o1(kernel_length);

    return;
}

void FiltersRGB::black_top_hat(int kernel_length)
{
    closing(kernel_length);

    for( int i = 0; i < 3*img_size; i++ )
    {
        if( filtered[i] >= previous_filtered[i] )
        {
            filtered[i] -= previous_filtered[i];
        }
        else
        {
            filtered[i] = 0;
        }
    }

    return;
}

void FiltersRGB::black_top_hat_o1(int kernel_length)
{
    closing_o1(kernel_length);

    for( int i = 0; i < 3*img_size; i++ )
    {
        if( filtered[i] >= previous_filtered[i] )
        {
            filtered[i] -= previous_filtered[i];
        }
        else
        {
            filtered[i] = 0;
        }
    }

    return;
}

void FiltersRGB::white_top_hat(int kernel_length)
{
    opening(kernel_length);

    for( int i = 0; i < 3*img_size; i++ )
    {
        if( previous_filtered[i] >= filtered[i] )
        {
            filtered[i] = previous_filtered[i]-filtered[i];
        }
        else
        {
            filtered[i] = 0;
        }
    }

    return;
}

void FiltersRGB::white_top_hat_o1(int kernel_length)
{
    opening_o1(kernel_length);

    for( int i = 0; i < 3*img_size; i++ )
    {
        if( previous_filtered[i] >= filtered[i] )
        {
            filtered[i] = previous_filtered[i]-filtered[i];
        }
        else
        {
            filtered[i] = 0;
        }
    }

    return;
}

void FiltersRGB::mean_filtering(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    int x, y; // position of the current pixel

    const int border_length = (kernel_length-1)/2;
    int sum;

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        for( int i = 0; i < img_size; i++ )
        {
            y = i/img_width;
            x = i-y*img_width;

            sum = 0;

            // if the current pixel is not in the border
            if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        sum += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }
            }
            // if in the border
            else
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        if( (x+dx >= 0 && x+dx < img_width) && (y+dy >= 0 && y+dy < img_height) )
                        {
                            sum += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                        }
                        if( (x+dx >= 0 && x+dx < img_width) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                        {
                            sum += int( filtered[ 3*((x+dx)+(y-dy)*img_width)+color_channel ] );
                        }
                        if( ( !(x+dx >= 0 && x+dx < img_width) ) && (y+dy >= 0 && y+dy < img_height) )
                        {
                            sum += int( filtered[ 3*((x-dx)+(y+dy)*img_width)+color_channel ] );
                        }
                        if( ( !(x+dx >= 0 && x+dx < img_width) ) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                        {
                            sum += int( filtered[ 3*((x-dx)+(y-dy)*img_width)+color_channel ] );
                        }
                    }
                }
            }
            filtered_modif[3*i+color_channel] = (unsigned char)(sum/(kernel_length*kernel_length));
        }
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

const double* const FiltersRGB::gaussian_kernel(int kernel_length, double sigma)
{
    // to protect the input : kernel_length impair et strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // to protect against /0
    if( sigma < 0.000000001 )
    {
        sigma = 0.000000001;
    }

    const int kernel_size = kernel_length*kernel_length;

    double* const kernel = new double[kernel_size];

    int x;
    int y;
    const int k = (kernel_length-1)/2;

    double sum = 0.0;

    for(int i = 0; i < kernel_size; i++)
    {
        y = i/kernel_length;
        x = i-y*kernel_length;

        kernel[i] = exp( - ((double(y)-double(k))*(double(y)-double(k))+(double(x)-double(k))*(double(x)-double(k))) / (2*sigma*sigma));

        sum += kernel[i];
    }

    // to normalize
    for( int i = 0; i < kernel_size; i++ )
    {
        kernel[i] = kernel[i]/sum;
    }

    return kernel;
}

void FiltersRGB::gaussian_filtering(int kernel_length, double sigma)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    // to protect against /0
    if( sigma < 0.000000001 )
    {
        sigma = 0.000000001;
    }

    const double* const mask = gaussian_kernel(kernel_length, sigma);

    int x, y; // position of the current pixel

    const int border_length = (kernel_length-1)/2;
    double sum;

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        for( int i = 0; i < img_size; i++ )
        {
            y = i/img_width;
            x = i-y*img_width;

            sum = 0.0;

            // if the current pixel is not in the border
            if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }
            }
            // if is in the border
            else
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        if( (x+dx >= 0 && x+dx < img_width) && (y+dy >= 0 && y+dy < img_height) )
                        {
                            sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                        }
                        if( (x+dx >= 0 && x+dx < img_width) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                        {
                            sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ 3*((x+dx)+(y-dy)*img_width)+color_channel ] );
                        }
                        if( ( !(x+dx >= 0 && x+dx < img_width) ) && (y+dy >= 0 && y+dy < img_height) )
                        {
                            sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ 3*((x-dx)+(y+dy)*img_width)+color_channel ] );
                        }
                        if( ( !(x+dx >= 0 && x+dx < img_width) ) && ( !(y+dy >= 0 && y+dy < img_height) ) ) {
                            sum += mask[ (border_length+dx)+(border_length+dy)*kernel_length ]*double( filtered[ 3*((x-dx)+(y-dy)*img_width)+color_channel ] );
                        }
                    }
                }
            }
            filtered_modif[3*i+color_channel] = (unsigned char)my_round(sum);
        }
    }

    delete[] mask;

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

inline void FiltersRGB::swap(unsigned char* const array, int a, int b)
{
    unsigned char temp = array[a];
    array[a] = array[b];
    array[b] = temp;

    return;
}

void FiltersRGB::quick_sort(unsigned char* const array, int begin, int end)
{
    int left = begin-1;
    int right = end+1;
    const unsigned char pivot = array[begin];

    // if the length of the array is 0, to do nothing
    if( begin >= end )
    {
        return;
    }

    while( 1 )
    {
        do
        {
            right--;
        }
        while( array[right] > pivot );

        do
        {
            left++;
        }
        while( array[left] < pivot );

        if( left < right )
        {
            swap(array, left, right);
        }
        else
        {
            break;
        }
    }

    quick_sort(array, begin, right);
    quick_sort(array, right+1, end);

    return;
}

void FiltersRGB::median_filtering_oNlogN(int kernel_length) {

    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0)
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    int x, y; // position of the current pixel

    const int border_length = (kernel_length-1)/2;
    int m;

    const int length = kernel_length*kernel_length;
    unsigned char* const median_kernel = new unsigned char[length];

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        // i, offset
        for( int i = 0; i < img_size; i++ )
        {
            y = i/img_width;
            x = i-y*img_width;

            // m : index of the kernel
            m = 0;

            // if the current pixel is not in the border
            if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        median_kernel[m++] = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                    }
                }
            }
            // if in the border
            else
            {
                for( int dy = -border_length; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= border_length; dx++ )
                    {
                        if( (x+dx >= 0 && x+dx < img_width) && (y+dy >= 0 && y+dy < img_height) )
                        {
                            median_kernel[m++] = filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ];
                        }
                        if( (x+dx >= 0 && x+dx < img_width) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                        {
                            median_kernel[m++] = filtered[ 3*((x+dx)+(y-dy)*img_width)+color_channel ];
                        }
                        if( ( !(x+dx >= 0 && x+dx < img_width) ) && (y+dy >= 0 && y+dy < img_height) )
                        {
                            median_kernel[m++] = filtered[ 3*((x-dx)+(y+dy)*img_width)+color_channel ];
                        }
                        if( ( !(x+dx >= 0 && x+dx < img_width) ) && ( !(y+dy >= 0 && y+dy < img_height) ) )
                        {
                            median_kernel[m++] = filtered[ 3*((x-dx)+(y-dy)*img_width)+color_channel ];
                        }
                    }
                }
            }

            quick_sort(median_kernel,0,length-1);

            filtered_modif[3*i+color_channel] = median_kernel[(length-1)/2];
        }
    }
    delete[] median_kernel;

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;

}

void FiltersRGB::median_filtering_o1(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int median_position = 1+kernel_length*kernel_length/2;

    // variable called radius r in the Perreault and Hébert's article
    // border_length = r
    // kernel_length = 2*r+1
    const int border_length = (kernel_length-1)/2;

    int I;

    int x, y; // position of the current pixel

    int n, m;

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        /////////////////////////////////////////
        // processing of the top of the image //
        ////////////////////////////////////////

        // clear
        for( I = 0; I < 256*img_width; I++ )
        {
            columns_histo[I] = 0;
        }

        // initialization
        for( x = 0; x < img_width; x++ )
        {
            for( y = 0; y < kernel_length; y++ )
            {
                columns_histo[ 256*x+filtered[3*(x+y*img_width)+color_channel] ]++;
            }
        }

        // downward moving in the image
        for( y = 0; y < border_length+1; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y+border_length+1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y+border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the value I of the median in the kernel histogram
                    n = 0;
                    I = -1;
                    while( n <= kernel_length*(x+1)/2 )
                    {
                        n += kernel_histo[++I];
                    }
                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = (unsigned char)(I);
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length+1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++)
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the value I of the median in the kernel histogram
                n = 0;
                I = -1;
                while( n <= (kernel_length*kernel_length-1)/2 )
                {
                    n += kernel_histo[++I];
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = (unsigned char)(I);
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            m = 1;
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the value I of the median in the kernel histogram
                n = 0;
                I = -1;
                while( n <= (kernel_length*(kernel_length-m))/2 )
                {
                    n += kernel_histo[++I];
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = (unsigned char)(I);
                m++;
            }
        }

        //////////////////////////////////////////////////////////////
        // processing of the image without top and bottom stripes  ///
        //////////////////////////////////////////////////////////////

        // clear
        for( I = 0; I < 256*img_width; I++)
        {
            columns_histo[I] = 0;
        }

        // initialization
        for( x = 0; x < img_width; x++ )
        {
            for( y = 0; y < kernel_length; y++ )
            {
                columns_histo[ 256*x+filtered[3*(x+y*img_width)+color_channel] ]++;
            }
        }

        // downward moving in the image
        for( y = border_length+1; y <= img_height-border_length-2; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y+border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the value I of the median in the kernel histogram
                    n = 0;
                    I = -1;
                    while( n <= kernel_length*(x+1)/2 )
                    {
                        n += kernel_histo[++I];
                    }
                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = (unsigned char)(I);
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y+border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++)
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the value I of the median in the kernel histogram
                n = 0;
                I = -1;
                while( n < median_position )
                {
                    n += kernel_histo[++I];
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = (unsigned char)(I);
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            m = 1;
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the value I of the median in the kernel histogram
                n = 0;
                I = -1;
                while( n <= (kernel_length*(kernel_length-m))/2 )
                {
                    n += kernel_histo[++I];
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = (unsigned char)(I);
                m++;
            }
        }

        ///////////////////////////////////////////
        // processing of the bottom of the image //
        ///////////////////////////////////////////

        // no need to clear and to initialize columns_histo

        for( y = img_height-border_length-1; y < img_height; y++ )
        {
            // clear
            for( I = 0; I < 256; I++ )
            {
                kernel_histo[I] = 0;
            }

            // initialization for each new current row
            for( x = 0; x < kernel_length; x++ )
            {
                columns_histo[ 256*x+filtered[3*(x+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*x+filtered[3*(x+(y-border_length)*img_width)+color_channel] ]++;
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*x+I ];
                }

                ///////////////////////////
                // left border processed //
                ///////////////////////////

                if( x >= border_length )
                {
                    // to find the value I of the median in the kernel histogram
                    n = 0;
                    I = -1;
                    while( n <= kernel_length*(x+1)/2 )
                    {
                        n += kernel_histo[++I];
                    }
                    filtered_modif[ 3*((x-border_length)+y*img_width)+color_channel ] = (unsigned char)(I);
                }
            }

            //////////////////////
            // center processed //
            //////////////////////

            // rightward moving in the image
            for( x = border_length+1; x <= img_width-border_length-2; x++ )
            {
                // to update the column histogram H(x+border_length)
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length-1)*img_width)+color_channel] ]--;
                columns_histo[ 256*(x+border_length)+filtered[3*((x+border_length)+(y-border_length)*img_width)+color_channel] ]++;

                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] += columns_histo[ 256*(x+border_length)+I ]-columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the value I of the median in the kernel histogram
                n = 0;
                I = -1;
                while( n < median_position )
                {
                    n += kernel_histo[++I];
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = (unsigned char)(I);
            }

            ////////////////////////////
            // right border processed //
            ////////////////////////////

            // rightward moving in the image
            m = 1;
            for( x = img_width-border_length-1; x < img_width; x++ )
            {
                // to update the mask histogram from the column histograms H(x+border_length) and H(x-border_length-1)
                for( I = 0; I < 256; I++ )
                {
                    kernel_histo[I] -= columns_histo[ 256*(x-border_length-1)+I ];
                }

                // to find the value I of the median in the kernel histogram
                n = 0;
                I = -1;
                while( n <= (kernel_length*(kernel_length-m))/2 )
                {
                    n += kernel_histo[++I];
                }

                filtered_modif[ 3*(x+y*img_width)+color_channel ] = (unsigned char)(I);
                m++;
            }
        }
    }

    if( kernel_length != 1 )
    {
        unsigned char* ptemp = filtered;
        filtered = filtered_modif;
        filtered_modif = ptemp;
    }

    return;
}

void FiltersRGB::gaussian_white_noise(double sigma)
{
    // generator
    boost::mt19937 rng(time(0));
    // distribution
    boost::normal_distribution<> nd(0.0, sigma);

    // link between the generator and the distribution
    boost::variate_generator<boost::mt19937&,
                             boost::normal_distribution<> > var_nor(rng, nd);

    int filtered_int;

    for( int i = 0; i < 3*img_size; i++ )
    {
        filtered_int = int(filtered[i]) + (int)(var_nor());

        if( filtered_int < 0 )
        {
            filtered_int = 0;
        }
        if( filtered_int > 255 )
        {
            filtered_int = 255;
        }
        filtered[i] = (unsigned char)(filtered_int);
    }

    return;
}

void FiltersRGB::salt_pepper_noise(double proba)
{
    // generator
    boost::mt19937 rng(time(0));
    // distribution
    boost::uniform_01<> ud;

    // link between the generator and the distribution
    boost::variate_generator<boost::mt19937&,
                             boost::uniform_01<> > var_uni(rng, ud);

    for( int i = 0; i < img_size; i++ )
    {
        if( var_uni() < proba )
        {
            if( var_uni() < 0.5 )
            {
                filtered[3*i] = 0;
            }
            else
            {
                filtered[3*i] = 255;
            }
            if( var_uni() < 0.5 )
            {
                filtered[3*i+1] = 0;
            }
            else
            {
                filtered[3*i+1] = 255;
            }
            if( var_uni() < 0.5 )
            {
                filtered[3*i+2] = 0;
            }
            else
            {
                filtered[3*i+2] = 255;
            }
        }

    }

    return;
}

void FiltersRGB::speckle(double sigma)
{
    // generator
    boost::mt19937 rng(time(0));
    // distribution
    boost::uniform_01<> ud;

    // link between the generator and the distribution
    boost::variate_generator<boost::mt19937&,
                             boost::uniform_01<> > var_uni(rng, ud);

    int filtered_int;

    for( int i = 0; i < 3*img_size; i++ )
    {
        // img = img + X * img
        // with X a random variable from a uniform distribution centered at 0
        // and of a standard deviation sigma
        // sigma * sqrt(12) is the interval of the uniform distribution
        filtered_int = int(filtered[i]) + int (double(filtered[i]) * sigma * sqrt(12.0) * (var_uni()-0.5));

        if( filtered_int < 0 )
        {
            filtered_int = 0;
        }
        if(filtered_int > 255 )
        {
            filtered_int = 255;
        }
        filtered[i] = (unsigned char)(filtered_int);
    }

    return;
}

void FiltersRGB::local_binary_pattern(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int border_length = (kernel_length-1)/2;

    int x, y; // position of the current pixel

    unsigned char Icentr;

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        for( int i = 0; i < img_size; i++ )
        {
            y = i/img_width;
            x = i-y*img_width;

            Icentr = filtered[3*i+color_channel];

            filtered_modif[3*i+color_channel] = 0;

            if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
            {
                if( filtered[ 3*((x-border_length)+(y-border_length)*img_width)+color_channel ] >= Icentr )
                {
                    filtered_modif[ 3*i+color_channel ] += 1;
                }
                if( filtered[ 3*(x+(y-border_length)*img_width)+color_channel ] >= Icentr )
                {
                    filtered_modif[ 3*i+color_channel ] += 2;
                }
                if( filtered[ 3*((x+border_length)+(y-border_length)*img_width)+color_channel] >= Icentr )
                {
                    filtered_modif[ 3*i+color_channel ] += 4;
                }
                if( filtered[ 3*((x-border_length)+y*img_width)+color_channel ] >= Icentr )
                {
                    filtered_modif[ 3*i+color_channel ] += 8;
                }
                if( filtered[ 3*((x+border_length)+y*img_width)+color_channel ] >= Icentr )
                {
                    filtered_modif[ 3*i+color_channel ] += 16;
                }
                if( filtered[ 3*((x-border_length)+(y+border_length)*img_width)+color_channel ] >= Icentr )
                {
                    filtered_modif[ 3*i+color_channel ] += 32;
                }
                if( filtered[ 3*(x+(y+border_length)*img_width)+color_channel ] >= Icentr )
                {
                    filtered_modif[ 3*i+color_channel ] += 64;
                }
                if( filtered[ 3*((x+border_length)+(y+border_length)*img_width)+color_channel ] >= Icentr )
                {
                    filtered_modif[ 3*i+color_channel ] += 128;
                }
            }
        }
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}


void FiltersRGB::nagao_filtering(int kernel_length)
{
    // to protect the input : kernel_length impair and strictly positive
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    int x, y; // position of the current pixel

    const int border_length = (kernel_length-1)/2;

    int var[9];
    int sum[9];

    int min_var;

    int nagao_size_init = square(border_length+1);
    if( nagao_size_init <= 0 ) {
        nagao_size_init = 1;
    }
    const int nagao_size = nagao_size_init;

    for( int color_channel = 0; color_channel < 3; color_channel++ )
    {
        for( int i = 0; i < img_size; i++ )
        {
            y = i/img_width;
            x = i-y*img_width;

            for( int n = 0; n < 9; n++ )
            {
                sum[n] = 0;
                var[n] = 0;
            }

            // if not in the border
            if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length)
            {
                for( int dy = 0; dy >= -border_length; dy-- )
                {
                    for( int dx = dy; dx <= -dy; dx++ )
                    {
                        sum[0] += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }
                for( int dx = 0; dx <= border_length; dx++ )
                {
                    for( int dy = -dx; dy <= dx; dy++ )
                    {
                        sum[1] += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }
                for( int dy = 0; dy <= border_length; dy++ )
                {
                    for( int dx = -dy; dx <= dy; dx++ )
                    {
                        sum[2] += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }
                for( int dx = 0; dx >= -border_length; dx-- )
                {
                    for( int dy = dx; dy <= -dx; dy++ )
                    {
                        sum[3] += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }

                for( int dy = -border_length; dy <= 0; dy++ )
                {
                    for( int dx = 0; dx <= border_length; dx++ )
                    {
                        sum[4] += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }
                for( int dy = 0; dy <= border_length; dy++ )
                {
                    for( int dx = 0; dx <= border_length; dx++ )
                    {
                        sum[5] += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }
                for( int dy = 0; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= 0; dx++ )
                    {
                        sum[6] += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }
                for( int dy = -border_length; dy <= 0; dy++ )
                {
                    for( int dx = -border_length; dx <= 0; dx++ )
                    {
                        sum[7] += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }

                for( int dy = -border_length+1; dy <= border_length-1; dy++ )
                {
                    for( int dx = -border_length+1; dx <= border_length-1; dx++ )
                    {
                        sum[8] += int( filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] );
                    }
                }
            }

            // if not in the border
            if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
            {
                for(int dy = 0; dy >= -border_length; dy-- )
                {
                    for( int dx = dy; dx <= -dy; dx++ )
                    {
                        var[0] += square( sum[0]-int(filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ]) );
                    }
                }
                for( int dx = 0; dx <= border_length; dx++ )
                {
                    for( int dy = -dx; dy <= dx; dy++ )
                    {
                        var[1] += square( sum[1]-int(filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ]) );
                    }
                }
                for( int dy = 0; dy <= border_length; dy++ )
                {
                    for(int dx = -dy; dx <= dy; dx++ )
                    {
                        var[2] += square( sum[2]-int(filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] ) );
                    }
                }
                for( int dx = 0; dx >= -border_length; dx-- )
                {
                    for( int dy = dx; dy <= -dx; dy++ )
                    {
                        var[3] += square( sum[3]-int(filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ] ) );
                    }
                }

                for( int dy = -border_length; dy <= 0; dy++ )
                {
                    for( int dx = 0; dx <= border_length; dx++ )
                    {
                        var[4] += square( sum[4]-int(filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ]) );
                    }
                }
                for( int dy = 0; dy <= border_length; dy++ )
                {
                    for( int dx = 0; dx <= border_length; dx++ )
                    {
                        var[5] += square( sum[5]-int(filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ]) );
                    }
                }
                for( int dy = 0; dy <= border_length; dy++ )
                {
                    for( int dx = -border_length; dx <= 0; dx++ )
                    {
                        var[6] += square( sum[6]-int(filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ]) );
                    }
                }
                for( int dy = -border_length; dy <= 0; dy++ )
                {
                    for( int dx = -border_length; dx <= 0; dx++ )
                    {
                        var[7] += square( sum[7]-int(filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ]) );
                    }
                }

                for( int dy = -border_length+1; dy <= border_length-1; dy++ )
                {
                    for( int dx = -border_length+1; dx <= border_length-1; dx++ )
                    {
                        var[8] += square(sum[8]-int(filtered[ 3*((x+dx)+(y+dy)*img_width)+color_channel ]));
                    }
                }
            }

            min_var = 999999999;
            for( int n = 0; n < 9; n++ )
            {
                if( var[n] < min_var )
                {
                    min_var = var[n];
                }
            }

            for( int n = 0; n < 9; n++ )
            {
                if( var[n] == min_var )
                {
                    filtered_modif[3*i+color_channel] = (unsigned char)(sum[n]/nagao_size);
                }
            }
        }
    }

    unsigned char* ptemp = filtered;
    filtered = filtered_modif;
    filtered_modif = ptemp;

    return;
}

//! Gradient morphologique avec un élement structurant carré de taille kernel_length
void FiltersRGB::morphological_gradient_yuv(int kernel_length, int alpha, int beta, int gamma)
{
    // pour blinder l'entrée : kernel_length impair et strictement positif
    if( kernel_length % 2 == 0 )
    {
        kernel_length--;
    }
    if( kernel_length < 1 )
    {
        kernel_length = 1;
    }

    const int border_length = (kernel_length-1)/2;

    int R, G, B;
    int Y, U, V;

    int x, y; // position of the current pixel

    int maxY, minY, maxU, minU, maxV, minV;

    for( int i = 0; i < img_size; i++ )
    {
        y = i/img_width;
        x = i-y*img_width;

        // initialisation
        maxY = 0;
        minY = 255;
        maxU = 0;
        minU = 255;
        maxV = 0;
        minV = 255;

        // si pas sur le bord
        if( x > border_length-1 && x < img_width-border_length && y > border_length-1 && y < img_height-border_length )
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    R = int(filtered[3*((x+dx)+(y+dy)*img_width)]);
                    G = int(filtered[3*((x+dx)+(y+dy)*img_width)+1]);
                    B = int(filtered[3*((x+dx)+(y+dy)*img_width)+2]);

                    Y = ( (  66 * R + 129 * G +  25 * B + 128) >> 8) +  16;
                    U = ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
                    V = ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;

                    // pas de tests des voisins
                    if( Y > maxY )
                    {
                        maxY = Y;
                    }
                    if( Y < minY )
                    {
                        minY = Y;
                    }

                    if( U > maxU )
                    {
                        maxU = U;
                    }
                    if( U < minU )
                    {
                        minU = U;
                    }

                    if( V > maxV )
                    {
                        maxV = V;
                    }
                    if( V < minV )
                    {
                        minV = V;
                    }
                }
            }
        }
        // si sur le bord
        else
        {
            for( int dy = -border_length; dy <= border_length; dy++ )
            {
                for( int dx = -border_length; dx <= border_length; dx++ )
                {
                    // tests des voisins
                    if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                    {
                        R = int(filtered[3*((x+dx)+(y+dy)*img_width)]);
                        G = int(filtered[3*((x+dx)+(y+dy)*img_width)+1]);
                        B = int(filtered[3*((x+dx)+(y+dy)*img_width)+2]);

                        Y = ( (  66 * R + 129 * G +  25 * B + 128) >> 8) +  16;
                        U = ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
                        V = ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;

                        // pas de tests des voisins
                        if( Y > maxY )
                        {
                            maxY = Y;
                        }
                        if( Y < minY )
                        {
                            minY = Y;
                        }

                        if( U > maxU )
                        {
                            maxU = U;
                        }
                        if( U < minU )
                        {
                            minU = U;
                        }

                        if( V > maxV )
                        {
                            maxV = V;
                        }
                        if( V < minV )
                        {
                            minV = V;
                        }
                    }
                }
            }
        }
        gradient[i] = (unsigned char)((alpha*(maxY-minY)+beta*(maxU-minU)+gamma*(maxV-minV))/(alpha+beta+gamma));
    }

    return;
}

void FiltersRGB::fas(int kernel_length)
{
    for( int i = 3; i < kernel_length; i += 2 )
    {
        opening(i);
        closing(i);
    }
}

int FiltersRGB::square(int value)
{
    return value*value;
}

const unsigned char* const FiltersRGB::get_filtered() const
{
    return filtered;
}

const unsigned char* const FiltersRGB::get_gradient() const
{
    return gradient;
}

}
