﻿/****************************************************************************
**
** Copyright (C) 2010-2011 Fabien Bessy.
** All rights reserved.
** Contact: fabien.bessy@gmail.com
**
** This file is part of Ofeli.
**
** You may use this file under the terms of the CeCILL license version 2 as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Fabien Bessy nor the names of its contributors
**     may be used to endorse or promote products derived from this
**     software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
**                    For all legal intents and purposes.
**
****************************************************************************/


#include "geodesic_ac.hpp"
#include <iostream> // for the object "std::cerr"

namespace ofeli
{

GeodesicAC::GeodesicAC(const unsigned char* img_gradient_data1, int img_width1, int img_height1) : ActiveContour(img_width1, img_height1, false, 0.9, 0.9, 0.0, 0.0, true, 5, 2.0, true, 30, 3), img_gradient_data(img_gradient_data1), otsu_threshold1(compute_threshold_with_Otsu_method(img_gradient_data1, img_width1*img_height1)), isInside(find_direction_evolution())
{
    if( img_gradient_data1 == NULL )
    {
        std::cerr << std::endl << "The pointer img_gradient_data1 must be a non-null pointer, it must be allocated." << std::endl;
    }
}

GeodesicAC::GeodesicAC(const unsigned char* img_gradient_data1, int img_width1, int img_height1, bool hasEllipse1, double init_width1, double init_height1, double center_x1, double center_y1, bool hasSmoothingCycle1, int kernel_length1, double sigma1,bool hasCleaning1, int Na1, int Ns1) : ActiveContour(img_width1, img_height1, hasEllipse1, init_width1, init_height1, center_x1, center_y1, hasSmoothingCycle1, kernel_length1, sigma1, hasCleaning1, Na1, Ns1), img_gradient_data(img_gradient_data1), otsu_threshold1(compute_threshold_with_Otsu_method(img_gradient_data1, img_width1*img_height1)), isInside(find_direction_evolution())
{
    if( img_gradient_data1 == NULL )
    {
        std::cerr << std::endl << "The pointer img_gradient_data1 must be a non-null pointer, it must be allocated." << std::endl;
    }
}

GeodesicAC::GeodesicAC(const unsigned char* img_gradient_data1, int img_width1, int img_height1, const char* phi_init1, bool hasSmoothingCycle1, int kernel_length1, double sigma1, bool hasCleaning1, int Na1, int Ns1) : ActiveContour(img_width1, img_height1, phi_init1, hasSmoothingCycle1, kernel_length1, sigma1, hasCleaning1, Na1, Ns1), img_gradient_data(img_gradient_data1), otsu_threshold1(compute_threshold_with_Otsu_method(img_gradient_data1, img_width1*img_height1)), isInside(find_direction_evolution())
{
    if( img_gradient_data1 == NULL )
    {
        std::cerr << std::endl << "The pointer img_gradient_data1 must be a non-null pointer, it must be allocated." << std::endl;
    }
}

inline int GeodesicAC::compute_external_speed_Fd(int i, int x, int y)
{
    int gradient = int(img_gradient_data[i]);
    int sum_local_Lout, n_local_Lout, sum_local_Lin, n_local_Lin;

    if( isInside )
    {
        if( gradient > 0.7*otsu_threshold1 )
        {
            sum_local_Lout = 0;
            n_local_Lout = 0;
            sum_local_Lin = 0;
            n_local_Lin = 0;

            // if not in the border of the image, no neighbor test
            if( x > 0 && x < img_width-1 && y > 0 && y < img_height-1 )
            {
                // scan through a 3Ã—3 neighborhood
                for( int dy = -1; dy <= 1; dy++ )
                {
                    for( int dx = -1; dx <= 1; dx++ )
                    {
                        if( phi[ (x+dx)+(y+dy)*img_width ] == 1 )
                        {
                            sum_local_Lout += int( img_gradient_data[ (x+dx)+(y+dy)*img_width ] );
                            n_local_Lout++;
                        }
                        if( phi [ (x+dx)+(y+dy)*img_width ] == -1 )
                        {
                            sum_local_Lin += int( img_gradient_data[ (x+dx)+(y+dy)*img_width ] );
                            n_local_Lin++;
                        }
                    }
                }
            }
            // if in the border, neighbors tests
            else
            {
                // scan through a 3Ã—3 neighborhood
                for( int dy = -1; dy <= 1; dy++ )
                {
                    for( int dx = -1; dx <= 1; dx++ )
                    {
                        // existence tests
                        if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                        {
                            if( phi[ (x+dx)+(y+dy)*img_width ] == 1 )
                            {
                                sum_local_Lout += int( img_gradient_data[ (x+dx)+(y+dy)*img_width ] );
                                n_local_Lout++;
                            }
                            if( phi [ (x+dx)+(y+dy)*img_width ] == -1 )
                            {
                                sum_local_Lin += int( img_gradient_data[ (x+dx)+(y+dy)*img_width ] );
                                n_local_Lin++;
                            }
                        }
                    }
                }
            }

            if( n_local_Lout != 0 && n_local_Lin != 0 )
            {
                if( sum_local_Lout/n_local_Lout-sum_local_Lin/n_local_Lin >= 0 )
                {
                    return 1;
                }
            }
        }

        return -1;
    }

    else
    {
        if( gradient > 0.7*otsu_threshold1 )
        {
            sum_local_Lout = 0;
            n_local_Lout = 0;
            sum_local_Lin = 0;
            n_local_Lin = 0;

            // if not in the border of the image, no neighbor test
            if( x > 0 && x < img_width-1 && y > 0 && y < img_height-1 )
            {
                // scan through a 3Ã—3 neighborhood
                for( int dy = -1; dy <= 1; dy++ )
                {
                    for( int dx = -1; dx <= 1; dx++ )
                    {
                        if( phi[ (x+dx)+(y+dy)*img_width ] == 1 )
                        {
                            sum_local_Lout += int( img_gradient_data[ (x+dx)+(y+dy)*img_width ] );
                            n_local_Lout++;
                        }
                        if( phi[ (x+dx)+(y+dy)*img_width ] == -1 )
                        {
                            sum_local_Lin += int( img_gradient_data[ (x+dx)+(y+dy)*img_width ] );
                            n_local_Lin++;
                        }
                    }
                }
            }

            // if in the border, neighbors tests
            else
            {
                // scan through a 3Ã—3 neighborhood
                for( int dy = -1; dy <= 1; dy++ )
                {
                    for( int dx = -1; dx <= 1; dx++ )
                    {
                        // existence tests
                        if( x+dx >= 0 && x+dx < img_width && y+dy >= 0 && y+dy < img_height )
                        {
                            if( phi[ (x+dx)+(y+dy)*img_width ] == 1 )
                            {
                                sum_local_Lout += int( img_gradient_data[ (x+dx)+(y+dy)*img_width ] );
                                n_local_Lout++;
                            }
                            if( phi[ (x+dx)+(y+dy)*img_width ] == -1 )
                            {
                                sum_local_Lin += int( img_gradient_data[ (x+dx)+(y+dy)*img_width ] );
                                n_local_Lin++;
                            }
                        }
                    }
                }
            }

            if( n_local_Lout != 0 && n_local_Lin != 0 )
            {
                if( sum_local_Lout/n_local_Lout-sum_local_Lin/n_local_Lin <= 0 )
                {
                    return -1;
                }
            }
        }

        return 1;
    }

}

// to find the strong gradient values and to converge
int GeodesicAC::compute_threshold_with_Otsu_method(const unsigned char* img_data1, int img_size1)
{
    int histogram[256];

    for( int x = 0; x < 256; x++ )
    {
        histogram[x] = 0;
    }

    for( int i = 0; i < img_size1; i++ )
    {
        for( int x = 0; x < 256; x++ )
        {
            if( int(img_data1[i]) == x )
            {
                histogram[x]++;
            }
        }
    }

    int size1, size2;

    double mean1, mean2, maximum;
    double interclass_variance[255];

    // S va de 1 Ã  255

    for( int S = 1; S < 256; S++ )
    {
        size1 = 0;
        size2 = 0;

        mean1 = 0.0;
        mean2 = 0.0;

        for( int x = 0; x < S; x++ )
        {
            size1 += histogram[x];
        }
        for( int x = S; x < 256; x++ )
        {
            size2 += histogram[x];
        }

        for( int x = 0; x < S; x++ )
        {
            if( size1 != 0 )
            {
                mean1 += double(x)*double(histogram[x])/double(size1);
            }
        }

        for( int x = S; x < 256; x++ )
        {
            if( size2 != 0 )
            {
                mean2 += double(x)*double(histogram[x])/double(size2);
            }
        }
        interclass_variance[S-1] = double(size1)*double(size2)*(mean1-mean2)*(mean1-mean2);

    }

    maximum = -1.0;

    for( int x = 0; x < 255; x++ )
    {
        if( interclass_variance[x] > maximum )
        {
            maximum = interclass_variance[x];
        }
    }

    int otsu_threshold;

    for( int x = 0; x < 255; x++ )
    {
        if( interclass_variance[x] == maximum )
        {
            otsu_threshold = x+1;
        }
    }

    return otsu_threshold;
}

inline bool GeodesicAC::find_direction_evolution()
{
    int sum_in = 0;
    int sum_out = 0;

    for( int i = 0; i < img_size; i++ )
    {
        if( phi[i] > 0 )
        {
            sum_out += int(img_gradient_data[i]);
        }
        else
        {
            sum_in += int(img_gradient_data[i]);
        }
    }

    if( sum_in > sum_out )
    {
        return true;
    }
    else
    {
        return false;
    }
}

unsigned char GeodesicAC::get_otsu_threshold() const
{
    return otsu_threshold1;
}

// initilization for video tracking
void GeodesicAC::initialize_for_each_frame(const unsigned char* img_gradient_data1)
{
    img_gradient_data = img_gradient_data1;
    do_general_initialization_for_each_frame();

    return;
}

}
