﻿/****************************************************************************
**
** Copyright (C) 2010-2011 Fabien Bessy.
** All rights reserved.
** Contact: fabien.bessy@gmail.com
**
** This file is part of Ofeli.
**
** You may use this file under the terms of the CeCILL license version 2 as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Fabien Bessy nor the names of its contributors
**     may be used to endorse or promote products derived from this
**     software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
**                    For all legal intents and purposes.
**
****************************************************************************/


#include "hausdorff_distance.hpp"
#include <cmath> // for the function "sqrt"
#include <iostream> // for the object "std::cerr"

namespace ofeli
{

HausdorffDistance::HausdorffDistance(const int* list1_1, int list1_1_length, int img1_1_width, int img1_1_height, const int* list2_1, int list2_1_length, int img2_1_width, int img2_1_height) : list1(list1_1), list1_length(list1_1_length), img1_width(img1_1_width), img1_height(img1_1_height), list2(list2_1), list2_length(list2_1_length), img2_width(img2_1_width), img2_height(img2_1_height), max_dist(calculate_max_dist(img1_1_width,img1_1_height,img2_1_width,img2_1_height))
{
    if( list1_1 == NULL )
    {
        std::cerr << std::endl << "The pointer list1_1 must be a non-null pointer, it must be allocated." << std::endl;
    }

    if( list2_1 == NULL )
    {
        std::cerr << std::endl << "The pointer list2_1 must be a non-null pointer, it must be allocated." << std::endl;
    }

    // initialization
    hausdorff_dist = -1.0;
    hausdorff_ratio = -1.0;
    modified_hausdorff_dist = -1.0;
    modified_hausdorff_ratio = -1.0;

    ////////////////  it calculates centers  ////////////////////

    int x1, y1;
    int sum_x1 = 0;
    int sum_y1 = 0;

    for( int i1 = 0; i1 < list1_length; i1++ )
    {
        y1 = list1[i1]/img1_width;
        x1 = list1[i1]-y1*img1_width;

        sum_x1 += x1;
        sum_y1 += y1;
    }

    // center of list 1
    if( list1_length != 0 )
    {
        mean_x1 = double(sum_x1)/double(list1_length);
        mean_y1 = double(sum_y1)/double(list1_length);
    }

    int x2, y2;
    int sum_x2 = 0;
    int sum_y2 = 0;

    for( int i2 = 0; i2 < list2_length; i2++ )
    {
        y2 = list2[i2]/img2_width;
        x2 = list2[i2]-y2*img2_width;

        sum_x2 += x2;
        sum_y2 += y2;
    }

    // center of list 2
    if( list2_length != 0 )
    {
        mean_x2 = double(sum_x2)/double(list2_length);
        mean_y2 = double(sum_y2)/double(list2_length);
    }

    ////////////////////////////////////////////////////////////////////

    // it computes centers_dist and centers_ratio

    if( ( list1_length != 0 ) && ( list2_length != 0 ) )
    {
        centers_dist = sqrt( (mean_x2-mean_x1)*(mean_x2-mean_x1) + (mean_y2-mean_y1)*(mean_y2-mean_y1) );
        centers_ratio = centers_dist/max_dist;
    }
    else
    {
        centers_dist = -1.0;
        centers_ratio = -1.0;
    }

    ////////////////////////////////////////////////////////////////////

    compute_hausdorff_dist();

}

void HausdorffDistance::compute_hausdorff_dist()
{

    if( ( list1_length != 0 ) && ( list2_length != 0 ) )
    {
        int x1, y1, x2, y2;
        double x1centered, y1centered, x2centered, y2centered;
        double euclidean_dist, sum_min_dist, min_dist;
        double hausdorff_dist1, modified_hausdorff_dist1;

        sum_min_dist = 0.0;

        // initialization in order to maximize
        hausdorff_dist1 = 0.0;

        // for each point of list 1
        for( int i1 = 0; i1 < list1_length; i1++ )
        {
            y1 = list1[i1]/img1_width;
            x1 = list1[i1]-y1*img1_width;

            // moves center of list 1
            x1centered = double(x1)-mean_x1;
            y1centered = double(y1)-mean_y1;

            // initialization in order to minimize
            min_dist = 999999.9;

            // computes the distance of all points of list 2
            for( int i2 = 0; i2 < list2_length; i2++ )
            {
                y2 = list2[i2]/img2_width;
                x2 = list2[i2]-y2*img2_width;

                // moves center of list 2
                x2centered = double(x2)-mean_x2;
                y2centered = double(y2)-mean_y2;


                // computes the euclidian distance
                euclidean_dist = sqrt( (x2centered-x1centered)*(x2centered-x1centered) + (y2centered-y1centered)*(y2centered-y1centered) );

                // minimizes of min_dist
                if( euclidean_dist < min_dist )
                {
                    min_dist = euclidean_dist;
                }
            }

            // maximizes of hausdorff_dist1
            if( min_dist > hausdorff_dist1 )
            {
                hausdorff_dist1 = min_dist;
            }

            sum_min_dist += min_dist;

        }

        // modified hausdorff distance = mean of min_dist
        modified_hausdorff_dist1 = sum_min_dist/list1_length;



        // backward distance


        double hausdorff_dist2, modified_hausdorff_dist2;

        sum_min_dist = 0.0;

        // initialization in order to maximize
        hausdorff_dist2 = 0.0;

        // for each point of list 1
        for( int i2 = 0; i2 < list2_length; i2++ )
        {
            y2 = list2[i2]/img2_width;
            x2 = list2[i2]-y2*img2_width;

            // moves center of list 2
            x2centered = double(x2)-mean_x2;
            y2centered = double(y2)-mean_y2;

            // initialization in order to minimize
            min_dist = 999999.9;

            for( int i1 = 0; i1 < list1_length; i1++ )
            {
                y1 = list1[i1]/img1_width;
                x1 = list1[i1]-y1*img1_width;

                // moves center of list 1
                x1centered = double(x1)-mean_x1;
                y1centered = double(y1)-mean_y1;


                // computes the euclidian distance
                euclidean_dist = sqrt ( (x2centered-x1centered)*(x2centered-x1centered) + (y2centered-y1centered)*(y2centered-y1centered) );

                // minimizes min_dist
                if( euclidean_dist < min_dist )
                {
                    min_dist = euclidean_dist;
                }
            }

            // maximizes hausdorff_dist2
            if( min_dist > hausdorff_dist2 )
            {
                hausdorff_dist2 = min_dist;
            }

            sum_min_dist += min_dist;
        }

        // modified hausdorff distance = mean of min_dist
        modified_hausdorff_dist2 = sum_min_dist/list2_length;


        // max of the distance and the backward distance
        if( hausdorff_dist1 >= hausdorff_dist2 )
        {
            hausdorff_dist = hausdorff_dist1;
        }
        else
        {
            hausdorff_dist = hausdorff_dist2;
        }

        // max of the distance and the backward distance
        if( modified_hausdorff_dist1 >= modified_hausdorff_dist2 )
        {
            modified_hausdorff_dist = modified_hausdorff_dist1;
        }
        else
        {
            modified_hausdorff_dist = modified_hausdorff_dist2;
        }

        hausdorff_ratio = hausdorff_dist/max_dist;
        modified_hausdorff_ratio = modified_hausdorff_dist/max_dist;
    }

    return;
}

double HausdorffDistance::calculate_max_dist(int width1, int height1, int width2, int height2)
{
    int max_width, max_height;

    if( width1 >= width2 )
    {
        max_width = width1;
    }
    else
    {
        max_width = width2;
    }

    if( height1 >= height2 )
    {
        max_height = height1;
    }
    else
    {
        max_height = height2;
    }

    return sqrt( double(max_width*max_width+max_height*max_height) );
}

double HausdorffDistance::get_hausdorff_dist() const
{
    return hausdorff_dist;
}

double HausdorffDistance::get_modified_hausdorff_dist() const
{
    return modified_hausdorff_dist;
}

double HausdorffDistance::get_hausdorff_ratio() const
{
    return hausdorff_ratio;
}

double HausdorffDistance::get_modified_hausdorff_ratio() const
{
    return modified_hausdorff_ratio;
}

double HausdorffDistance::get_centers_dist() const
{
    return centers_dist;
}

double HausdorffDistance::get_centers_ratio() const
{
    return centers_ratio;
}

}
