﻿/****************************************************************************
**
**
****************************************************************************/

#include "hybrid.hpp"
#include <iostream> // for the object "std::cerr"


//TEMP
#include "opencv2/highgui/highgui.hpp"

namespace ofeli
{

Hybrid::Hybrid(const unsigned char* img_data1, int img_width1, int img_height1) 
  : ActiveContour(img_width1, img_height1, false, 0.9, 0.9, 0.0, 0.0, true, 5, 2.0, true, 30, 3), 
    img_data(img_data1), lambda_out(1), lambda_in(1)
{
  initialize_sums();
  calculate_means();
}

Hybrid::Hybrid(const unsigned char* img_data1, int img_width1, int img_height1, bool hasEllipse1, double init_width1, 
                               double init_height1, double center_x1, double center_y1, bool hasSmoothingCycle1, int kernel_length1, 
                               double sigma1,bool hasCleaning1, int Na1, int Ns1, int lambda_out1, int lambda_in1) 
  : ActiveContour(img_width1, img_height1, hasEllipse1, init_width1, init_height1, center_x1, center_y1, hasSmoothingCycle1, kernel_length1, 
                  sigma1, hasCleaning1, Na1, Ns1), 
    img_data(img_data1), lambda_out(lambda_out1), lambda_in(lambda_in1)
{
  initialize_sums();
  calculate_means();
}


  // All terms enabled
Hybrid::Hybrid(const unsigned char* img_data1, int img_width1, int img_height1, const char* phi_init1, bool hasSmoothingCycle1, 
               int kernel_length1, double sigma1, bool hasCleaning1, int Na1, int Ns1, int lambda_out1, int lambda_in1,
               const unsigned char* prior_img_data1,
               double weightCV1, double weightPrior1, double weightVFC1, double gammaVFC, dmat vfc_x1, dmat vfc_y1, bool verbose) 
  : ActiveContour(img_width1, img_height1, phi_init1, hasSmoothingCycle1, kernel_length1, sigma1, hasCleaning1, Na1, Ns1), 
    img_data(img_data1), lambda_out(lambda_out1), lambda_in(lambda_in1), 
    prior_img_data(prior_img_data1), weightCV(weightCV1), weightPrior(weightPrior1), weightVFC(weightVFC1),
    vfc_x(vfc_x1), vfc_y(vfc_y1)
{
  eliminate_redundant_in_Lout(); // clean Lout boundary
  eliminate_redundant_in_Lin(); // clean Lin boundary
  
  initialize_sums();
  calculate_means();
}

int Hybrid::compute_external_speed_Fd(int i, int x, int y)
{
  int cv = compute_external_speed_Fd_Chan_Vese(i, x, y);
  int prior = compute_external_speed_Fd_Prior(i, x, y);
  double vfc = compute_external_speed_Fd_VFC(i, x, y);

  // for TESTING
  // std::cout << "CV: " << cv << "\t weight: " << weightCV << "\t total: " << weightCV * cv << std::endl;
  // std::cout << "Prior: " << prior << "\t weight: " << weightPrior << "\t total: " << weightPrior * prior << std::endl;
  // std::cout << "VFC: " << vfc << "\t weight: " << weightVFC << "\t total: " << weightVFC * vfc << std::endl;

  // double total =  weightCV * cv / 65535.0 +
  //   weightPrior * prior / 65535.0 + 
  //   weightVFC * vfc;

  // std::cout << "Total term: " << total << std::endl;
  // std::cout << "Sing(total): " << sign(total) << std::endl;

  return sign( weightCV * cv / 65535.0 +
               weightPrior * prior / 65535.0 + 
               weightVFC * vfc );
}

int Hybrid::compute_external_speed_Fd_Chan_Vese(int i, int x, int y)
{
    // Chan & Vese
  
    // intensity of the current pixel
    I = int(img_data[i]);

    return lambda_out*square(I-Cout) - lambda_in*square(I-Cin);
}

int Hybrid::compute_external_speed_Fd_Prior(int i, int x, int y)
{
    // Chan & Vese
  
    // intensity of the current pixel
    // Careful: P is a class variable and it is used elsewhere
    P = int(prior_img_data[i]);

    return prior_lambda_out*square(P-prior_Cout) - prior_lambda_in*square(P-prior_Cin);
}

double Hybrid::compute_external_speed_Fd_VFC(int i, int x, int y)
{
  // vfc
  double vfc;
  cv::Vec2f grad;

  if (null_neighbor_norms(i))
  {
    vfc = -1;
    // std::cout << "null_neighbor" << std::endl;
  }
  else
  {
    grad = evaluate_gradient(i);
    double norm = sqrt(double(square(grad[0]) + square(grad[1])));
    if (norm != 0)
    {
      grad[0] /= norm;
      grad[1] /= norm;
    }
    else
      return -1;

    vfc = vfc_x[y][x] * grad[0] + vfc_y[y][x] * grad[1];

  //   if ( i % 1 == 0){
  // cv::Point dest = cv::Point(x + 5*grad[0], y + 5*grad[1]);
  // cv::circle(show_gradient, cv::Point(x,y), 0, CV_RGB(255,0,0));
  // // cv::circle(show_gradient, dest, 1, CV_RGB(255,0,0));
  // // cv::line(show_gradient, cv::Point(x,y), dest, CV_RGB(255, 0, 0), 1, 8, 0 );
  //   }
  }

  // std::cout << "i: " << i << std::endl;
  // std::cout << "g[0]: " << grad[0] << std::endl;
  // std::cout << "g[1]: " << grad[1] << std::endl;
  // std::cout << "vfc[0]: " << vfc_x[y][x] << std::endl;
  // std::cout << "vfc[1]: " << vfc_y[y][x] << std::endl;
  // std::cout << "VFC term: " << vfc << std::endl;
  // std::cout << std::endl;

  return vfc;
}

bool Hybrid::null_neighbor_norms(int offset)
{
  cv::Vec2i grad;

  grad = evaluate_gradient(offset + 1);
  if (grad[0] == 0 && grad[1] == 0)
    return true;

  grad = evaluate_gradient(offset - 1);
  if (grad[0] == 0 && grad[1] == 0)
    return true;

  grad = evaluate_gradient(offset + img_width);
  if (grad[0] == 0 && grad[1] == 0)
    return true;

  grad = evaluate_gradient(offset - img_width);
  if (grad[0] == 0 && grad[1] == 0)
    return true;

  return false;
}

cv::Vec2i Hybrid::evaluate_gradient(const int offset ) const
{
	cv::Vec2i grad;
        int xa, xb, ya, yb, valueA, valueB;

	if (offset >= 0 && offset < img_size)
	{
		xa = offset + 1;
		xb = offset - 1;
		if (xa % img_width == 0)
			xa = offset;
		if (xb % img_width == img_width - 1)
			xb = offset;
		valueA = phi[xa];
		valueB = phi[xb];
		grad[0] = valueA - valueB;

		ya = offset + img_width;
		if (ya >= img_size)
			ya = offset;
		yb = offset - img_width;
		if (yb < 0)
			yb = offset;
		valueA = phi[ya];
		valueB = phi[yb];
		grad[1] = valueA - valueB;
		//grad[1] = valueB - valueA; // in an image buffer UP and DOWN are inverted, so B - A and not A - B
	}
	else
	{
		grad[0] = 0;
		grad[1] = 0;
	}

  return grad;
}

  
void Hybrid::initialize_sums()
{
    sum_in = 0;
    sum_out = 0;
    n_in = 0;
    n_out = 0;

    for( int i = 0; i < img_size; i++ )
    {
        if( phi[i] > 0 )
        {
            sum_out += int(img_data[i]);
            n_out++;
        }
        else
        {
            sum_in += int(img_data[i]);
            n_in++;
        }
    }

    prior_sum_in = 0;
    prior_sum_out = 0;
    prior_n_in = 0;
    prior_n_out = 0;

    for( int i = 0; i < img_size; i++ )
    {
        if( phi[i] > 0 )
        {
            prior_sum_out += int(prior_img_data[i]);
            prior_n_out++;
        }
        else
        {
            prior_sum_in += int(prior_img_data[i]);
            prior_n_in++;
        }
    }

    return;
}

void Hybrid::calculate_means()
{
    // protection against /0
    if( n_out == 0 )
    {
        Cout = 0;
    }
    else
    {
        Cout = sum_out/n_out;
    }

    // protection against /0
    if( n_in == 0 )
    {
        Cin = 255;
    }
    else
    {
        Cin = sum_in/n_in;
    }

    // protection against /0
    if( prior_n_out == 0 )
    {
        prior_Cout = 0;
    }
    else
    {
        prior_Cout = prior_sum_out/prior_n_out;
    }

    // protection against /0
    if( prior_n_in == 0 )
    {
        prior_Cin = 255;
    }
    else
    {
        prior_Cin = prior_sum_in/prior_n_in;
    }

    return;
}



void Hybrid::do_specific_in_step(int i, int x, int y)
{
    sum_out -= I;
    n_out--;

    sum_in += I;
    n_in++;

    prior_sum_out -= P;
    prior_n_out--;

    prior_sum_in += P;
    prior_n_in++;

    return;
}

void Hybrid::do_specific_out_step(int i, int x, int y)
{
    sum_in -= I;
    n_in--;

    sum_out += I;
    n_out++;

    prior_sum_in -= P;
    prior_n_in--;

    prior_sum_out += P;
    prior_n_out++;

    return;
}

void Hybrid::do_specific_in_step2(int i, int x, int y)
{
    // intensity of the current pixel
    I = int(img_data[i]);

    sum_out -= I;
    n_out--;

    sum_in += I;
    n_in++;

    // intensity of the current pixel in the prior
    P = int(prior_img_data[i]);

    prior_sum_out -= P;
    prior_n_out--;

    prior_sum_in += P;
    prior_n_in++;


    return;
}

void Hybrid::do_specific_out_step2(int i, int x, int y)
{
    // intensity of the current pixel
    I = int(img_data[i]);

    sum_in -= I;
    n_in--;

    sum_out += I;
    n_out++;

    // intensity of the current pixel in the prior
    P = int(prior_img_data[i]);

    prior_sum_in -= P;
    prior_n_in--;

    prior_sum_out += P;
    prior_n_out++;

    return;
}

void Hybrid::do_specific_beginning_step()
{
    calculate_means();

    return;
}

int Hybrid::get_Cout() const
{
    return Cout;
}

int Hybrid::get_Cin() const
{
    return Cin;
}

// initilization for video tracking
void Hybrid::initialize_for_each_frame()
{
    do_general_initialization_for_each_frame();
    initialize_sums();
    calculate_means();

    return;
}

// initilization for video tracking
void Hybrid::initialize_for_each_frame(const unsigned char* img_data1)
{
    img_data = img_data1;
    this->initialize_for_each_frame();

    return;
}

int Hybrid::get_prior_Cout() const
{
    return prior_Cout;
}

int Hybrid::get_prior_Cin() const
{
    return prior_Cin;
}

}
