﻿/****************************************************************************
**
**
****************************************************************************/

#ifndef HYBRID_HPP
#define HYBRID_HPP

#include "activecontour.hpp"
#include "filters.hpp"
#include "../vfc/vfc.h"

namespace ofeli
{

class Hybrid : public ActiveContour
{

public :

    //! Constructor to initialize the active contour with a centered rectangle and the default values of the algorithm parameters.
    Hybrid(const unsigned char* img_data1, int img_width1, int img_height1);

    //! Constructor to initialize the active contour from geometrical parameters of an unique shape, an ellipse or a rectangle.
    Hybrid(const unsigned char* img_data1, int img_width1, int img_height1, 
           bool hasEllipse1, double init_width1, double init_height1, double center_x1, 
           double center_y1, bool hasSmoothingCycle1, int kernel_length1, double sigma1, 
           bool hasCleaning1, int Na1, int Ns1, int lambda_out1, int lambda_in1);

    //! Constructor to initialize the active contour from an initial phi level-set function.
    Hybrid(const unsigned char* img_data1, int img_width1, int img_height1, 
           const char* phi_init1, bool hasSmoothingCycle1, int kernel_length1, double sigma1, 
           bool hasCleaning1, int Na1, int Ns1, int lambda_out1, int lambda_in1, 
           const unsigned char* prior_img_data, double weightCV, double weightPrior, double weightVFC,
           double gammaVFC, dmat vfc_x1, dmat vfc_y1, bool verbose);

    //! Initialization for each new frame buffer, used for video tracking.
    void initialize_for_each_frame();

    //! Initialization for each new frame buffer, used for video tracking.
    void initialize_for_each_frame(const unsigned char* img_data1);

    // 
    // Chan & Vese 
    // 

    //! Getter function for #Cout
    int get_Cout() const;
    //! Getter function for #Cin
    int get_Cin() const;

    // Prior

    //! Getter function for #Cout
    int get_prior_Cout() const;
    //! Getter function for #Cin
    int get_prior_Cin() const;


private :

    //! Computes total external speed 
    int compute_external_speed_Fd(int i, int x, int y);

    //
    // Chan & Vese
    //

    //! Initializes the variables #sum_in, #sum_out, #n_in and #n_out with scanning through the image.
    void initialize_sums();

    //! Calculates means #Cout and #Cin in \a O(1) or accounting for the previous updates of #sum_out and #sum_in, in \a O(#lists_length) and not in \a O(#img_size).
    void calculate_means();

    //! Computes external speed \a Fd with the Chan-Vese model for a current point \a (x,y) of #Lout or #Lin.
    int compute_external_speed_Fd_Chan_Vese(int i, int x, int y);

    //! Calculates means #Cout and #Cin in \a O(1) or accounting for the previous updates of #sum_out and #sum_in, in \a O(#lists_length) and not in \a O(#img_size).
    void do_specific_beginning_step();

    //! Updates variables #sum_in, #sum_out, #n_in and #n_out, before each #switch_in, in the cycle 1, in order to calculate means #Cout and #Cin.
    void do_specific_in_step(int i, int x, int y);

    //! Updates variables #sum_in, #sum_out, #n_in and #n_out, before each #switch_out, in the cycle 1, in order to calculate means #Cout and #Cin.
    void do_specific_out_step(int i, int x, int y);

    //! Updates variables #sum_in, #sum_out, #n_in and #n_out, before each #switch_in, in the cycle 2, in order to calculate means #Cout and #Cin.
    void do_specific_in_step2(int i, int x, int y);

    //! Updates variables #sum_in, #sum_out, #n_in and #n_out, before each #switch_out, in the cycle 2, in order to calculate means #Cout and #Cin.
    void do_specific_out_step2(int i, int x, int y);

    //! Input pointer on the grayscale or grey-level image data buffer. This buffer must be row-wise.
    const unsigned char* img_data;

    //! Mean of the intensities or grey-levels of the pixels outside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) >0\f$ .
    int Cout;
    //! Mean of the intensities or grey-levels of the pixels inside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) <0\f$ .
    int Cin;

    //! Intensity or grey-level of the current pixel.
    int I;

    //! Weight of the outside homogeneity criterion in the Chan-Vese model
    const int lambda_out;
    //! Weight of the inside homogeneity criterion in the Chan-Vese model
    const int lambda_in;

    //! Sum of the intensities or grey-levels of the pixels outside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) >0\f$ .
    int sum_out;
    //! Sum of the intensities or grey-levels of the pixels intside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) <0\f$ .
    int sum_in;

    //! Number of pixels outside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) >0\f$ .
    int n_out;
    //! Number of pixels inside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) <0\f$ .
    int n_in;

    double weightCV;
    double weightPrior;
    double weightVFC;

    inline int sign( double x )
    {
      if (x == 0)
        return 0;

      if (x < 0)
        return -1;
      else
        return 1;
    }

  // Prior term, computed as Chan & Vese term on the prior image

    //! Computes external speed \a Fd with the Chan-Vese model for a current point \a (x,y) of #Lout or #Lin.
    int compute_external_speed_Fd_Prior(int i, int x, int y);

    //! Input pointer on the grayscale or grey-level image data buffer. This buffer must be row-wise.
    const unsigned char* prior_img_data;

    //! Mean of the intensities or grey-levels of the pixels outside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) >0\f$ .
    int prior_Cout;
    //! Mean of the intensities or grey-levels of the pixels inside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) <0\f$ .
    int prior_Cin;

    //! Intensity or grey-level of the current pixel.
    int P;

    //! Weight of the outside homogeneity criterion in the Chan-Vese model
    static const int prior_lambda_out = 1;
    //! Weight of the inside homogeneity criterion in the Chan-Vese model
    static const int prior_lambda_in = 1;

    //! Sum of the intensities or grey-levels of the pixels outside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) >0\f$ .
    int prior_sum_out;
    //! Sum of the intensities or grey-levels of the pixels intside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) <0\f$ .
    int prior_sum_in;

    //! Number of pixels outside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) >0\f$ .
    int prior_n_out;
    //! Number of pixels inside the curve, i.e. pixels \f$i\f$ with \f$\phi \left( i\right) <0\f$ .
    int prior_n_in;
  
    // Vector Field Convolution term
    dmat vfc_x;
    dmat vfc_y;

    double compute_external_speed_Fd_VFC(int i, int x, int y);

    cv::Vec2i evaluate_gradient(const int offset ) const;

    bool null_neighbor_norms(int offset);

};



}



#endif // HYBRID_CPP
