function phi = ac_new_hybrid_model(P, phi, propagation_weight, GAC_weight, g, ...
    delta_t, n_iters, show_result)
% FUNCTION U = AC_HYBRID_MODEL(P, U, PROPAGATION_WEIGHT, GAC_WEIGHT, G, ...
%     DELTA_T, N_ITERS, SHOW_RESULT)
% The hybrid method combining both boundary and regional terms. PDE for
% this model is u_t = \alpha * P * |\nabla u| + \beta *  div(g \nabla u /
% |\nabla u|)|\nabla u|.
%   Inputs:
%       P --- Propagation field indicating the magnitues and directions of
%           evolving curves. Positive value for expansion and negative for
%           contraction.
%       U --- Initial level set.
%       PROPAGATION_WEIGHT --- \alpha in the PDE.
%       GAC_WEIGHT --- \beta in the PDE.
%       G --- Gradient map related to GAC. (default value 1 for curvature
%           flow)
%       DELTA_T --- Time step for each iteration (default value 1).
%       N_ITERS --- Maximum number of iterations (default value 50).
%       SHOW_RESULT --- 1 for showing the results after each iteration
%       (only for 2D image) (default value 0).
%   Output:
%       U --- resulting level set after DELTA_T*N_ITERS.
%
%
% Reference:
%   Y. Zhang, B. J. Matuszewski, L.-K. Shark, C. J. Moore, Medical Image
%       Segmentation Using New Hybrid Level-Set Method. IEEE International
%       Conference on Biomedical Visualisation, MEDi08VIS, London, pp.71-76,
%       July, 2008.
%
%
%%
if nargin < 5 || isempty(g);
    g = 1;
end

if nargin < 6 || isempty(delta_t)
    delta_t = 1;
end

if nargin < 7 || isempty(n_iters)
    n_iters = 50;
end

if nargin < 8 || isempty(show_result)
    show_result = 0;
end

% Displaying of evolution process only works for 2D image.
if ndims(P) ~= 2
    show_result = 0;
end

P = im2double(P);

%%
for i = 1:n_iters
    % phi has been changed when called to this function: phi = phi-.5
    phi_old = phi;
    
    %     Re-initialisation of an embedding
    % function phi is the process to make |phi| = 1
    % while remaining the embedded curve (the zero set)
    % unchanged. One of the efficient methods for reinitialisation
    % is the fast marching method.
    phi = ac_reinit(phi);
    
    % Update phi with the predefined time step.
    % - delta_t: time step for each iteration
    % - propagation_weight: alpha parameter of the Chan-Vese approach (\alpha in the PDE).
    % - P: (target_image - mu), being mu a predefined parameter indicating
    % the lower bound of the gray-level of the target object.
    % In this paper, the target objects are always assumed to have
    % relatively high gray-level values for convenience of discussion.
    phi = phi + delta_t*propagation_weight*P;
    if GAC_weight
        phi = ac_reinit(phi);
        % - g: gradient map related to GAC
        % - GAC_weight: \beta in the PDE
        if isscalar(g)
            phi = ac_linear_diffusion_AOS(phi, delta_t*GAC_weight*g);
        else
            phi = ac_div_AOS(phi, g, delta_t*GAC_weight);
%             phi = ac_reinit(phi); 
%             phi = phi + delta_t*(-1)*g;
        end
    end
    
    % Show evolution process.
    if show_result
        if i == 1
            figure, imshow(P,[]); hold on;
        end
        c = contours(phi,[0,0]);
        if isempty(c), break; end
        if exist('h','var') && all(ishandle(h)), delete(h); end;
        h = zy_plot_contours(c,'linewidth',2);
        set(gcf,'name',sprintf('#iters = %d',i));
        drawnow;
    end
    
    % Terminate the process if no more evolution.
    if isequal(phi > 0, phi_old > 0), break; end
end