function [ANOVA_pValue,c] = ...
    parametricMoreThanTwoMethods(M,significanceLevel,problemType,methodUsed,...
    outputPath,inputFile,numberOfMethodsToCompare, saveFigures, fileID)

%% MORE THAN TWO METHODS TO COMPARE

%% ONE WAY ANOVA
% Note: I'm not using anova2 (two-way analysis of variance) because for this test
% the data in different columns represent changes in one
% factor and the data in different rows represent changes in the other factor. But, in our data, not
% all rows indicate a possible influence factor in the results, i.e. a
% group of rows (a function/dataset/problem) it's a factor, but not every
% single row: we're not using averages, we're using all the results.

% anova1(X) performs balanced one-way ANOVA for comparing the means
% of two or more columns of data in the matrix X, where each column
% represents an independent sample containing mutually independent
% observations. The function returns the p value under the null
% hypothesis that all samples in X are drawn from populations with
% the same mean.
[ANOVA_pValue,t,st] = anova1(M,methodUsed);
doPairwise = 0;
if ANOVA_pValue < significanceLevel
    label = sprintf('ANOVA: There are significant differences between the means of the methods');
    disp(label);
    doPairwise = 1;
else
    label = sprintf('ANOVA: There are NOT significant differences between the means of the methods');
    disp(label);
end;
if saveFigures
    title('One-way analysis of variance: ANOVA');
    saveas(gcf,strcat(outputPath,'ANOVAplot_',inputFile,'.jpg'))
end;
close all;
fprintf(fileID,strcat(label,'\n'));
label = sprintf('p-value: %d\n',ANOVA_pValue);
fprintf(fileID,strcat(label,'\n'));

%% POST-HOC ANALYSIS USING TUKEY-KRAMER CORRECTION (NXN, no control method)
% multcompare(stats) performs a multiple comparison test using the
% information in the stats structure, and returns a matrix c of
% pairwise comparison results.
correction = 'tukey-kramer'; % The Nemenyi test is similar to Tukey's range test in ANOVA.
[c,m,h,nms] = multcompare(st,'display','on','alpha',significanceLevel,'ctype',correction);
if doPairwise
    % How to analyze results?
    % For example, if one row in c contains the following entries.
    % 2.0000  5.0000  1.9442  8.2206  14.4971
    % These numbers indicate that the mean of group 2 minus the mean
    % of group 5 is estimated to be 8.2206, and a 95% confidence
    % interval for the true difference of the means is [1.9442,
    % 14.4971]. In this example the confidence interval does not
    % contain 0.0, so the difference is significant at the 0.05 level.
    % If the confidence interval did contain 0.0, the difference would
    % not be significant at the 0.05 level.
    
    differencesArray = ~(0.0 > c(1:size(c,1),3) & 0.0 < c(1:size(c,1),5));
    pairWise_StatisticalDifferences = methodUsed(c(differencesArray,1:2));
    label = sprintf('Pairwise Statistical Differences using ANOVA with the %s correction: ', correction);
    disp(label);
    fprintf(fileID,strcat(label,'\n'));
    temporal = c(differencesArray,1:2);
    if ~isempty(temporal)
        for i=1:size(pairWise_StatisticalDifferences,1)
            if problemType == 1
                [valuePairWise, indexPairWise]=sort([mean(M(:,temporal(i,1))),mean(M(:,temporal(i,2)))],'descend');
            else
                [valuePairWise, indexPairWise]=sort([mean(M(:,temporal(i,1))),mean(M(:,temporal(i,2)))],'ascend');
            end;
            label = sprintf('%s | %s -> %s',char(pairWise_StatisticalDifferences(i,1)),char(pairWise_StatisticalDifferences(i,2)),char(pairWise_StatisticalDifferences(i,indexPairWise(1))));
            disp(label)
            fprintf(fileID,strcat(label,'\n'));
        end;
    else
        label = sprintf('There are NOT pairwise differences between methods');
        disp(label);
        fprintf(fileID,strcat(label,'\n'));
    end;
    if saveFigures
        title(sprintf('Comparison Intervals Plot of the results obtained by each method using the %s correction',correction))
        saveas(gcf,strcat(outputPath,'ComparisonIntervalsPlot_',inputFile,'.jpg'))
    end;
    close all;
    
    %% HOLM-SIDAK PROCEDURE FOR MULTIPLE STUDENT'S T-TESTS
    disp(' ')
    disp('ALTERNATIVE CHECK: ')
    label = sprintf('Holm-Sidak procedure for multiple Student''s t tests');
    disp(label);
    disp('Methods used: ')
    disp(methodUsed)
    
    M_test = reshape(M,size(M,1)*size(M,2),1);
    AA = repmat([1],size(M,1),1);
    for j=2:numberOfMethodsToCompare
        AA = [AA;repmat([j],size(M,1),1)];
    end;
    newM = [M_test, AA];
    holm(newM)
end;
fprintf(fileID,'\n');
